/*
 *     nanoMir
 *
 *     "db.c"           Databaser for X
 */
#define XLIB_ILLEGAL_ACCESS

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xresource.h>

#include "nm.h"
#include "xi.h"
#include "db.h"
#ifdef UNIX
#include <pwd.h>
#define APP_DEFAULTS_PATH "/usr/lib/X11/app-defaults/"
#define X_DEFAULTS_FILE "/.Xdefaults"
#endif
#ifdef VMS
#define APP_DEFAULTS_PATH "SYS$LOGIN:"
#define VMS_DEFAULTS_FILE "_DEFAULT.DAT"
#endif

extern char *getenv (char *);

local char *
homeDir (char *dest)
{
#ifdef UNIX
#ifdef SVR3
  int uid;
  extern int getuid ();
  extern struct passwd *getpwuid (int);
  extern struct passwd *getpwnam (char *);
#endif
#ifdef MIM_SVR4
  uid_t uid;
#endif
  struct passwd *pw;
  register char *p;

  if ((p = getenv ("HOME")) != NULL)
    (void) strcpy (dest, p);
  else
    {
      if ((p = getenv ("USER")) != NULL)
	pw = getpwnam (p);
      else
	{
	  uid = getuid ();
	  pw = getpwuid (uid);
	}
      if (pw)
	(void) strcpy (dest, pw->pw_dir);
      else
	*dest = ' ';
    }
#endif
#ifdef VMS
  (void) strcpy (dest, "SYS$LOGIN:");
#endif
  return dest;
}

local XrmDatabase rDB;
local void
getDatabases ()
{
  XrmDatabase homeDB, serverDB, appDB;
  char fnbuf[1024];
  char *fn = fnbuf;
  char *env;
  char name[255];

  (void) strcpy (name, APP_DEFAULTS_PATH);
  (void) strcat (name, NMCLASS);
#ifdef VMS
  (void) strcat (name, VMS_DEFAULTS_FILE);
#endif
  /* get application defaults file, if any */
  appDB = XrmGetFileDatabase (name);
  (void) XrmMergeDatabases (appDB, &rDB);

#ifdef UNIX
  /* MERGE server defaults, these ara created by xrdb, loaded as a property
   * of the root window when the server initializes, and loaded into
   * the display structure on XOpenDisplay. If not defined, use .Xdefaults
   */
  if (dpy->xdefaults != NULL)
    serverDB = XrmGetStringDatabase (dpy->xdefaults);
  else
    {
      /* Open .Xdefaults file and merge into existing database */
      (void) homeDir (fn);
      (void) strcat (fn, X_DEFAULTS_FILE);
      serverDB = XrmGetFileDatabase (fn);
    }
  XrmMergeDatabases (serverDB, &rDB);

  /* Open XENVIRONMENT file, or if not defined, the ~/.Xdefaults,
     and merge into existing database */
  if ((env = getenv ("XENVIRONMENT")) == NULL)
    {
      env = homeDir (fn);
      (void) strcat (env, "/.Xdefaults");
    }
  homeDB = XrmGetFileDatabase (env);
  XrmMergeDatabases (homeDB, &rDB);
#endif
}

#define T_STRING 1
#define T_INT    2
#define T_BOOL   3
struct ptable
{
  char *name;
  char *class;
  small type;
  struct dbvalue *val;
}
pTable[] =
{
  {
  "xnm.geometry", "Xnm.Geometry", T_STRING, &dbGeostr,}
  ,
  {
  "xnm.iconGeometry", "Xnm.IconGeometry", T_STRING, &dbIconGeostr,}
  ,
  {
  "xnm.foreground", "Xnm.Foreground", T_STRING, &dbForeground,}
  ,
  {
  "xnm.background", "Xnm.Background", T_STRING, &dbBackground,}
  ,
  {
  "xnm.cursorForeground", "Xnm.CursorForeground", T_STRING, &dbTcForeground,}
  ,
  {
  "xnm.cursorBackground", "Xnm.CursorBackground", T_STRING, &dbTcBackground,}
  ,
  {
  "xnm.statusForeground", "Xnm.StatusForeground", T_STRING, &dbStForeground,}
  ,
  {
  "xnm.statusBackground", "Xnm.StatusBackground", T_STRING, &dbStBackground,}
  ,
  {
  "xnm.errorForeground", "Xnm.ErrorForeground", T_STRING, &dbErForeground,}
  ,
  {
  "xnm.errorBackground", "Xnm.ErrorBackground", T_STRING, &dbErBackground,}
  ,
  {
  "xnm.iconForeground", "Xnm.IconForeground", T_STRING, &dbIcForeground,}
  ,
  {
  "xnm.iconBackground", "Xnm.IconBackground", T_STRING, &dbIcBackground,}
  ,
  {
  "xnm.borderWidth", "Xnm.BorderWidth", T_INT, &dbBorderWidth,}
  ,
  {
  "xnm.pad", "Xnm.Pad", T_INT, &dbPad,}
  ,
  {
  "xnm.font", "Xnm.Font", T_STRING, &dbFontName,}
  ,
  {
  NIL, NIL, 0, NIL,}
,};

void
dbGetPreferences ()
{
  char *stype[256];
  char buf[256];
  XrmValue value;
  register struct ptable *p;
  Bool s;

  getDatabases ();
  for (p = pTable; p->name; p++)
    {
      s = XrmGetResource (rDB, p->name, p->class, stype, &value);
      switch (p->type)
	{
	case T_STRING:
	  if (s)
	    {
	      *sncpy (value.addr, p->val->dbv_value, value.size) = 0;
	      p->val->dbv_changed = TRUE;
	    }
	  else
	    p->val->dbv_changed = FALSE;
	  break;
	case T_INT:
	  if (s)
	    {
	      char buf[128];
	      *sncpy (value.addr, buf, value.size) = 0;
	      dtol (buf, (large *) p->val->dbv_value);
	      p->val->dbv_changed = TRUE;
	    }
	  else
	    p->val->dbv_changed = FALSE;
	  break;
	}
    }
}
