/*
 *      nanoMir
 *
 *      "fs_unix.c      File system (UNIX)
 */

#include "nm.h"
#include "fs.h"
#include "dq.h"
#include "tx.h"
#include "me.h"
#include <stdio.h>
#include <errno.h>
#ifdef MIM_SVR4
#include <unistd.h>
#endif
#ifdef SVR3
#define W_OK    02
#endif

char fsDefExt[MAXFILELN + 1];


small Fsddlen;
small Fsnmlen;
small Fsext;
bool Fsiswild;			/* Wildcard present (*,?) */

char *FsName;
char *FsUname;
char *FsGname;

static char Fs_buf[MAXPATH];

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <dirent.h>
#include <grp.h>
#include <pwd.h>

static void Fs_getpwuid (ushort);
static void Fs_getgrgid (ushort);

static deq *pw_deq;
static small pw_count;

struct pw_name
{
  ushort uid;
  char name[MAXUSRLN];
};

static deq *gr_deq;
static small gr_count;

struct gr_name
{
  ushort gid;
  char name[MAXGRPLN];
};

static char cwdbuf[MAXPATH + 2], *cwdptr;
extern char *getcwd ();
static struct stat statbuf;

void
fsInit ()
{
  if ((cwdptr = getcwd (cwdbuf, MAXPATH + 2)) == NULL)
    otkaz ("getcwd fail");
  tzset ();
  pw_deq = NIL;
  gr_deq = NIL;
}

bool
FsParse (register char *ip, register char *op)
{
  register char c;
  register small i;
  register char *iop;

  iop = op;
  Fsiswild = FALSE;

  if (*ip == '/')
    ip++;
  else
    op = scpy (cwdptr, op);

  *op++ = '/';

  for (;;)
    {
      Fsddlen = op - iop;
      if ((*ip == '.') && ((c = *(ip + 1)) == '/' || c == 0))
	goto dots;
      else
	if ((*ip == '.') && (*(ip + 1) == '.')
	    && ((c = *(ip + 2)) == '/' || c == 0))
	{
	  if (--op == iop)
	    {
	      c = FE_NSDIR;
	      goto err;
	    }
	  while (*--op != '/');
	  op++;
	  ip++;			/* Skip second dot */
	dots:
	  ip++;			/* Skip first dot */
	  if (c)
	    ip++;		/* If slash then skip it */
	}
      else
	{
	  i = MAXFILELN;
	  for (;;)
	    {
	      *op++ = c = *ip++;
	      if (c == 0)
		{
		  Fsnmlen = (op - iop) - Fsddlen - 1;
		  goto ok;
		}
	      else if (c == '/')
		{
		  if (Fsiswild)
		    goto synterr;
		  break;
		}
	      else
		{
		  if ((c == '*') || (c == '?'))
		    Fsiswild++;
		  if (--i < 0)
		    goto synterr;
		}
	    }
	}
    }
ok:
  return (TRUE);
synterr:c = FE_SYNT;
err:errno = c;
  return (FALSE);
}

static int umodes[] =
  { O_RDONLY, O_WRONLY, O_RDWR, O_WRONLY | O_CREAT | O_TRUNC };
fid
FsOpen (char *filnam, register small mode)
{
  mode = open (filnam, umodes[mode], 0644);
  /* mode = open( filnam, umodes[mode], 0755 ); */
  if (mode < 0)
    return 0;
  return mode;
}

large
FsRead (fid f, large size, char *buf)
{
  large s;
  s = read (f, buf, size);
  if (s == -1)
    s = 0;
  return s;
}

large
FsWrite (fid f, large size, char *buf)
{
  return write (f, buf, size);
}

bool
FsClose (fid f)
{
  return !close (f);
}

bool
FsDelete (char *filnam)
{
  int code;
  if (stat (filnam, &statbuf))
    return FALSE;
  if (statbuf.st_mode & S_IFMT == S_IFDIR)
    code = rmdir (filnam);
  else
    code = unlink (filnam);
  return (bool) ! code;
}

static char rbuf[MAXPATH];
static struct stat stfo, stfn;
char *
FsRename (char *fo, char *fn)
{
  if (stat (fo, &stfo))
    return NIL;
  if (stat (fn, &stfn))
    {
      if (errno != ENOENT)
	return NIL;
    }
  else
    {
      if (stfo.st_dev != stfn.st_dev)
	return NIL;
      if (stfo.st_ino == stfn.st_ino)
	return NIL;
      if (stfn.st_mode & S_IFMT == S_IFDIR)
	return NIL;
    }
  unlink (fn);
  if (link (fo, fn))
    return NIL;
  if (unlink (fo))
    return NIL;
  if (FsParse (fn, rbuf))
    return rbuf;
  else
    return NIL;
}

bool
FsInfo (register txt * t)
{
  register struct stat *d;

  if (!((t->txstat & TS_DIRLST) && (t->fiddl == 1)))
    {
      if (t->txstat & TS_DIRLST)
	*sncpy (t->finame, Fs_buf, t->fiddl - 1) = 0;
      else
	*scpy (t->finame, Fs_buf) = 0;

      d = &statbuf;
      if (stat (Fs_buf, d))
	{
	  Fs_buf[t->fiddl - 1] = 0;
	  if (!stat (Fs_buf, d) && !access (Fs_buf, W_OK))
	    errno = FE_NSF;
	  return FALSE;
	}
      t->fiattr = d->st_mode;
      t->fisize = d->st_size;
      t->fitime = d->st_mtime;
      if (!(t->txstat & TS_DIRLST) && (t->fiattr & 040000))
	{
	  if (*(t->finame + t->fiddl + t->finml - 1) == '/')
	    {
	      *scpy ("*", t->finame + t->fiddl + t->finml) = 0;
	      t->fiddl += t->finml;
	    }
	  else
	    {
	      *scpy ("/*", t->finame + t->fiddl + t->finml) = 0;
	      t->fiddl += (t->finml + 1);
	    }
	  t->finml = 1;
	  t->txstat |= TS_DIRLST;
	}
    }
  if (t->txstat & TS_DIRLST)
    t->txredit = TXED_NONO;
  else if (access (Fs_buf, W_OK))
    t->txredit = TXED_NONO;
  return TRUE;
}

static char pattern[MAXPATH];
char fullname[MAXPATH];
static DIR *de;

bool
FsSearch (register txt * t, bool qfirst)
{
  register struct stat *d;
  struct dirent *ee;
  if (qfirst)
    {
/*
 * Precompile pattern - convert "?**?" to "??*"
 */
      register int quest = 0, star = 0;
      register char *p = pattern, *s = t->finame + t->fiddl, c;
      do
	switch (c = *s++)
	  {
	  case '*':
	    star++;
	    break;
	  case '?':
	    quest++;
	    break;
	  default:
	    while (quest--)
	      *p++ = '?';
	    quest = 0;
	    if (star)
	      *p++ = '*';
	    star = 0;
	    *p++ = c;
	  }
      while (c);

      *(FsName = sncpy (t->finame, fullname, t->fiddl)) = 0;
      de = opendir (fullname);
      if (de == NULL)
	return FALSE;
    }


  while ((ee = readdir (de)) != NULL)
    {
/*
 * Match
 */
      struct pn
      {
	char *p, *n;
      }
      stack[65];
      register struct pn *stkptr = stack;
      register char *p = pattern, *n = ee->d_name, c, c1;

    match:
      switch (c = *p++)
	{
	case '?':
	  if (*n++ == 0)
	    goto ng;
	  goto match;
	case '\0':
	  if (*n++ == 0)
	    goto ok;
	  goto ng;
	case '*':
	  if ((c = *p) == 0)
	    goto ok;
	  while (c1 = *n++)
	    if (c1 == c)
	      {
		stkptr->p = p - 1;
		stkptr->n = n;
		stkptr++;
		p++;
		goto match;
	      }
	  goto ng;
	default:
	  if (c != *n++)
	    goto ng;
	  goto match;
	}
    ng:if (stkptr == stack)
	continue;
      p = (--stkptr)->p;
      n = stkptr->n;
      goto match;
    }
  closedir (de);
  return FALSE;
ok:
  d = &statbuf;
  *scpy (ee->d_name, FsName) = 0;
  /*    ,    -     link */
  if (stat (fullname, d) && lstat (fullname, d))
    {
      closedir (de);
      endgrent ();
      endpwent ();
      return FALSE;
    }

  t->fiattr = d->st_mode;
  t->fisize = d->st_size;
  t->fitime = d->st_mtime;

  Fs_getpwuid (d->st_uid);
  Fs_getgrgid (d->st_gid);

  return TRUE;
}

void
FsSetDef (register txt * t)
{
  if (t->fiddl > 1)
    {
      *sncpy (t->finame, cwdptr, t->fiddl - 1) = 0;
      chdir (cwdptr);
    }
  else
    {
      *cwdptr = 0;
      chdir ("/");
    }
}
/*
#define chkdeb
*/

bool
FsCheck (register txt * t)
{

#ifdef chkdeb
  printf ("Check %s %s", t->finame, dttm (t->fitime));
#endif

  if (t->txstat & TS_DIRLST)
    *sncpy (t->finame, Fs_buf, (t->fiddl) - 1) = 0;
  else
    *scpy (t->finame, Fs_buf) = 0;
  if (stat (Fs_buf, &statbuf))
    return FALSE;

#ifdef chkdeb
  printf (" %s\n", dttm (statbuf.st_mtime));
#endif

  if (t->fitime < statbuf.st_mtime)
    return FALSE;
  return (TRUE);
}

time_t
FsTime ()
{
  return GetTime ();
}

static char idtmp[8] = "???";
void
Fs_getpwuid (register ushort uid)
{
  struct passwd *getpwent ();
  register struct passwd *ptr;
  register int i;
  register char *p;
  register struct pw_name *unames;

  if (pw_deq == NIL)
    {
      pw_deq = DqNew (DT_NONE, 0, 100);
      pw_count = 0;
      setpwent ();
      for (pw_count = 0;; pw_count++)
	{
	  ptr = getpwent ();
	  if (ptr == NULL)
	    break;
	  DqExtend (pw_deq, sizeof (struct pw_name));
	  unames = (struct pw_name *) pw_deq->dbeg;
	  unames[pw_count].uid = (ushort) ptr->pw_uid;
	  p = scpy (ptr->pw_name, unames[pw_count].name);
	  while (p < unames[pw_count].name + MAXUSRLN)
	    *p++ = ' ';
	}
      endpwent ();
    }
  else
    unames = (struct pw_name *) pw_deq->dbeg;

  for (i = 0; i < pw_count; i++)
    if (unames[i].uid == uid)
      {
	FsUname = unames[i].name;
	return;
      }
  sprintf (idtmp, "%#6d    ", uid);
  FsUname = idtmp;
}

static char grtmp[8] = "???";
void
Fs_getgrgid (register ushort gid)
{
  extern struct group *getgrent ();
  register struct group *ptr;
  register small i;
  register char *p;
  register struct gr_name *gnames;

  if (gr_deq == NIL)
    {
      gr_deq = DqNew (DT_NONE, 0, 100);
      gr_count = 0;
      setgrent ();
      for (gr_count = 0;; gr_count++)
	{
	  ptr = getgrent ();
	  if (ptr == NULL)
	    break;
	  DqExtend (gr_deq, sizeof (struct gr_name));
	  gnames = (struct gr_name *) gr_deq->dbeg;
	  gnames[gr_count].gid = (ushort) ptr->gr_gid;
	  p = scpy (ptr->gr_name, gnames[gr_count].name);
	  while (p < gnames[gr_count].name + MAXGRPLN)
	    *p++ = ' ';
	}
      endgrent ();
    }
  else
    gnames = (struct gr_name *) gr_deq->dbeg;

  for (i = 0; i < gr_count; i++)
    if (gnames[i].gid == gid)
      {
	FsGname = gnames[i].name;
	return;
      }
  sprintf (grtmp, "%#6d    ", gid);
  FsGname = grtmp;
}

bool
FsAccess (char *s, int mode)
{
  return !access (s, mode);
}
