/*
 *     nanoMir
 *
 *     "fs_vms.c        File System for RMS VMS
 */

#include "nm.h"
#include "fs.h"
#include "dq.h"
#include "tx.h"
#include <stdio.h>
#include <string.h>
#include <types.h>
#include <stat.h>
#include <unixlib.h>
#include <unixio.h>
#include <errno.h>
#include <file.h>
#include <fab.h>
#include <nam.h>
#include <rab.h>
#include <rmsdef.h>
char fsDefExt[MAXFILELN + 1] = "";
small Fsddlen;
small Fsnmlen;
small Fsext;
bool Fsiswild;			/* Wildcard present (*,?) */
char *FsName;
char *FsUname;
char *FsGname;
static char Fs_buf[MAXPATH];

static struct FAB fab;
static struct NAM nam;
static struct RAB rab;

char esa[NAM$C_MAXRSS];
char rsa[NAM$C_MAXRSS];

static void Fs_getpwguid (ushort, ushort);	/* ( gid )  -   FsUname */
					    /* ( uid )  -   FsGname */
static char cwdbuf[MAXPATH + 2], *cwdptr;
extern char *getcwd ();
static struct stat statbuf;

void
fsInit ()
{
  if ((cwdptr = getcwd (cwdbuf, MAXPATH + 2)) == NULL)
    otkaz ("getcwd fail\n");
  fab = cc$rms_fab;
  nam = cc$rms_nam;
  rab = cc$rms_rab;

  fab.fab$v_sqo = 1;
  fab.fab$v_cr = 1;
  fab.fab$v_mxv = 1;
  fab.fab$w_mrs = MAXLUP;
  fab.fab$b_org = FAB$C_SEQ;
  fab.fab$b_rfm = FAB$C_VAR;
  fab.fab$l_nam = &nam;
/*nam.nam$v_noconceal = 1;      */
  nam.nam$v_pwd = 1;
  nam.nam$l_esa = esa;
  nam.nam$b_ess = sizeof esa;
  nam.nam$l_rsa = rsa;
  nam.nam$b_rss = sizeof rsa;
  rab.rab$l_fab = &fab;
  rab.rab$b_rac = RAB$C_SEQ;
  rab.rab$v_tpt = 1;
  rab.rab$v_uif = 1;
}

bool
FsParse (register char *ip, register char *op)
{
  register int s, l;
  register char *p;

  fab.fab$l_fna = ip;
  fab.fab$b_fns = slen (ip);
  s = sys$parse (&fab);
  if (!(s & 1))
    {
    err:errno = EVMSERR;
      vaxc$errno = s;
      return FALSE;
    }
  Fsiswild = nam.nam$v_wildcard;
  if ((!nam.nam$v_exp_name) && (!nam.nam$v_exp_type) && (!nam.nam$v_exp_ver))
    if (nam.nam$v_exp_dir)
      {
	Fsiswild = TRUE;
	nam.nam$b_name = 1;
	nam.nam$b_type = nam.nam$b_ver = 2;
	*scpy ("*.*;*", nam.nam$l_name) = 0;
	nam.nam$b_esl = 5 + nam.nam$l_name - nam.nam$l_esa;
	nam.nam$l_type = 1 + nam.nam$l_name;
	nam.nam$l_ver = 3 + nam.nam$l_name;
      }
    else
      {
	s = RMS$_FNF;
	goto err;
      }
/*   if( nam.nam$v_wild_dir ) {   s = RMS$_FNF; goto err; }  */
  if (!Fsiswild)
    {
      s = sys$search (&fab);
      if (!(s & 1) && s != RMS$_FNF)
	goto err;
      *sncpy (nam.nam$l_rsa, op, nam.nam$b_rsl) = 0;
    }
  else
    *sncpy (nam.nam$l_esa, op, nam.nam$b_esl) = 0;
  Fsddlen = nam.nam$b_node + nam.nam$b_dev + nam.nam$b_dir;
  Fsext = nam.nam$b_type + nam.nam$b_ver;
  Fsnmlen = nam.nam$b_name + Fsext;
  return TRUE;
}

fid
FsOpen (char *filnam, register small mode)
{
  register int s;

  fab.fab$l_fna = filnam;
  fab.fab$b_fns = slen (filnam);

#if DBFS
  fprintf (stderr, "FsOpen %d \"%s\"\n", mode, filnam);
  fflush (stderr);
#endif

  switch (mode)
    {
    case FO_READ:
      fab.fab$v_sup = 0;
      fab.fab$v_cif = 0;
      fab.fab$v_tef = 0;
      fab.fab$l_alq = 0;
      fab.fab$w_deq = 0;
      fab.fab$v_get = 1;
      fab.fab$v_put = 0;
      fab.fab$v_trn = 0;
      fab.fab$v_nil = 0;
      s = sys$open (&fab);
      break;
    case FO_NEW:
      fab.fab$v_sup = 1;
      fab.fab$v_cif = 1;
      fab.fab$v_tef = 1;
      fab.fab$l_alq = 1;
      fab.fab$w_deq = 1;
      fab.fab$v_put = 1;
      fab.fab$v_get = 0;
      fab.fab$v_trn = 1;
      fab.fab$v_nil = 0;
      s = sys$create (&fab);
      break;
    default:
      errno = EVMSERR;
      vaxc$errno = 99176;	/* operation not supported by rms extention */
      return FALSE;		/* ( not documented error )                 */
    }
  if (!(s & 1))
    goto err;
  s = sys$connect (&rab);
  if (!(s & 1))
    {
    err:errno = EVMSERR;
      vaxc$errno = s;
      return FALSE;
    }
  return TRUE;
}

large
FsRead (fid f, large size, char *buf)
{
  large l = 0;
  int s;

  rab.rab$w_usz = MAXLUP;
  sys$rewind (&rab);
  while (l < size)
    {
      rab.rab$l_ubf = buf + l;
      s = sys$get (&rab);
      if (!(s & 1))
	break;
      l += rab.rab$w_rsz;
      *(buf + l) = LF;
      l++;
    }
  return l;
}

large
FsWrite (fid f, large size, char *buf)
{
  large l;
  int s;
  char *p;
  l = 0;
  rab.rab$l_rbf = buf;
  s = sys$rewind (&rab);
  if (!(s & 1))
    goto err;
/* s = sys$find( &rab );
   if(!(s&1) && s != RMS$_EOF ) goto err;
   if( s != RMS$_EOF )
   {
      s = sys$truncate( &rab );
      if(!(s&1)) goto err;
   }
*/ while (l < size)
    {
      rab.rab$l_rbf = buf;
      for (p = buf; *p != LF; p++);
      rab.rab$w_rsz = p - buf;
      s = sys$put (&rab);
      if (!(s & 1))
	{
	err:errno = EVMSERR;
	  vaxc$errno = s;
	  return l;
	}
      l += (++p) - buf;
      buf = p;
    }
  return size;
}

bool
FsClose (fid f)
{
  int s;
  errno = 0;
  s = sys$close (&fab);
  if (!(s & 1))
    {
      errno = EVMSERR;
      vaxc$errno = s;
      return FALSE;
    }
  return TRUE;
}

bool
FsDelete (char *filnam)
{
  int s;
  fab.fab$l_fna = filnam;
  fab.fab$b_fns = slen (filnam);
  s = sys$erase (&fab);
  if (!(s & 1))
    {
      errno = EVMSERR;
      vaxc$errno = s;
      return FALSE;
    }
  return TRUE;
}
char *
FsRename (char *fo, char *fn)
{
  int s;
  struct FAB nfab;
  nfab = fab;
  fab.fab$l_fna = fo;
  fab.fab$b_fns = slen (fo);
  nfab.fab$l_fna = fn;
  nfab.fab$b_fns = slen (fn);
  s = sys$rename (&fab, 0, 0, &nfab);
  if (!(s & 1))
    {
      errno = EVMSERR;
      vaxc$errno = s;
/*    fprintf( stderr, "nm:%s\n", strerror( EVMSERR, s ) ); */
      return 0;
    }
  Fsddlen =
    nfab.fab$l_nam->nam$b_node + nfab.fab$l_nam->nam$b_dev +
    nfab.fab$l_nam->nam$b_dir;
  Fsnmlen =
    nfab.fab$l_nam->nam$b_name + nfab.fab$l_nam->nam$b_type +
    nfab.fab$l_nam->nam$b_ver;
  nfab.fab$l_nam->nam$l_esa[nfab.fab$l_nam->nam$b_esl] = 0;
  return nfab.fab$l_nam->nam$l_esa;
/* return (bool)!rename( fo,fn );       */
}

local char *
fsd2f (register txt * t)
{
  register char *p, *p1, *p2;
  static char buf[MAXPATH];

  *(p = sncpy (t->finame, buf, t->fiddl)) = 0;
  for (p1 = p; *p1 != '.' && *p1 != '['; p1--);
  if (*p1 == '[')
    {
      p = buf + t->fiddl + 6;
      for (p2 = buf + t->fiddl - 2; p2 != p1; p2--)
	p2[7] = *p2;
      scpy ("[000000]", p1);
    }
  else
    {
      p--;
      *p1 = ']';
    }
  *scpy (".DIR", p) = 0;
  return buf;
}

bool
FsInfo (register txt * t)
{
  register struct stat *d;
  char *p, *p1;
  if (t->txstat & TS_DIRLST)
    p = fsd2f (t);
  else
    {
      *scpy (t->finame, Fs_buf) = 0;
      p = Fs_buf;
    }
  d = &statbuf;
  if (stat (p, d))
    if (t->txstat & TS_DIRLST)
      {
	t->fiattr = 0;
	t->fisize = 0;
	t->fitime = 0;
      }
    else
      return FALSE;
  else
    {
      t->fiattr = d->st_mode;
      t->fisize = d->st_size;
      t->fitime = d->st_mtime;
    }
  if (!(t->txstat & TS_DIRLST) && (t->fiattr & 040000))
    {
      p = t->finame + t->fiddl - 1;
      *p++ = '.';
      for (p1 = p; *p1 != '.' && *p1 != ';' && *p1 != '\0'; p1++);
      *p1++ = ']';
      t->fiddl = ((char *) p1 - (char *) (t->finame));
      *scpy ("*.*;*", p1) = 0;
      t->finml = 5;
      t->txstat |= TS_DIRLST;
    }
  if (t->txstat & TS_DIRLST)
    t->txredit = TXED_NONO;
  else if (!FsAccess (Fs_buf, 2))
    t->txredit = TXED_NONO;
  return TRUE;
}

char fullname[MAXPATH];
bool
FsSearch (register txt * t, bool qfirst)
{
  register int s = 1;
  if (qfirst)
    {
      fab.fab$l_fna = t->finame;
      fab.fab$b_fns = slen (t->finame);
      s = sys$parse (&fab);
      if (!(s & 1) /*|| nam.nam$v_wild_dir */ )
	{
	  errno = EVMSERR;
	  /*if( nam.nam$v_wild_dir ) vaxc$errno = RMS$_SYN;
	     else */ vaxc$errno = s;
	  return FALSE;
	}
    }
  s = sys$search (&fab);
  if (!(s & 1))
    {
      errno = EVMSERR;
      vaxc$errno = s;
      return FALSE;
    }
  *sncpy (nam.nam$l_rsa, fullname, nam.nam$b_rsl) = 0;
  FsName = fullname + (nam.nam$l_name - nam.nam$l_rsa);
  if (stat (fullname, &statbuf))
    {
      t->fiattr = 0;
      t->fisize = 0;
      t->fitime = 0;
      Fs_getpwguid (0, 0);
    }
  else
    {
      t->fiattr = statbuf.st_mode;
      t->fisize = statbuf.st_size;
      t->fitime = statbuf.st_mtime;
      Fs_getpwguid (statbuf.st_gid, statbuf.st_uid);
    }
  return TRUE;
}

void
FsSetDef (register txt * t)
{
  register char *p;

  *sncpy (t->finame, cwdptr, t->fiddl) = 0;
  while ((p = strstr (cwdptr, "][")) != 0)
    *scpy (p + 2, p) = 0;
  chdir (cwdptr);
}

bool
FsCheck (register txt * t)
{
  register char *p;
  if (t->txstat & TS_DIRLST)
    return TRUE;
/*if( t->txstat & TS_DIRLST ) return FALSE;
    p = fsd2f( t );
  else
*/
  *scpy (t->finame, Fs_buf) = 0;
  if (stat (Fs_buf, &statbuf))
    return FALSE;
#ifdef chkdeb
  printf (" %s\n", dttm (statbuf.st_mtime));
#endif
#if DBFS
  fprintf (stderr, "FsCheck \"%s\" %d %d\n", Fs_buf, t->fitime,
	   statbuf.st_mtime);
#endif
  return (t->fitime >= statbuf.st_mtime);
}

time_t
FsTime ()
{
  return GetTime ();
}

static char idtmp[10] = "[000,000]";
void
Fs_getpwguid (register ushort gid, register ushort uid)
{
  sprintf (idtmp, "[%03o,%03o]", (unsigned char) gid, (unsigned char) uid);
  FsUname = idtmp;
}

#include <acldef.h>
#include <jpidef.h>
#include <chpdef.h>
#include <dnsdef.h>
#include <descrip.h>
unsigned a1, a2;
char user_name[16];
$DESCRIPTOR (file, user_name);
$DESCRIPTOR (user, user_name);
struct $dnsitmdef file_item[2] = { {4, 1, &a1, &a2}, {0, 0, 0, 0} };
struct $dnsitmdef proc_item[2] =
  { {16, JPI$_USERNAME, user_name, &user.dsc$w_length}, {0, 0, 0, 0} };

bool
FsAccess (char *st, int mode)
{
  int s;
  s = getpid ();
  sys$getjpiw (1, &s, 0, proc_item, 0, 0, 0);
  a1 = mode;
  file.dsc$a_pointer = st;
  file.dsc$w_length = strlen (st);
  s = sys$check_access (&ACL$C_FILE, &file, &user, file_item);
  return s & 1;
}
