/*
 *     
 *
 *     "kt.c"            
 */

#include "nm.h"
#include "kb.h"
#ifdef X11
#include "xi.h"
#endif

#include <stdio.h>
#ifdef UNIX
#include <fcntl.h>
#endif
#ifdef VMS
#include <file.h>
#endif
int columns;
int lines_;
int regis_level;
#ifndef VERSION_2
bool alphanumeric_term;
#endif /* ! VERSION_2 */
char *bell;
char *clear_screen;
char *clr_bol;
char *clr_bos;
char *clr_eol;
char *clr_eos;
char *carriage_return;
char *cursor_address;
char *cursor_home;
char *cursor_down;
char *cursor_invisible;
char *cursor_left;
char *cursor_normal;
char *cursor_right;
char *cursor_up;
char *delete_character;
char *delete_line;
char *enter_alt_charset_mode;
char *enter_attribute_mode;
char *enter_blink_mode;
char *enter_bold_mode;
char *enter_dim_mode;
char *enter_graph_mode;
char *enter_reverse_mode;
char *enter_underline_mode;
char *erase_chars;
char *exit_alt_charset_mode;
char *exit_attribute_mode;
char *exit_graph_mode;
char *insert_character;
char *insert_line;
char *parm_dch;
char *parm_delete_line;
char *parm_ich;
char *parm_insert_line;
char *parm_index;
char *parm_rindex;
char *roll_disable;
char *roll_enable;
char *screen_attach;
char *screen_detach;
char *scroll_forward;
char *scroll_reverse;
char *set_top_and_bottom_margins;
unsigned char *tv_ASCII;
unsigned char *kb_lat_ASCII;
unsigned char *kb_rus_ASCII;

static void del_rd (register token_tabs * tp);

static short
get_one (register token_tabs * tp, register short (*gt) ())
{
  if (tp->next_in_buf == tp->buf_count)
    tp->buf_elem[tp->buf_count++] = (*gt) ();
  return tp->buf_elem[tp->next_in_buf++];
}

static void
del_rd (register token_tabs * tp)
{
  register int i, j, k;
  for (i = 0, j = tp->next_in_buf, k = tp->buf_count; j < k; i++, j++)
    tp->buf_elem[i] = tp->buf_elem[j];
  tp->buf_count = i;
  tp->next_in_buf = 0;
}

static void
del_one (register token_tabs * tp)
{
  register int i;
  for (i = 1; i < tp->buf_count; i++)
    tp->buf_elem[i - 1] = tp->buf_elem[i];
  tp->buf_count--;
  tp->next_in_buf = 0;
}

short
get_token (register token_tabs * tp, register short (*gt) ())
{

  register int i, j, k, from, to;
  register short c, cc;
  for (;;)
    {
      c = get_one (tp, gt);

      /*+++
         printf("*** get_token: c=%o \n", c);
         +++ */

#ifndef X11
      if ((c & 0140) && !(c & ~0377) && (c != 0177))
	{
#else
      if ((c & 0140) && !(c & ~0377) && (c != 0177) && (c != 0377))
	{
#endif
	  del_one (tp);
	  return c;
	}
      from = 0;
      to = tp->table_pointer[0];
      /*                 */
      /*   from -    */
      /*   to -       */
      for (;;)
	{
	  for (i = from, k = to + 1; i < k - 1;)
	    {
	      j = (i + k) >> 1;
	      if (c > tp->big_tab[j - 1].token)
		i = j;
	      else
		k = j;
	    }
	  j = i;

	  if (j <= to && tp->big_tab[j].token == c)
	    {
	      cc = tp->big_tab[j].code;
	      if (cc >= 0)
		{
		  del_rd (tp);
		  return cc;
		}
	      else
		{
		  i = -cc;
		  from = (tp->table_pointer[i - 1]) + 1;
		  to = tp->table_pointer[i];
		  c = get_one (tp, gt);
		}
	    }
	  else
	    {
	      for (i = 0; i < tp->special_count; i++)
		{
		  if (tp->buf_elem[0] == tp->special_tab[i].token)
		    {
#ifdef  VERSION_2
		      if (tp->next_in_buf > 1)
			tp->next_in_buf--;
#else
		      tp->next_in_buf--;
#endif /* VERSION_2 */
		      del_rd (tp);
		      return tp->special_tab[i].code;
		    }
		}
	      del_one (tp);
	      break;
	    }
	}
    }
}

#define tread(w) read(fd, (char *) &(w), sizeof(w))

int
load_table (char *file_name, token_tabs * tp)
{
  short tab_count, i, j, k;
  int fd;
  tab_elem t;
  if (access (file_name, 04) != 0)
    return -1;
  fd = open (file_name, O_RDONLY);
  tread (tab_count);
  tread (i);
  tp->table_pointer = (short *) GetMain (sizeof (short) * tab_count);
  tp->big_tab = (tab_elem *) GetMain (sizeof (tab_elem) * i);
  i = 0;
  for (i = 0, j = 0; j < tab_count; j++)
    {
      tread (k);
      while (k--)
	{
	  tread (t);
	  tp->big_tab[i++] = t;
	}
      tp->table_pointer[j] = i - 1;
    }
  tread (tp->special_count);
  tp->special_tab =
    (tab_elem *) GetMain (sizeof (tab_elem) * (tp->special_count));
  for (j = 0; j < tp->special_count; j++)
    {
      tread (t);
      tp->special_tab[j] = t;
    }
  close (fd);
  tp->buf_count = 0;
  tp->next_in_buf = 0;
  return 0;
}

#define tget(oper) oper=mem+p;tread(l);if(l) read(fd,oper,l);p+=l;mem[p]='\000';p++

int
load_codes (char *file_name)
{
  short l, p;
  int fd;
  char *mem;
  if (access (file_name, 04) != 0)
    return -1;
  fd = open (file_name, O_RDONLY);
  tread (columns);
  tread (lines_);
  tread (regis_level);
#ifndef VERSION_2
  if (regis_level == ALPHANUMERIC_TERM)
    alphanumeric_term = TRUE;
  else
    alphanumeric_term = FALSE;
#endif /* ! VERSION_2 */
  tread (l);
  mem = GetMain (l);
  p = 0;

  tget (bell);
  tget (clear_screen);
  tget (clr_bol);
  tget (clr_bos);
  tget (clr_eol);
  tget (clr_eos);
  tget (carriage_return);
  tget (cursor_address);
  tget (cursor_home);
  tget (cursor_down);
  tget (cursor_invisible);
  tget (cursor_left);
  tget (cursor_normal);
  tget (cursor_right);
  tget (cursor_up);
  tget (delete_character);
  tget (delete_line);
  tget (enter_alt_charset_mode);
  tget (enter_attribute_mode);
  tget (enter_blink_mode);
  tget (enter_bold_mode);
  tget (enter_dim_mode);
  tget (enter_graph_mode);
  tget (enter_reverse_mode);
  tget (enter_underline_mode);
  tget (erase_chars);
  tget (exit_alt_charset_mode);
  tget (exit_attribute_mode);
  tget (exit_graph_mode);
  tget (insert_character);
  tget (insert_line);
  tget (parm_dch);
  tget (parm_delete_line);
  tget (parm_ich);
  tget (parm_insert_line);
  tget (parm_index);
  tget (parm_rindex);
  tget (roll_disable);
  tget (roll_enable);
  tget (screen_attach);
  tget (screen_detach);
  tget (scroll_forward);
  tget (scroll_reverse);
  tget (set_top_and_bottom_margins);


  tv_ASCII = (unsigned char *) mem + p;
  read (fd, tv_ASCII, 256);
  p += 256;
  kb_rus_ASCII = (unsigned char *) mem + p;
  read (fd, kb_rus_ASCII, 256);
  p += 256;
  kb_lat_ASCII = (unsigned char *) mem + p;
  read (fd, kb_lat_ASCII, 256);
  close (fd);
  return 0;
}
