/*
 *     nanoMir
 *
 *     "sc.c"           Screen
 */

#include "nm.h"
#include "me.h"
#include "sc.h"
#include "kb.h"
#include "trm.h"
#include "ac.h"
#include "tv.h"
#include <signal.h>
#ifdef VMS
#include <iodef.h>
#include <stdio.h>
#include <errno.h>
#endif
#ifdef UNIX
#include <termio.h>
#endif

#ifdef UNIX
#include <sys/ioctl.h>

#ifdef MIM_SVR4
#include <termios.h>

#if ! defined( TCSBRK ) && defined( TIOCSBRK )
#define TCSBRK TIOCSBRK
#endif

#else
#include <termio.h>
#endif /* def MIM_SVR4 */
#endif /* def UNIX */

tchar *scimage;
#define SCSMAX 2
local tchar *scs[SCSMAX], *scl;
static tchar *scbuff;
static small *lnlen;
small scx, scy;
small cx, cy;
local bool gop;

static tchar cur_attr;
#define MAXTVBUFF 1024
char tv_buffer[MAXTVBUFF + 10];
small tv_buff_count = 0;
local void tv_spaces (small);

local void tv_drawline_h (small, small, small),
tv_drawline_v (small, small, small);
local small get_lnlen (small);

static tchar spaces[MAXLPAC];
static tchar bline[MAXLPAC];

void
scInit ()
{
  register small i;

  for (i = 0; i < MAXLPAC; i++)
    {
      spaces[i] = (tchar) ' ';
      bline[i] = (tchar) WND_H_LINE;
    }
/* if ( !have_tty ) term_attach();      */
  tv_features ();
  scimage = (tchar *) GetMain (scx * scy * sizeof (tchar));
  for (i = 0; i < SCSMAX; i++)
    scs[i] = (tchar *) GetMain (scx * scy * sizeof (tchar));
  scl = (tchar *) GetMain (scx * scy * sizeof (tchar));
  lnlen = (small *) GetMain (scy * sizeof (small));
  cur_attr = 0;
  scAttach ();
  have_tty = TRUE;
  scErase ();
#ifdef COLOR
  SetTextColor ();
#endif
#ifdef ALT_CHAR_SET_FONT
  SetAltCharSetFont ();
#endif
}
void
scAttach ()
{
  tv_oper (exit_alt_charset_mode);
  tv_oper (exit_attribute_mode);
  tv_flush ();
  if (!have_tty)
    kbAttach ();
}
void
scDetach ()
{
  if (have_tty)
    kbDetach ();
}
void
ScTerm ()
{
  if (have_tty)
    {
      tv_flush ();
      KbTerm ();
    }
#ifdef COLOR
  SetDefaultTextColor ();
#endif
#ifdef ALT_CHAR_SET_FONT
  SetAltCharSetFont ();
#endif
}

void
scResize ()
{
  register int i;
  if (columns == scx && lines_ == scy)
    return;
#if 0
  scx = columns;
  scy = lines_;
  scimage = (tchar *) meAlloc (scimage, scx * scy * sizeof (tchar));
  for (i = 0; i < SCSMAX; i++)
    scs[i] = (tchar *) meAlloc (scs[i], scx * scy * sizeof (tchar));
  scl = (tchar *) meAlloc (scl, scx * scy * sizeof (tchar));
  lnlen = (small *) meAlloc (lnlen, scy * sizeof (small));
#endif
}

local bool
qinbox (box b, small x, small y)
{
  return x >= b.x0 && x < b.x1 && y >= b.y0 && y < b.y1;
}

void
scblkon (box block, box win)
{
  register small i, j;

  for (j = win.y0; j < win.y1; j++)
    for (i = win.x0; i < win.x1; i++)
      if (!qinbox (block, i, j) && scim (i, j) & AT_BLOCK)
	{
	  scpos (i, j);
	  tv_char (scim (i, j) &= ~AT_BLOCK);
	}
      else if (qinbox (block, i, j) && !(scim (i, j) & AT_BLOCK))
	{
	  scpos (i, j);
	  tv_char (scim (i, j) |= AT_BLOCK);
	}
}

void
scblkoff ()
{
  small i, j;
  for (j = 0; j < scy; j++)
    for (i = 0; i < scx; i++)
      if (scim (i, j) & AT_BLOCK)
	{
	  scpos (i, j);
	  tv_char (scim (i, j) &= ~AT_BLOCK);
	}
}

bool ScLocked = FALSE;
void
sclock ()
{
  if (!ScLocked)
    tv_flush ();
  ScLocked = TRUE;
}

void
scunlock (bool q)
{
  register small i, j;
  small tcx = cx, tcy = cy;

  ScLocked = FALSE;
  if (q && 1)
    {
      for (j = 0; j < scy; j++)
	{
	  scpos (0, j);
	  for (i = 0; i < scx; i++)
	    tv_char (scim (i, j));
	}
    }
  else
    {
      blkmov ((char *) scimage, (char *) scl, scx * scy * sizeof (tchar));
      scErase ();
      for (j = 0; j < scy; j++)
	sctext (scl + scx * j, scx, 0, j);
    }
  scpos (tcx, tcy);
  tv_flush ();
}

void
scdrwnd (register small x0, register small y0,
	 register small dx, register small dy)
{
  register small x1, y1, m;
  x1 = x0 + dx;
  y1 = y0 + dy;
  --x0;
  --y0;
  m = 0;
  if (x1 >= scx)
    {
      m |= ScRBLK;
      x1 = scx - 1;
    }
  if (y0 <= 0)
    {
      m |= ScUBLK;
      y0 = 0;
    }
  if (x0 <= 0)
    {
      m |= ScLBLK;
      x0 = 0;
    }
  if (y1 >= scy)
    {
      m |= ScDBLK;
      y1 = scy - 1;
    }
  if (!(m & ScUBLK))
    tv_drawline_h (x0, y0, x1 - x0 + 1);
  if (!(m & ScRBLK))
    tv_drawline_v (x1, y0, y1 - y0 + 1);
}

local void
tv_drawline_h (register small x, register small y, register small l)
{
#ifdef COLORB
  SetBorderTextColor ();
#endif
  scmvtext (bline, l, x, y);
  tv_flush ();
#ifdef COLORB
  SetTextColor ();
#endif
}

local void
tv_drawline_v (register small x, register small y, register small l)
{
  while (l--)
    {
      if (scim (x, y) != WND_V_LINE)
	{
	  scpos (x, y);
	  tv_char (WND_V_LINE);
	  scim (x, y) = WND_V_LINE;
	}
      y++;
    }
  tv_flush ();
}

void
scErase ()
{
  register int i;
  cx = 0;
  cy = 0;
  i = scx * scy - 1;
  while (i--)
    scimage[i] = (tchar) ' ';
  for (i = 0; i < scy; i++)
    lnlen[i] = 0;
  if (!ScLocked)
    tv_oper (clear_screen);
}
void
scbell ()
{
  if (!ScLocked)
    tv_oper (bell);
}
void
scurson ()
{
  if (!ScLocked)
    {
      tv_oper (cursor_normal);
      tv_flush ();
    }
}
void
scursoff ()
{
  if (!ScLocked)
    tv_oper (cursor_invisible);
}

#define MAXS 3
void
scpos (register small x, register small y)
{
  register small i;
  small l;
  if (x < 0 || x >= scx || y < 0 || y >= scy)
    return;
  if (cy == y && (l = (x - cx)) >= 0 && l < MAXS)
    {
      for (i = 0; i < l; i++)
	if (scim (cx + i, y) & ~0377)
	  goto fail;
      for (i = 0; i < l; i++)
	tv_char (scim (cx, y));
      return;
    }
fail:
  (*tv_pos) (x, y);
}

local small
sclen (tchar * tp, small l)
{
  tp += l;
  for (; l; l--)
    if (*(--tp) != ' ')
      break;
  return l;
}

local small
scsplen (tchar * tp, small l)
{
  small i;
  for (i = 0; i < l; l++)
    if (*(tp++) != ' ')
      break;
  return i;
}

void
sctext (tchar * tp, small l, small x, small y)
{
  small i, l0, l1, l2, l3, l4, l5;
  small ll, ll0, ll1;
  l1 = sclen (tp, l);
  if (l1 == 0)
    {
      scspaces (x, y, l);
      return;
    }
  l0 = scsplen (tp, l1);
  ll1 = sclen (scimage + (y * scx), scx);
  ll0 = scsplen (scimage + (y * scx), ll1);
  if (ll0 >= x)
    {
      if (ll0 < x + l0)
	{
	  small i;
	  if (!ScLocked)
	    (*tv_clr_bol) (y, x + l0);
	  for (i = 0; i < x + l0; i++)
	    scim (i, y) = ' ';
	}
      tp += l0;
      l -= l0;
      l1 -= l0;
      x += l0;
      l0 = 0;
    }
  if (ll1 <= x + l)
    {
      if (x + l1 < ll1)
	{
	  small i;
	  if (!ScLocked)
	    (*tv_clr_eol) (y, scx - x - l1);
	  for (i = x + l1; i < scx; i++)
	    scim (i, y) = ' ';
	}
      l = l1;
    }
  if (l > 0)
    scmvtext (tp, l, x, y);
  lnlen[y] = get_lnlen (y);
}

void
scmvtext (tchar * tp, small l, small x, small y)
{
  small i;
  tv_mv_text (tp, l, x, y);
  for (i = 0; i < l; i++)
    scim (x++, y) = tp[i];
}

void
tv_mv_text (tchar * tp, small l, small x, small y)
{
  small i;
  if (ScLocked)
    return;
  for (i = 0; i < l; i++)
    {
      if (scim (x, y) != tp[i])
	{
	  scpos (x, y);
	  tv_char (tp[i]);
	}
      x++;
    }
}

local bool sq[SCSMAX];
void
ScLoad (int n)
{
  int i;
  if (!sq[n])
    return;
  for (i = 0; i < scy; i++)
    sctext (scs[n] + scx * i, scx, 0, i);
  sq[n] = FALSE;
}

void
ScSave (int n)
{
  if (sq[n])
    return;
  blkmov ((char *) scimage, (char *) scs[n], scx * scy * sizeof (tchar));
  sq[n] = TRUE;
}


#ifdef RUSSIAN
local char justmin[] = "         . . .";
#endif
#ifdef ENGLISH
local char justmin[] = "Just a moment, please...";
#endif
local small freq;
void
ScJustMin (bool save, bool rdrw, box b)
{
  register char *p;
  register tchar *tp;
  register small x, y, l;
  tchar tbuf[MAXLPAC];
  if (save)
    {
      if (rdrw && !freq)
	{
	  ScSave (1);
	  scErase ();
	  for (p = justmin, tp = tbuf; *p; *tp++ = ctotc (*p++));
	  l = slen (justmin);
	  x = (scx - l) / 2;
	  y = scy / 2;
	  sctext (tbuf, slen (justmin), x, y);
	  if (!ScLocked)
	    tv_flush ();
	}
      freq++;
    }
  else
    {
      if (rdrw && !freq)
	{
	  scErase ();
	  ScLoad (1);
	  if (!ScLocked)
	    tv_flush ();
	}
      else
	freq--;
    }
}

void
scspaces (small x, small y, small l)
{
  small i;
  small ll0, ll1;
  if (l <= 0)
    return;
  ll1 = sclen (scimage + (y * scx), scx);
  ll0 = scsplen (scimage + (y * scx), ll1);
  if (x + l <= ll0 || x >= ll1)
    return;
  if (x <= ll0)
    if (x + l >= ll1)
      {
	if (!ScLocked)
	  (*tv_clr_eol) (y, scx);
      }
    else
      {
	if (!ScLocked)
	  (*tv_clr_bol) (y, x + l);
      }
  else if (x + l >= ll1)
    {
      if (!ScLocked)
	(*tv_clr_eol) (y, scx - x);
    }
  else
    scmvtext (spaces, l, x, y);
  for (i = x; i < x + l; i++)
    scim (i, y) = ((tchar) ' ');
  lnlen[y] = get_lnlen (y);
}

void
scclear (register small x0, register small x1,
	 register small y0, register small y1)
{
  register small i, j;
  small x, y;

  if (!ScLocked)
    {
      x = cx;
      y = cy;
      if (x0 == 0 && x1 == scx)
	if (y1 == scy)
	  (*tv_clr_eos) (y0);
	else if (y0 == 0)
	  (*tv_clr_bos) (y1);
	else
	  (*tv_clr_lines) (y0, y1);
      else
	for (j = y0; j < y1; j++)
	  tv_mv_text (spaces, x1 - x0, x0, j);
    }
  for (i = x0; i < x1; i++)
    for (j = y0; j < y1; j++)
      scim (i, j) = ((tchar) ' ');
  for (j = y0; j < y1; j++)
    lnlen[j] = get_lnlen (j);
  scpos (x, y);
}

void
sctxblt (register small x0, register small x1,
	 register small y0, register small y1,
	 register small dx, register small dy)
{
  register small i, j, k, l, x = cx, y = cy;

  if (dx == 0 && x0 == 0 && x1 == scx)
    {
      if (dy < 0)
	{
	  if (!ScLocked)
	    {
	      gop = TRUE;
	      if (y1 == scy && y0 + dy == 0)
		(*tv_roll_up) (-dy);
	      else if (y1 == scy)
		(*tv_eos_up) (y0, -dy);
	      else
		(*tv_lines_up) (y0, y1, -dy);
	    }
	  for (j = y0; j < y1; j++)
	    {
	      for (i = 0; i < scx; i++)
		scim (i, j + dy) = scim (i, j);
	      lnlen[j + dy] = lnlen[j];
	    }
	  if (gop)
	    for (j = y1 + dy; j < y1; j++)
	      {
		small i;
		for (i = 0; i < scx; i++)
		  scim (i, j) = ' ';
		lnlen[j] = 0;
	      }
	}
      if (dy > 0)
	{
	  if (!ScLocked)
	    {
	      gop = TRUE;
	      if (y0 == 0 && y1 + dy == scy)
		(*tv_roll_down) (dy);
	      else if (y1 + dy == scy)
		(*tv_eos_down) (y0, dy);
	      else
		(*tv_lines_down) (y0, y1, dy);
	    }
	  for (j = y1 - 1; j >= y0; j--)
	    {
	      for (i = 0; i < scx; i++)
		scim (i, j + dy) = scim (i, j);
	      lnlen[j + dy] = lnlen[j];
	    }
	  if (gop)
	    for (j = y0 + dy - 1; j >= y0; j--)
	      {
		small i;
		for (i = 0; i < scx; i++)
		  scim (i, j) = ' ';
		lnlen[j] = 0;
	      }
	}
    }
  else
    {
      if (dy == 0)
	{
	  if (dx < 0)
	    {
	      if (!ScLocked)
		if (x1 == scx)
		  for (j = y0; j < y1; j++)
		    (*tv_eol_left) (x0, j, -dx);
		else
		  for (j = y0; j < y1; j++)
		    (*tv_text_left) (x0, x1, j, -dx);
	      for (j = y0; j < y1; j++)
		{
		  for (i = x0; i < x1; i++)
		    scim (i + dx, j) = scim (i, j);
		  for (i = x1 + dx; i < x1; i++)
		    scim (i, j) = (tchar) (-1);
		  lnlen[j] = get_lnlen (j);
		}
	    }
	  if (dx > 0)
	    {
	      if (!ScLocked)
		if (x1 + dx == scx)
		  for (j = y0; j < y1; j++)
		    (*tv_eol_right) (x0, j, dx);
		else
		  for (j = y0; j < y1; j++)
		    (*tv_text_right) (x0, x1, j, dx);
	      for (j = y0; j < y1; j++)
		{
		  for (i = x1 - 1; i >= x0; i--)
		    scim (i + dx, j) = scim (i, j);
		  for (i = x0; i < x0 + dx; i++)
		    scim (i, j) = (tchar) (-1);
		  lnlen[j] = get_lnlen (j);
		}
	    }
	}
      else
	{
	  i = x1 - x0;
	  for (j = y0; j < y1; j++)
	    tv_remove_text (x0, j, x0 + dx, j + dy, i);
	  if (dx < 0)
	    if (dy < 0)
	      for (j = y0; j < y1; j++)
		for (i = x0; i < x1; i++)
		  scim (i + dx, j + dy) = scim (i, j);
	    else
	      for (j = y1 - 1; j >= y0; j--)
		for (i = x0; i < x1; i++)
		  scim (i + dx, j + dy) = scim (i, j);
	  else if (dy < 0)
	    for (j = y0; j < y1; j++)
	      for (i = x1 - 1; i >= x0; i--)
		scim (i + dx, j + dy) = scim (i, j);
	  else
	    for (j = y1 - 1; j >= y0; j--)
	      for (i = x1 - 1; i >= x0; i--)
		scim (i + dx, j + dy) = scim (i, j);
	}
      for (j = y0 + dy; j < y1 + dy; j++)
	lnlen[j] = get_lnlen (j);
    }
  scpos (x, y);
}

void
tv_remove_text (register small xi, register small yi,
		register small xo, register small yo, register small l)
{
  gop = FALSE;
  tv_mv_text (scimage + yi * scx + xi, l, xo, yo);
}

local void
tv_attr (register tchar c)
{
  if (((~0377) & cur_attr) == c)
    return;
  cur_attr = (cur_attr & 0377) | c;
  tv_oper (exit_attribute_mode);
  if (c & AT_LIGHT)
    tv_oper (enter_dim_mode);
  if (c & AT_ITALIC)
    tv_oper (enter_blink_mode);
  if (c & AT_UNDERL)
    tv_oper (enter_underline_mode);
  if (c & AT_BOLD)
    tv_oper (enter_bold_mode);
  if (!(c & AT_INVERT) ^ !(c & AT_BLOCK))
    tv_oper (enter_reverse_mode);
}

local bool
tv_need_graph_mode (register tchar c_in)
{
  register tchar c = c_in & 0377;
  return c >= LDC_FIRST && c <= LDC_LAST;
}

static bool isSetGraphMode = FALSE;
void
tv_char (register tchar c_in)
{
  register small c;
  bool q;
  if (!ScLocked)
    {
      if (tv_buff_count >= MAXTVBUFF - 16)
	tv_flush ();
      if (q = tv_need_graph_mode (c_in))
	{
	  if (!isSetGraphMode)
	    {
	      tv_oper (enter_graph_mode);
	      isSetGraphMode = TRUE;
	      cur_attr &= ~0200;
	    }
	}
      else
	{
	  if (isSetGraphMode)
	    {
	      tv_oper (exit_graph_mode);
	      isSetGraphMode = FALSE;
	      cur_attr &= ~0200;
	    }
	}
      c = tv_ASCII[c_in & 0377];
      if (!q)
	if ((c & 0200) && !(cur_attr & 0200))
	  {
	    cur_attr |= 0200;
	    tv_oper (enter_alt_charset_mode);
	    isSetGraphMode = FALSE;
	  }
	else if (!(c & 0200) && (cur_attr & 0200))
	  {
	    cur_attr &= ~0200;
	    tv_oper (exit_alt_charset_mode);
	    isSetGraphMode = FALSE;
	  }
      tv_attr (c_in & ~0377);
      if (cy == scy - 1 && cx == scx - 1)
	{
	  tv_oper (roll_disable);
	  tv_write_one ((char) c);
	  tv_oper (roll_enable);
	}
      else
	tv_write_one ((char) c);
    }
  cx++;
}

static void
tv_here (char c)
{
  tv_buffer[tv_buff_count++] = c;
}

void
tv_flush ()
{
  int i;
  char *t;
  void (*p) (int);
  if (tv_buff_count)
    {
#ifdef UNIX
      p = signal (SIGINT, SIG_IGN);
      write (term_file, tv_buffer, (int) tv_buff_count);
      ioctl (term_file, TCSBRK, 1);
      tv_buff_count = 0;
      signal (SIGINT, p);
#endif
      tputs (exit_graph_mode, 1, tv_here);
      isSetGraphMode = FALSE;
      cur_attr &= ~0200;
#ifdef VMS
      sys$qiow (0, term_file, IO$_WRITEVBLK | IO$M_NOFORMAT, 0, 0, 0,
		&tv_buffer, tv_buff_count, 0, 0, 0, 0);
      tv_buff_count = 0;
#endif
    }
}

void
tv_write_one (char c)
{
  if (tv_buff_count >= MAXTVBUFF)
    tv_flush ();
  tv_buffer[tv_buff_count++] = c;
}

local small
get_lnlen (register small y)
{
  register small i;
  for (i = scx - 1; i >= 0 && scim (i, y) == (tchar) ' '; i--);
  return i + 1;
}

local void
tv_spaces (register small l)
{
  if (cx + l == scx)
    (*tv_clr_eol) (cy, l);
  else if (cx == 0)
    (*tv_clr_bol) (cy, l);
  else
    (*tv_clr_text) (cx, cy, l);
}
