/*
 *      nanoMir
 *
 *      "stuf_unix.c      OS dependent stuffs ( UNIX )
 *
 */

#include "nm.h"
#include "sc.h"
#include "kb.h"
extern char *getenv (char *);
local int xsystem (char *cmd);
local void onCLD (int sig);

void
fileexec (register char *file)
{
  register char *p;
  register int stat;
  local char fnbuf[MAXPATH + 100];
  local char base[MAXPATH];

  if (*file)
    {
      for (p = file; *p++;);
      p--;
      while (*--p != '.')
	if (*p == '/')
	  {
	    *scpy (file, fnbuf) = 0;
	    goto execute;
	  }
      p++;
      *sncpy (file, base, p - file - 1) = 0;
      sprintf (fnbuf, "%s/%s.com %s", nmlocal, p, base);
    }
  else
    {
      char *shell = getenv (MIMSHELL);
      if (shell == NIL)
	shell = "sh";
      /* sprintf( fnbuf, "env PS1=\'NM> \' %s ", shell ); */
      sprintf (fnbuf, "env PS1=\'MIM:%s\' %s ", getenv ("PS1"), shell);
    }
execute:
#ifdef COLOR
  SetDefaultTextColor ();
#endif
#ifndef X11
  scErase ();
  scDetach ();
#endif
  kbDetach ();
#ifdef COLOR
  SetDefaultTextColor ();
#endif
  if (*file)
    printf ("+ %s\n", fnbuf);
  {
    char exbuf[MAXPATH + 128];
    sprintf (exbuf, "exec %s", fnbuf);
#ifdef X11
    xsystem (exbuf);
  }
#else
#if 1
    /*   2  ! */
    stat = system (exbuf);
#else
#ifdef  VERSION_2
    stat = xsystem (exbuf);
#else
    stat = system (exbuf);
#endif /* VERSION_2 */
#endif
  }
  if (*file)
    {
      printf ("NM: stat=%d - press Return\n", stat);
      getchar ();
    }
#endif
  kbAttach ();
#ifndef X11
  scAttach ();
  scErase ();
#endif
#ifdef COLOR
  SetTextColor ();
#endif
}

#include <signal.h>
#ifdef MIM_SVR4
#include <sys/wait.h>
#endif

local bool shell_active = FALSE;
#ifdef MIM_SVR4
local pid_t shell_pid;
#else
local int shell_pid;
#endif

#ifdef X11
void
stufTerm ()
{
  if (shell_active)
    {
      kill (shell_pid, SIGQUIT);
    }
}

local void
onCLD (int sig)
{
  int stat;
#ifdef MIM_SVR4
  waitpid (-1, &stat, WNOHANG);
  if (WIFSIGNALED (stat))
    {
      stat = WTERMSIG (stat);
      printf ("\nNM: Shell terminated due to the receipt of the signal %d \n",
	      stat);
      goto done;
    }
  else if (WIFEXITED (stat))
    {
      stat = WEXITSTATUS (stat);
      printf ("\nNM: Return status is %d \n", stat);
    done:shell_active = FALSE;
      signal (SIGCLD, SIG_DFL);
    }
#else /* not def MIM_SVR4 */
  wait (&stat);
  printf ("\nNM: Return status is %d \n", stat);
#endif /* def MIM_SVR4 */
}
#endif /* def X11 */

local int
xsystem (char *cmd)
{
  if (shell_active)
    return;
  shell_pid = fork ();
  switch (shell_pid)
    {
    case 0:
      signal (SIGINT, SIG_DFL);
      signal (SIGTERM, SIG_DFL);
      signal (SIGQUIT, SIG_DFL);
      execlp ("sh", "sh", "-c", cmd, 0);
      exit (1);
    case -1:
      return -1;
    default:
#ifdef MIM_SVR4
      tcsetpgrp (0, shell_pid);
#endif
      shell_active = TRUE;
#ifdef X11
      signal (SIGCLD, onCLD);
      return 0;
#else
      {
	int stat;
	wait (&stat);
	shell_active = FALSE;
	return stat;
      }
#endif
    }
}
