/*
 *     nanoMir
 *
 *     "te.c"           Text editor
 */

#include "nm.h"
#include "tx.h"
#include "wd.h"
#include "cm.h"
#include "cmn.h"
#include "le.h"
#include "te.h"
#include "fs.h"
#include "tm.h"
#include "cp.h"
#include "kb.h"
#include "ud.h"
#include "er.h"
#ifndef X11
#include "sc.h"
#else
#include "xi.h"
#endif

wnd *Twnd;
txt *Ttxt;
large Ty;
small Tx;
bool Twwrap;
bool Tradjust;
bool tecss;			/* Case sensitive search */
small fplm, fpl1m, fprm;
small (*telsearch) (small), (*tersearch) (small);
bool teHyphen;
int teSpaced;

tchar *lfbuf;
txt *Atxt;

	/* Return     */
bool mechMathReturn = FALSE;

static tchar *spatbuf;
static small spatlen;
static tchar *rpatbuf;
static small rpatlen;
static tchar *spa, *rpa;
small spl;
static small rpl;
static bool tehyp;

#ifdef RUSSIAN
#define sprompt " :"
#define lsprompt 11
#define rprompt " :"
#define lrprompt 13
#endif
#ifdef ENGLISH
#define sprompt "Search pattern:"
#define lsprompt 15
#define rprompt "Replace pattern:"
#define lrprompt 16
#endif

local small lcssearch (small), rcssearch (small), lcusearch (small),
rcusearch (small);
local bool cspateq (tchar *), cupateq (tchar *);
local void tereplace (void);
local small mcdpat (small);

static small fcl1m, fclm, fcrm, folm;
static tchar *fword;
static small flword, fiptr, fictr, folen;
static bool flongword;

local small leftspac (void), ihyphen (tchar *, small, small);
local bool fgetword (void);
local void gformat (bool), foflush (bool);
local small radjust (tchar *, small, small, small);
local bool tewwrap (void);

local void glinedraw (small);
local char textldc (small, large);

void tesro (void), tesrw (void), teswight (void), teptab (void),
tepform (void), tecentr ();
void tesmark (void), tecmark (void), teup (void), tedown (void),
tecup (void), tecdown (void),
tetbeg (void), tetend (void), tecr (void), tercr (void),
tesentr (void), tesup (void), tesdown (void), terdown (void),
terup (void),
tenpage (void), teppage (void), telprint (void), tefprint (void);
#ifdef X11
void tesetxy (void);
#endif
void teil (void), tedl (void), teblin (void), teslin (void),
tenblin (void), tenslin (void),
teipage (void), teformat (void), tepformat (void);
void teldup (void), telddown (void), teldright (void), teldleft (void);

extern void scbell (void);

#define FSTECOM 0451
static comdesc tecmds[] = {
  teundo, /*451 */ CA_CHANGE,
  teunundo, /*452 */ CA_CHANGE,
  tesundo, /*453 */ CA_CHANGE,
  tesunundo, /*454 */ CA_CHANGE,
  tesro, /*455 */ CA_RPT,
  tesrw, /*456 */ CA_RPT,
  teswight, /*457 */ CA_RPT,
  tesmark, /*460 */ CA_RPT,
  tecmark, /*461 */ 0,
  teptab, /*462 */ CA_RPT,
  tepform, /*463 */ CA_RPT,
  teup, /*464 */ CA_NBEG,
  tedown, /*465 */ CA_NEND,
  tecup, /*466 */ CA_RPT,
  tecdown, /*467 */ CA_RPT,
  tetbeg, /*470 */ CA_RPT,
  tetend, /*471 */ CA_RPT,
  teil, /*472 */ CA_CHANGE,
  tedl, /*473 */ CA_CHANGE | CA_NEND,
#ifdef  VERSION_2
  tecr, /*474 */ CA_CHANGE,
#else
  tecr, /*474 */ CA_NEND,
#endif				/* VERSION_2 */
  tercr, /*475 */ CA_NBEG,
  tesentr, /*476 */ 0,
  tesup, /*477 */ 0,
  tesdown, /*500 */ CA_NEND,
  tecentr, /*501 */ CA_RPT,
  teclin, /*502 */ CA_NEND | CA_LCUT,
  tecdlin, /*503 */ CA_NEND | CA_CHANGE | CA_LCUT,
  teplin, /*504 */ CA_CHANGE | CA_LCUT,
  tecclin, /*505 */ CA_LCUT,
  teblin, /*506 */ CA_CHANGE | CA_NEND | CA_RPT,
  teslin, /*507 */ CA_CHANGE | CA_NEND | CA_RPT,
  tenblin, /*510 */ CA_CHANGE | CA_NEND | CA_RPT,
  tenslin, /*511 */ CA_CHANGE | CA_NEND,
  teipage, /*512 */ CA_CHANGE,
  tenpage, /*513 */ CA_NEND,
  teppage, /*514 */ CA_NBEG,
  temcblock, /*515 */ CA_NEND | CA_RPT | CA_LCUT,
  tecdblock, /*516 */ CA_CHANGE | CA_NEND | CA_RPT | CA_LCUT,
  teccblock, /*517 */ CA_CHANGE | CA_NEND | CA_RPT | CA_LCUT,
  tepblock, /*520 */ CA_CHANGE | CA_NEND | CA_LCUT,
  teprblock, /*521 */ CA_CHANGE | CA_NEND | CA_LCUT,
  tepiblock, /*522 */ CA_CHANGE | CA_LCUT,
  tepirblock, /*523 */ CA_CHANGE | CA_LCUT,
  telprint, /*524 */ 0,
  tefprint, /*525 */ 0,
  teformat, /*526 */ CA_NEND | CA_CHANGE,
  tepformat, /*527 */ CA_NEND | CA_CHANGE,
  teldup, /*530 */ CA_CHANGE | CA_EXT | CA_NBEG,
  telddown, /*531 */ CA_CHANGE | CA_EXT,
  teldleft, /*532 */ CA_CHANGE | CA_EXT,
  teldright, /*533 */ CA_CHANGE | CA_EXT,
  terup, /*534 */ CA_CHANGE | CA_NBEG,
  terdown, /*535 */ CA_CHANGE | CA_NEND,
#ifdef X11
  tesetxy, /*536 */ CA_RPT,
#else
  NIL, /*536 */ 0,
#endif
  NIL, /*537 */ 0,
  teblkmod, /*540 */ CA_CHANGE | CA_NEND | CA_RPT | CA_LCUT,
#ifdef SPELL
  spNext, /*541 */ CA_NEND,
  spPrev, /*542 */ 0,
  spAdd, /*543 */ 0,
  spRem, /*544 */ 0,
#endif
};

#ifdef SPELL
#define LSTECOM 0544
#else
#define LSTECOM 0540
#endif

void teCorrect ();		/* Correct the command description table */

void
teInit ()
{
  register char *p;
  register tchar *pt;

#ifdef  VERSION_2
  teCorrect ();
#endif /* VERSION_2 */

  Twnd = NIL;
  Atxt = NIL;
  lfbuf = (tchar *) GetMain (MAXLPAC * sizeof (tchar));

  spatbuf = pt = (tchar *) GetMain (MAXLPAC * sizeof (tchar));
  for (p = sprompt; (*pt++) = ctotc (*p++););
  spatlen = lsprompt;
  spl = 0;
  spa = NIL;
  rpatbuf = pt = (tchar *) GetMain (MAXLPAC * sizeof (tchar));
  for (p = rprompt; (*pt++) = ctotc (*p++););
  rpatlen = lrprompt;
  rpl = 0;
  rpa = NIL;
  if (tecss)
    {
      telsearch = lcssearch;
      tersearch = rcssearch;
    }
  else
    {
      telsearch = lcusearch;
      tersearch = rcusearch;
    }
  fword = (tchar *) GetMain (MAXFWORD * sizeof (tchar));
}

/* Added by VVB */
void
teCorrect ()
{
  /* tecr,   *474*  CA_NEND,   */
  /* tecr,   *474*  CA_CHANGE, */
  /* Correct "Return" entry in command descriptor table */
  if (mechMathReturn)
    {
      tecmds[0474 - FSTECOM].cattr = CA_NEND;
    }
  else
    {
      tecmds[0474 - FSTECOM].cattr = CA_CHANGE;
    }
}

comdesc *
Tdecode ()
{
  register kchar code;
  if ((code = KbCode - FSTECOM) >= 0 && code <= (LSTECOM - FSTECOM))
    return (tecmds + code);
  return (NIL);
}

static void
tekbwait ()
{
  register wnd *w;
  register txt *t;
  wadjust (w = Twnd, Tx, Ty);
  Lx = Tx;
  if (BlockMark)
    cpblockmark ();
  wpos (w, Tx, Ty);
/*stCXYSym( Lx+1, Ty+1 ); */
  stCXYSym (Tx, Ty);
}

#define erret(  code ) { er=code; goto ret;  }
#define erret2( code ) { er=code; goto ret2; }

small
TExec (small typ)
{
  register comdesc *cp;
  register txt *t;
  register small attr;
  register small er;
  large r;			/* Can't be register due to setjmp */
  register void (*kboldhook) ();
  jmp_buf teenv;
  register jmp_buf *nextexc;
#ifdef NO_MARGIN_BELL
  small IsMargin = 0;
#endif

  t = Ttxt;
  nextexc = excptr;
  excptr = (jmp_buf *) teenv;
  kboldhook = KbWaitIn;
  if (er = setjmp (teenv))
    goto ret2;
  KbWaitIn = tekbwait;

  TxSetY (t, Ty);
  if (typ & RQ_EXEC)
    goto noinp;
loop:
  KbGet ();
noinp:
  if ((cp = Tdecode ()) != NIL)
    {
      attr = cp->cattr;
      r = (attr & CA_RPT) ? 1 : KbCount;
      if (attr & CA_LCUT && (t == LCtxt))
	erret (E_LCUT);
      if (attr & CA_CHANGE)
	{
	  if (t->txredit != TXED_YES)
	    erret (E_CHANGE);
	  UdMark = TRUE;
	}
      for (; r; r--)
	{
	  TxSetY (t, Ty);
#ifdef NO_MARGIN_BELL
	  if (attr & CA_NEND && qTxDown (t))
	    {
	      IsMargin = 1;
	      erret2 (E_MOVDOWN);
	    }
	  if (attr & CA_NBEG && qTxUp (t))
	    {
	      IsMargin = 1;
	      erret2 (E_MOVUP);
	    }
#else
	  if (attr & CA_NEND && qTxDown (t))
	    erret2 (E_MOVDOWN);
	  if (attr & CA_NBEG && qTxUp (t))
	    erret2 (E_MOVUP);
#endif
/*    if( qKbReq() == TK_BREAK ) erret2( E_KBREAK ); */
	  if ((attr & CA_EXT) && qTxDown (t))
	    {
	      teil ();
	      t->txstat |= TS_CHANGE;
	      TxSetY (t, Ty);
	    }
	  (*cp->cfunc) ();
	  if (attr & CA_CHANGE)
	    {
	      t->txstat |= TS_CHANGE;
	      stModification (STM_YES);
	    }
	}
    }
  else
  down:
  if ((cp = Ldecode ()) != NIL)
    {
      TxSetY (t, Ty);
      tleload ();
      undodir = 0;
      er = LExec (typ | RQ_EXEC);
      if (er == E_UP)
	undodir = (-1);
      tleunload ();
      undodir = 0;
      switch (er)
	{
	case E_UP:
	case E_NOCOM:
	  goto noinp;
	case E_CHANGE:
	  if ((Ldecode ()->cattr & CA_EXT) &&
	      qTxDown (t) && t->txredit == TXED_YES)
	    {
	      teil ();
	      t->txstat |= TS_CHANGE;
	      goto down;
	    }
	  goto ret;
	case E_MOVEND:
	case E_EDTEND:
	  if (((attr = KbCode) < 0400 || attr == LE_PCHAR) &&
	      Tx == t->txrm && t->txredit == TXED_YES && Twwrap && tewwrap ())
	    {
	      if (Tx == 0 && (char) KbCode == ' ')
		goto loop;
	      goto down;
	    }
	  goto ret;
	default:
	  goto ret;
	}
    }
  else
    erret (E_NOCOM);
  if (typ & RQ_WAIT)
    goto loop;
  erret (E_OK);
ret2:
  if (!(attr & CA_RPT))
    KbCount = r;
ret:
  if (er == E_DOWN)
    {
      undodir = 1;
      goto down;
    }
#ifdef NO_MARGIN_BELL
  if (typ & RQ_WAIT && !Event[er])
    {
      if (!IsMargin)
	scbell ();
      else
	IsMargin = 0;
      goto loop;
    }
#else
  if (typ & RQ_WAIT && !Event[er])
    {
      scbell ();
      goto loop;
    }
#endif
  KbWaitIn = kboldhook;
  excptr = nextexc;
  return (er);
}

void
tleload ()
{
  register small l;
  register txt *t;
  assert (Lwnd == NIL, "leload");
  Lwnd = Twnd;
  Ltxt = t = Ttxt;
  Lleng = TxFRead (t, Lebuf);
  Lchange = FALSE;
  Lredit = (qTxDown (t) ? FALSE : t->txredit == TXED_YES);
  Ly = Ty;
  Lx = Tx;
  Lxlm = Lxle = 0;
  Lxre = t->txrm;
  Lxrm = t->txrm - 1;
  tundoload (t);
#ifdef TEDEBUG
  dbprintf ("LOAD\n");
#endif
}

void
tleunload ()
{
  register txt *t;
  assert (Lwnd != NIL, "leunload");
  Lwnd = NIL;
  Tx = Lx;
  t = Ttxt;
  tundounload (t);
#ifdef TEDEBUG
  dbprintf ("UNLOAD\n");
#endif
  if (Lchange)
    {
/*    UdMark = TRUE;   */
      TxSetY (t, Ty);
      TxTRep (t, Lebuf, Lleng);
      t->txstat |= TS_CHANGE;
    }
}

/*
 *  Enter line
 *    Parameters:
 *       lptr   (in, out): the line
 *       plen   (in, out): lenght of the line
 *       lprompt(in)     : lenght of prompt
 *       terms  (in)     : array of commands that finalize input
 *       style  (in)     : style of input:
 */
void
Lenter (tchar * lptr, small * plen, small lprompt, small * terms,
	bool tcstyle)
{
  register wnd *w;
  register small *p;
  register small c;
  small x;
  register void (*kboldhook) () = KbWaitIn;

#ifdef COLORE
  SetEnterTextColor ();
#endif
  KbWaitIn = tekbwait;
  if (Atxt == NIL)
    {
      while ((Atxt = TxNew (TRUE)) == NIL)
	tmswap (TRUE);
      Atxt->txstat |= TS_PERM1;
    }
  TxEmpt (Atxt);

  Lleng = *plen;
  Lwnd = w = Twnd;
  x = w->wtx;
  if (lptr != NIL)
    blkmov ((char *) lptr, (char *) (Lebuf + x), Lleng * sizeof (tchar));
  else
    blkmov ((char *) Lebuf, (char *) (Lebuf + x), Lleng * sizeof (tchar));
  Lleng += x;
  blktspac (Lebuf + Lleng, MAXLPAC - Lleng);
  Ltxt = Atxt;
  Lchange = FALSE;
  Lredit = TRUE;
  Ly = Ty;
  Lx = Lxlm = Lxle = lprompt + x;
  Lxrm = Lxre = MAXLPAC;
  wxmin = 0;
  wxmax = w->wsw;
  wymin = Ly - w->wty;
  wymax = wymin + 1;
  wrdrw (w);
  c = RQ_WAIT;
  if (tcstyle)
    {
#ifdef X11
      wpos (Lwnd, Lx, Ly);
#endif
      c = Tx;
      Tx = Lx = Lleng;
      KbGet ();
      Tx = c;
      /*      ,    */
      if (KbCode < 0400 || KbCode == LE_PCHAR)
	{
	  Lx = lprompt + x;
	  ledeol ();
	}
      c = RQ_WAIT | RQ_EXEC;
    }
  for (;;)
    {
      if (LExec (c) == E_NOCOM)
	for (p = terms; *p;)
	  if (*p++ == KbCode)
	    goto ok;
      scbell ();
      c = RQ_WAIT;
    }
ok:
  Lwnd = NIL;			/* tundounload( Ttxt ); */
  wxmin = 0;
  wxmax = w->wsw;
  wymin = Ly - w->wty;
  wymax = wymin + 1;
  wrdrw (w);
  Lleng -= x;
  blkmov ((char *) (Lebuf + x), (char *) Lebuf, Lleng * sizeof (tchar));
  if (lptr != NIL)
    blkmov ((char *) Lebuf, (char *) lptr, Lleng * sizeof (tchar));
  *plen = Lleng;
  KbWaitIn = kboldhook;
#ifdef COLORE
  SetTextColor ();
#endif
}

void
tesro ()
{
  if (Ttxt->txredit == TXED_YES)
    Ttxt->txredit = TXED_NO;
}
void
tesrw ()
{
  if (Ttxt->txredit == TXED_NO)
    Ttxt->txredit = TXED_YES;
}

void
teswight ()
{
  register small x;
  if (KbRadix == 0 || (x = KbCount) < 0 || x > MAXLPAC)
    exc (E_WIGHT);
  Ttxt->txrm = x;
  if (Tx > x)
    Tx = x;
}

static small tpterms[] = { TE_CR, TK_BREAK, 0 };

void
teptab ()
{
  register small i;
  register tchar *pt;
  register char *p;
  small tplen;
  for (i = tplen = Ttxt->txrm, p = tabuf, pt = Lebuf; i; i--)
    *pt++ = ctotc ((*p++) & TABSTOP ? '*' : '-');
  Lenter (NIL, &tplen, 0, tpterms, FALSE);
  if (KbCode != TK_BREAK)
    for (i = tplen, p = tabuf, pt = Lebuf; i; i--, p++)
      {
	*p &= ~TABSTOP;
	if ((char) (*pt++) == '*')
	  *p |= TABSTOP;
      }
}


void
qsety (register large y)
{
  register bool q;
  register txt *t;
  t = Ttxt;
  q = TxSetY (t, y);
  Ty = t->txy;
  if (!q)
    exc (E_SETY);
}

void
tecentr ()
{
  register txt *t;
  register wnd *w;
  register large y;
  register small whh;
  t = Ttxt;
  w = Twnd;
  if (KbRadix != 0)
    {
      if (KbCount == 0)
	qsety (0);
      else
	qsety (KbCount - 1);
    }
  else
    {
      wadjust (w, Tx, Ty);
      whh = (w->wsh - 1) >> 1;
      if ((y = Ty - whh) < 0)
	y = 0;
      wadjust (w, Tx, y);
      while (!qTxDown (t) && whh--)
	TxDown (t);
      wadjust (w, Tx, t->txy);
    }
}

void
tesmark ()
{
  register txt *t;
  t = Ttxt;
  t->txmarkx = Tx;
  t->txmarky = Ty;
}

void
tecmark ()
{
  register txt *t;
  register small x;
  register large y;
  t = Ttxt;
  x = t->txmarkx;
  y = t->txmarky;
  t->txmarkx = Tx;
  t->txmarky = Ty;
  Tx = x;
  qsety (y);
}

void
teup ()
{
  Ty--;
}
void
tedown ()
{
  Ty++;
}

#ifdef X11
void
tesetxy ()
{
  large l;
  Tx = min (xiTx, Ttxt->txrm);
  Tx = max (0, Tx);
  l = TxLines (Ttxt);
  Ty = min (xiTy, l);
  Ty = max (0, Ty);
}
#endif

static void
cupdown (small dy)
{
  kchar k;
  for (;;)
    {
      qsety (Ty + dy);
      wadjust (Twnd, Tx, Ty);
#ifdef SLOW_VERSION
#if defined( ALPHA ) && !defined( X11 )
      if (!ScLocked)
	usleep (20000);		/* very FAST machine */
#endif
#endif
      k = qKbReq ();
      if (k)
	break;
    }
}

void
tecup ()
{
  cupdown (-1);
}
void
tecdown ()
{
  cupdown (1);
}

void
tetbeg ()
{
  qsety (0L);
}
void
tetend ()
{
  qsety (MAXLARGE);
}

#ifdef  VERSION_2
void
tecr ()
{
  if (mechMathReturn)
    {
      Ty++;
      Tx = 0;			/* Return of mech-math */
    }
  else
    {
      if (!qTxDown (Ttxt))
	{
	  teblin ();
	}
      else
	{
	  teil ();
	}
      Ty++;
      Tx = 0;
    }
}
#else
void
tecr ()
{
  Ty++;
  Tx = 0;
}
#endif /* VERSION_2 */

void
tercr ()
{
  Ty--;
  Tx = 0;
}

#define s_upper(c) if(c>='a'&&c<='z')c-=040;else if((uchar)(c)>=(uchar)''&&(uchar)(c)<=(uchar)0xdf)c+=040;

static small sterms[] = { TK_BREAK, TE_UP, TE_SUP,
  TE_DOWN, TE_SDOWN, TE_CR, TE_SENTR, 0
};
static small rterms[] = { TK_BREAK, TE_UP, TE_SUP,
  TE_DOWN, TE_SDOWN, TE_CR, 0
};

#if 0
void
tesentr ()
{
  register tchar *p, *pe;
  register char c;
  register small e;
  Lenter (spatbuf, &spatlen, lsprompt, sterms, TRUE);
  e = KbCode;
  for (spa = p = spatbuf + lsprompt, pe = spatbuf + spatlen; p < pe; p++)
    {
      c = (char) *p;
      if (c == '/')
	{
	  spl = p - spa;
	  rpa = ++p;
	  rpl = pe - p;
	  goto l1;
	}
      if (!tecss)
	{
	  s_upper (c);
	  *p = ctotc (c);
	}
    }
  spl = p - spa;
  rpa = NIL;
l1:
  if (e == TE_SUP || e == TE_UP)
    tesup ();
  if (e == TE_SDOWN || e == TE_DOWN || e == TE_CR)
    tesdown ();
}
#endif

void
tesentr ()
{
  register tchar *p, *pe;
  register small e, i;
  Lenter (spatbuf, &spatlen, lsprompt, sterms, TRUE);
  e = KbCode;
  spa = spatbuf + lsprompt;
  spl = spatlen - lsprompt;
  if (!tecss)
    for (p = spa, i = spl; i--; p++)
      {
	s_upper (*p);
	*p = ctotc (*p);
      }
  switch (e)
    {
    case TE_SUP:
    case TE_UP:
      tesup ();
      rpa = NIL;
      break;
    case TE_SDOWN:
    case TE_DOWN:
    case TE_CR:
      tesdown ();
      rpa = NIL;
      break;
    case TE_SENTR:
      Lenter (rpatbuf, &rpatlen, lrprompt, rterms, TRUE);
      e = KbCode;
      rpa = rpatbuf + lrprompt;
      rpl = rpatlen - lrprompt;
      switch (e)
	{
	case TE_SUP:
	case TE_UP:
	  tesup ();
	  break;
	case TE_SDOWN:
	case TE_DOWN:
	case TE_CR:
	  tesdown ();
	  break;
	}
    }
}

void
TxSRead ()
{
  register small i, m;
  register tchar *p, *p1, *pe;
  register txt *t;

  Lleng = TxTRead (t = Ttxt, p = Lebuf);
  teSpaced = leftspac ();
  p += Lleng;
  if (teSpaced >= 0 && (char) *(p - 1) == ' ' && !qTxDown (t))
    {
      tehyp = TRUE;
      TxDown (t);
      m = TxTRead (t, p1 = lfbuf);
      TxUp (t);
      while ((char) (*p1++) == ' ' && --m >= 0);
      p1--;
      p--;
    }
  else
    {
      tehyp = FALSE;
      m = 0;
    }
  for (i = spl, pe = Lebuf + MAXLPAC; --i >= 0 && p < pe;)
    *p++ = (--m >= 0) ? (*p1++) : (tchar) ' ';
  Lleng = p - Lebuf;
}


void
te_sdown (small d)
{
  register small x;
  register txt *t;
  register tchar *p;
  t = Ttxt;
  x = Tx + d;
  teHyphen = FALSE;
  if (qTxDown (t))
    exc (E_MOVDOWN);
  for (;;)
    {
      TxSRead ();
      if ((x = (*tersearch) (x)) >= 0)
	{
	  Ty = t->txy;
	  Tx = x;
	  break;
	}
      else
	{
	  teHyphen = tehyp;
	  if (qKbBreak ())
	    exc (E_KBREAK);
	  TxDown (t);
	  if (qTxDown (t))
	    exc (E_SFAIL);
	}
    }
}

void
te_sup (small d)
{
  register small x;
  register txt *t;
  t = Ttxt;
  if ((x = Tx - d) < 0)
    goto prevline;
  for (;;)
    {
      TxSRead ();
      if ((x = (*telsearch) (x)) >= 0)
	{
	  Ty = t->txy;
	  Tx = x;
	  break;
	}
      else
	{
	prevline:
	  if (qKbBreak ())
	    exc (E_KBREAK);
	  if (qTxUp (t))
	    exc (E_SFAIL);
	  TxUp (t);
	}
    }
}

void
tesup ()
{
  te_sup (1);
}
void
tesdown ()
{
  te_sdown (1);
}

void
terdown ()
{
  te_sdown (0);
  tereplace ();
  Tx += rpl;
  te_sdown (0);

}

void
terup ()
{
  te_sup (0);
  tereplace ();
  Tx -= spl;
  te_sup (0);
}

/*
 * Case sensitive line search
 */
static small
rcssearch (register small x)
{
  register tchar *p, *pe;
  register char s;
  if (spl == 0)
    exc (E_NOSPAT);
  if (x < 0)
    x = 0;
  for (p = Lebuf + x, pe = Lebuf + (Lleng - spl), s = (char) (*spa); p <= pe;)
    if (((char) *p++) == s && cspateq (p))
      return ((p - Lebuf) - 1);
  return (-1);
}

static small
lcssearch (register small x)
{
  register tchar *p, *pe;
  register char s;
  if (spl == 0)
    exc (E_NOSPAT);
  if (x < 0 || x > Lleng - spl)
    x = Lleng - spl;
  for (p = Lebuf + x, pe = Lebuf, s = (char) (*spa); p >= pe; p--)
    if (((char) *p) == s && cspateq (p + 1))
      return ((p - pe));
  return (-1);
}

static bool
cspateq (register tchar * p)
{
  register tchar *p1;
  register small i;
  for (p1 = spa + 1, i = spl - 1; --i >= 0;)
    if ((char) (*p++) != (char) (*p1++))
      return (FALSE);
  return (TRUE);
}

/*
 * Case unsensitive line search
 */
static small
rcusearch (register small x)
{
  register tchar *p, *pe;
  register char s, c;
  if (spl == 0)
    exc (E_NOSPAT);
  if (x < 0)
    x = 0;
  for (p = Lebuf + x, pe = Lebuf + (Lleng - spl), s = (char) (*spa); p <= pe;)
    {
      c = (char) (*p++);
      s_upper (c);
      if (c == s && cupateq (p))
	return ((p - Lebuf) - 1);
    }
  return (-1);
}

static small
lcusearch (register small x)
{
  register tchar *p, *pe;
  register char s, c;
  if (spl == 0)
    exc (E_NOSPAT);
  if (x < 0 || x > Lleng - spl)
    x = Lleng - spl;
  for (p = Lebuf + x, pe = Lebuf, s = (char) (*spa); p >= pe; p--)
    {
      c = (char) (*p);
      s_upper (c);
      if (c == s && cupateq (p + 1))
	return ((p - pe));
    }
  return (-1);
}

static bool
cupateq (register tchar * p)
{
  register tchar *p1;
  register small i;
  register char c;
  for (p1 = spa + 1, i = spl - 1; i; i--)
    {
      c = (char) (*p++);
      s_upper (c);
      if (c != (char) (*p1++))
	return (FALSE);
    }
  return (TRUE);
}

/*
 * Replace substring
 */
static void
tereplace ()
{
  register small l;
  if (rpa == NIL)
    exc (E_NORPAT);
  if (Tx + max (spl, rpl) > MAXLPAC)
    exc (E_ROUT);
  TxFRead (Ttxt, Lebuf);
  if (rpl != spl)
    {
      l = max (spl, rpl);
      blkmov ((char *) (Lebuf + (Tx + spl)), (char *) (Lebuf + (Tx + rpl)),
	      (MAXLPAC - Tx - l) * sizeof (tchar));
      if ((l = (spl - rpl)) > 0)
	blktspac (Lebuf + (MAXLPAC - l), l);
    }
  blkmov ((char *) rpa, (char *) (Lebuf + Tx), rpl * sizeof (tchar));
  TxFRep (Ttxt, Lebuf);
  Ttxt->txstat |= TS_CHANGE;
  stModification (STM_YES);
}

void
tenpage ()
{
  qsety (Ty + Twnd->wsh);
}
void
teppage ()
{
  qsety (Ty - Twnd->wsh);
}

static tchar ffpa[1] = { (tchar) 014 };
static small ffpl = 1;

void
teipage ()
{
  TxTIL (Ttxt, ffpa, ffpl);
}

void
tepform ()
{
  register small i, oldrm;
  register tchar *pt;

  oldrm = fprm;
  for (i = 0, pt = Lebuf; i < oldrm; i++)
    *pt++ = ctotc (i < fplm ? ' ' : (i < fpl1m ? '+' : '-'));
  Lenter (NIL, &fprm, 0, tpterms, FALSE);
  if (KbCode != TK_BREAK)
    {
      for (i = 0, pt = Lebuf; i < fprm && (char) (*pt) == ' '; i++)
	++pt;
      fplm = i;
      for (; i < fprm && (char) (*pt++) == '+'; i++);
      fpl1m = i;
    }
  else
    fprm = oldrm;
}

void
teformat ()
{
  gformat (TRUE);
}
void
tepformat ()
{
  gformat (FALSE);
}

static void
gformat (bool fai)
{
  register txt *t;
  register bool qproc;
  register small lw, i, j;

  TxDown (t = Ttxt);
  if (!qTxDown (t))
    {
      Lleng = TxTRead (t, Lebuf);
      if ((i = leftspac ()) < 0)
	i = 0;
    }
  else
    i = 0;
  folm = fclm = i;
  TxUp (t);
  Lleng = TxTRead (t, Lebuf);
  if ((i = leftspac ()) < 0)
    exc (E_FORMAT);
  fcl1m = i;
  fcrm = t->txrm;

  if (!fai)
    {
      fclm = fplm;
      fcl1m = fpl1m;
      fcrm = min (fprm, fcrm);
    }
  if (fcrm <= fclm)
    exc (E_FORMAT);

  folen = 0;
  foflush (FALSE);
  flongword = FALSE;
  flword = 0;
  fiptr = Lleng = 0;
  fictr = 0;
  do
    {
      qproc = fgetword ();
      if ((lw = flword) == 0)
	break;
      for (i = 1;;)
	{
	  if ((j = folen) == fcrm)
	    goto noh;
	  if (j != fcl1m)
	    lfbuf[j++] = (tchar) ' ';
	  if (j + lw <= fcrm)
	    break;
	  if (i = ihyphen (fword, lw, fcrm - j - 1))
	    {
	      blkmov ((char *) fword, (char *) (lfbuf + j),
		      i * sizeof (tchar));
	      j += i;
	      lfbuf[j++] = HYPHEN;
	      folen = j;
	      blkmov ((char *) (fword + i), (char *) fword,
		      (flword = (lw -= i)) * sizeof (tchar));
	    }
	noh:
	  foflush (TRUE);
	  if (lw > fcrm - fcl1m && i == 0)
	    {
	      flongword = TRUE;
	      goto fstop;
	    }
	}
      blkmov ((char *) fword, (char *) (lfbuf + j), lw * sizeof (tchar));
      folen = j + lw;
      flword = 0;
    }
  while (qproc);
fstop:
  foflush (FALSE);
  if (flongword)
    {
      TxSetY (t, Ty);
      TxTIL (t, Lebuf, Lleng);
    }
}

static bool
fgetword ()
{
  register txt *t;
  register small l, i;
  register tchar c;
  t = Ttxt;

  if (flword)
    return (TRUE);
  c = 0;
  if ((i = fiptr) == Lleng)
    {
    nextl:
      TxSetY (t, Ty);
      if (qTxDown (t))
	goto fstop;
      Lleng = TxTRead (t, Lebuf);
      if ((i = leftspac ()) < 0)
	goto fstop;
    l1:if (fictr > 0 && i != folm)
	goto fstop;
      TxDL (t);
      fictr++;
    }
  for (;;)
    {
      if (i == Lleng)
	{
	  if (((small) c & 0377) == HYPHEN)
	    {
	      flword--;
	      goto nextl;
	    }
	  break;
	}
      if ((char) (c = Lebuf[i++]) == ' ')
	break;
      if (flword >= MAXFWORD)
	{
	  flongword = TRUE;
	  goto fstop;
	}
      fword[flword++] = c;
    }
  while (i < Lleng && (char) Lebuf[i] == ' ')
    i++;
  fiptr = i;
  return (TRUE);
fstop:return (FALSE);
}

static small
leftspac ()
{
  register small l;
  register tchar *pt;
  for (pt = Lebuf + (l = Lleng);; l--)
    {
      if (l == 0)
	goto err;
      if ((char) (*--pt) != ' ')
	break;
    }
  Lleng = l;
  for (pt = Lebuf; (char) (*pt++) == ' ';);
  return ((pt - Lebuf) - 1);
err:return (-1);
}


static void
foflush (bool qradjust)
{
  register tchar *pt;
  register small i;

  if (folen > fcl1m)
    {
      if (qradjust)
	folen = radjust (lfbuf, fcl1m, folen, fcrm);
      wadjust (Twnd, Tx, Ty);
      TxSetY (Ttxt, Ty);
      TxTIL (Ttxt, lfbuf, folen);
      Ty++;
      fcl1m = fclm;
    }
  for (pt = lfbuf, i = fcl1m; i; i--)
    *pt++ = (tchar) ' ';
  folen = fcl1m;
}

static bool
tewwrap ()
{
  register small i, l, j;
  register tchar *p;
  register txt *t;
  t = Ttxt;
  j = leftspac ();
  l = 0;
  i = Lleng;
  if (j >= 0)
    {
      p = Lebuf + Tx;
      if ((char) KbCode != ' ' && (char) (*--p) != ' ')
	{
	  while (p > Lebuf)
	    if ((char) (*--p) == ' ')
	      goto ok;
	  return (FALSE);
	ok:
	  p++;
	  l = Tx - (p - Lebuf);
	  if ((i = ihyphen (p, l, l - 1)) != 0)
	    {
	      p += i;
	      l -= i;
	    }
	  blkmov ((char *) p, (char *) lfbuf, l * sizeof (tchar));
	  if (i)
	    *p++ = HYPHEN;
	  while (p > Lebuf)
	    if ((char) (*--p) != ' ')
	      {
		p++;
		break;
	      }
	  i = (p - Lebuf);
	}
      i = radjust (Lebuf, j, i, Lx);
      TxTRep (t, Lebuf, i);
    }
  l = mcdpat (Tx = l);
  TxDown (t);
  TxTIL (t, lfbuf, l);
  Ty++;
  return (TRUE);
}

static small
radjust (register tchar * pt, small l, small m, small r)
{
  register small i, es, ns;

  if (Tradjust && (es = r - m) > 0 && m > l)
    {
      for (pt += l, i = m - l, ns = 0; i; i--)
	if ((char) (*pt++) == ' ')
	  ns++;
      if (ns)
	{
	  m = r;
	  do
	    {
	      for (i = 1; (char) (*--pt) != ' '; i++);
	      blkmov ((char *) pt, (char *) (pt + es), i * sizeof (tchar));
	      i = (es + ns - 1) / ns;
	      es -= i;
	      blktspac (pt + es, i);
	    }
	  while (--ns);
	}
    }
  return (m);
}

static struct letlist
{
  int typ;
  char *list;
}
letters[] =
{
  'V', "",
    'C', "",
    'S', "\357", '#', NIL};

#define RULES "#***N*#**N**#*N***#N*S**YCVCVYVCCVY*VV*Y*CC*E"

static small
tletter (tchar c)
{
  register struct letlist *p;
  for (p = letters; p->list != NIL && sindex (p->list, (char) c) == NIL; ++p);
  return (p->typ);
}

static bool
hasvowel (register tchar * p1, register tchar * p2)
{
  register small t;
  while (p1 != p2)
    {
      if (p1 > p2)
	p1--;
      if ((t = tletter (*p1)) == 'V')
	return (TRUE);
      if (t == '#')
	break;
      if (p1 < p2)
	p1++;
    }
  return (FALSE);
}

static small
ihyphen (register tchar * pw, register small lw, register small im)
{
  register char *r;
  register small i, j;
  register tchar *pw0;
  small c[4];
  pw0 = pw;

  for (pw += lw, i = lw; i > 0;)
    {
      --i;
      c[3] = c[2];
      c[2] = c[1];
      c[1] = c[0];
      c[0] = tletter (*--pw);
      if (i <= lw - 4 && i <= im - 2)
	{
	  for (r = RULES; *r; r += 5)
	    {
	      for (j = 0; j < 4; ++j)
		if (r[j] != '*' && r[j] != c[j])
		  goto nextrule;
	      switch (r[4])
		{
		case 'E':
		  if ((char) *(pw + 1) != (char) *(pw + 2))
		    goto nextplac;
		case 'Y':
		  if (hasvowel (pw + 2, pw0) && hasvowel (pw + 2, pw0 + lw))
		    return (i + 2);
		case 'N':
		  goto nextplac;
		}
	    nextrule:;
	    }
	}
    nextplac:;
    }
  return (0);
}

void
telprint ()
{
#if 0
  if (qTxDown (Ttxt))
    exc (E_MOVDOWN);
  PrLine (Lebuf, TxTRead (Ttxt, Lebuf));
  Ty++;
  ErOff ();
#endif
}

void
tefprint ()
{
#if 0
  for (;;)
    {
      if (qTxDown (Ttxt))
	exc (E_MOVDOWN);
      PrLine (Lebuf, TxTRead (Ttxt, Lebuf));
      Ty++;
      wadjust (Twnd, Tx, Ty);
      TxSetY (Ttxt, Ty);
      if (qKbBreak ())
	break;
    }
  ErOff ();
#endif
}

static small
mcdpat (register small i)
{
  register tchar *p;
  if (qmcd (Ttxt) && i < MCD_LEFT)
    {
      for (p = lfbuf + i; i < MCD_LEFT; i++)
	*p++ = (tchar) ' ';
      *p++ = MICROS_DIR_FIELDS_SEPARATOR;
      *p++ = (tchar) ' ';
      *p++ = (tchar) ' ';
      *p++ = (tchar) ' ';
      *p++ = MICROS_DIR_FIELDS_SEPARATOR;
      i = MCD_RIGHT + 1;
    }
  return (i);
}

void
teil ()
{
  TxTIL (Ttxt, lfbuf, mcdpat (0));
}

void
tedl ()
{
  TxDL (Ttxt);
}

static void
teslice (bool vert)
{
  register small l, l1;
  register txt *t;
  register tchar *p;

  t = Ttxt;
  l = TxTRead (t, p = Lebuf);
  l1 = min (l, Tx);
  TxTRep (t, p, l1);
  TxDown (t);
  if (vert)
    {
      if (l > l1)
	blktspac (p, l1);
      else
	l = 0;
    }
  else
    {
      l -= l1;
      p += l1;
    }
  TxTIL (t, p, l);
}

void
teblin ()
{
  teslice (FALSE);
}
void
teslin ()
{
  teslice (TRUE);
}

static void
tepaste (bool vert)
{
  register txt *t;
  register tchar *p;
  register small l, l1;

  t = Ttxt;
  l = TxFRead (t, Lebuf);
  if (vert)
    l = Tx;
  TxDown (t);
  if (qTxDown (t))
    exc (E_LPASTE);
  l1 = TxTRead (t, p = lfbuf);
  if (l + l1 > MAXLPAC)
    exc (E_LPASTE);
  if (vert)
    {
      p += l;
      if ((l1 -= l) < 0)
	l1 = 0;
    }
  blkmov ((char *) p, (char *) (Lebuf + l), l1 * sizeof (tchar));
  TxDL (t);
  TxSetY (t, Ty);
  TxTRep (t, Lebuf, l + l1);
}

void
tenblin ()
{
  tepaste (FALSE);
}
void
tenslin ()
{
  tepaste (TRUE);
}

small
TxFRead (txt * t, tchar * p)
{
  register small l;
  l = TxTRead (t, p);
  blktspac (p + l, MAXLPAC - l);
  return (l);
}

void
TxFRep (txt * t, tchar * p)
{
  TxTRep (t, p, lstrlen (MAXLPAC, p));
}

#ifdef KOI8
unsigned char asctoldc[] = {
/* 0x83 */ LD_UP | LD_DOWN,
/* 0x84 */ LD_UP | LD_DOWN | LD_LEFT,
/* 0x85 */ ' ',
/* 0x86 */ ' ',
/* 0x87 */ ' ',
/* 0x88 */ ' ',
/* 0x89 */ ' ',
/* 0x8a */ ' ',
/* 0x8b */ ' ',
/* 0x8c */ ' ',
/* 0x8d */ ' ',
/* 0x8e */ ' ',
/* 0x8f */ LD_DOWN | LD_LEFT,
/* 0x90 */ LD_UP | LD_RIGHT,
/* 0x91 */ LD_UP | LD_LEFT | LD_RIGHT,
/* 0x92 */ LD_DOWN | LD_LEFT | LD_RIGHT,
/* 0x93 */ LD_UP | LD_DOWN | LD_RIGHT,
/* 0x94 */ LD_RIGHT | LD_LEFT,
/* 0x95 */ LD_UP | LD_DOWN | LD_LEFT | LD_RIGHT,
/* 0x96 */ ' ',
/* 0x97 */ ' ',
/* 0x98 */ ' ',
/* 0x99 */ ' ',
/* 0x9a */ ' ',
/* 0x9b */ ' ',
/* 0x9c */ ' ',
/* 0x9d */ ' ',
/* 0x9e */ ' ',
/* 0x9f */ ' ',
/* 0xa0 */ ' ',
/* 0xa1 */ ' ',
/* 0xa2 */ ' ',
/* 0xa3 */ ' ',
/* 0xa4 */ ' ',
/* 0xa5 */ ' ',
/* 0xa6 */ ' ',
/* 0xa7 */ ' ',
/* 0xa8 */ ' ',
/* 0xa9 */ LD_UP | LD_LEFT,
/* 0xaa */ LD_DOWN | LD_RIGHT
};
#else
unsigned char asctoldc[] = {
  LD_DOWN | LD_RIGHT, LD_DOWN | LD_LEFT, LD_UP | LD_RIGHT, LD_UP | LD_LEFT,
  LD_UP | LD_DOWN | LD_RIGHT, LD_UP | LD_DOWN | LD_LEFT,
  LD_LEFT | LD_RIGHT | LD_DOWN, LD_LEFT | LD_RIGHT | LD_UP,
  LD_LEFT | LD_RIGHT, LD_UP | LD_DOWN, LD_UP | LD_DOWN | LD_LEFT | LD_RIGHT
};
#endif /* KOI8 */

#ifdef KOI8
unsigned char ldctoasc[] = {
  ' ',				/*   0000 */
  ' ',				/*   0001 */
  ' ',				/*   0010 */
  0x83,				/*   0011 LD_UP|LD_DOWN */
  ' ',				/*   0100 */
  0xa9,				/*   0101 LD_UP|LD_LEFT */
  0x8f,				/*   0110 LD_DOWN|LDLEFT  */
  0x84,				/*   0111 LD_UP|LD_DOWN|LD_LEFT */
  ' ',				/*   1000 */
  0x90,				/*   1001 LD_UP|LD_RIGHT */
  0xaa,				/*   1010 LD_DOWN|LD_RIGHT */
  0x93,				/*   1011 LD_UP|LD_DOWN|LD_RIGHT */
  0x94,				/*   1100 LD_RIGHT|LD_LEFT     */
  0x91,				/*   1101 LD_UP|LD_LEFT|LD_RIGHT */
  0x92,				/*   1110 LD_DOWN|LD_LEFT|LD_RIGHT */
  0x95				/*   1111 ALL */
};
#else
unsigned char ldctoasc[] = {
  ' ',				/*   0000 */
  ' ',				/*   0001 */
  ' ',				/*   0010 */
  0211,				/*   0011 LD_UP|LD_DOWN */
  ' ',				/*   0100 */
  0203,				/*   0101 LD_UP|LD_LEFT */
  0201,				/*   0110 LD_DOWN|LDLEFT  */
  0205,				/*   0111 LD_UP|LD_DOWN|LD_LEFT */
  ' ',				/*   1000 */
  0202,				/*   1001 LD_UP|LD_RIGHT */
  0200,				/*   1010 LD_DOWN|LD_RIGHT */
  0204,				/*   1011 LD_UP|LD_DOWN|LD_RIGHT */
  0210,				/*   1100 LD_RIGHT|LD_LEFT     */
  0207,				/*   1101 LD_UP|LD_LEFT|LD_RIGHT */
  0206,				/*   1110 LD_DOWN|LD_LEFT|LD_RIGHT */
  0212				/*   1111 ALL */
};
#endif /* KOI8 */

void
teldup ()
{
  glinedraw (LD_UP);
  Ty--;
}

void
telddown ()
{
  glinedraw (LD_DOWN);
  Ty++;
}

void
teldright ()
{
  glinedraw (LD_RIGHT);
  Tx++;
}

void
teldleft ()
{
  if (Tx == 0)
    exc (E_EDTBEG);
  glinedraw (LD_LEFT);
  Tx--;
}

static void
glinedraw (register small ldc)
{
  register small x;
  register large y;
  if ((x = Tx) == Ttxt->txrm)
    exc (E_EDTEND);
  y = Ty;
  if (textldc (x, y - 1) & LD_DOWN)
    ldc |= LD_UP;
  if (textldc (x, y + 1) & LD_UP)
    ldc |= LD_DOWN;
  if (textldc (x + 1, y) & LD_LEFT)
    ldc |= LD_RIGHT;
  if (textldc (x - 1, y) & LD_RIGHT)
    ldc |= LD_LEFT;

  if (ldc == LD_UP || ldc == LD_DOWN)
    ldc = LD_UP | LD_DOWN;
  if (ldc == LD_LEFT || ldc == LD_RIGHT)
    ldc = LD_LEFT | LD_RIGHT;
  Lebuf[x] = ctotc (ldctoasc[ldc]);
  if (++x > Lleng)
    Lleng = x;
  TxTRep (Ttxt, Lebuf, Lleng);
}

static char
textldc (register small x, register large y)
{
  register txt *t;
  register small c;
  t = Ttxt;
  if (y < 0 || x < 0 || x >= t->txrm)
    goto none;
  TxSetY (t, y);
  if (qTxDown (t))
    goto none;
  if (x >= (Lleng = TxFRead (t, Lebuf)))
    goto none;
  if ((c = (small) Lebuf[x] & 0377) >= LDC_FIRST && c <= LDC_LAST)
    return (asctoldc[c - LDC_FIRST]);
none:return (0);
}
