#include "cmn.h"
#include "kblinux-console.h"

#define MAXTABS 50
#define MAXLENTAB 200
#define FILE_NAME "linux-console.cm"

short origin_table[] = {
/* Line edit commands */
LE_RIGHT   ,RIGHT,             -1,/*  0401    */        /* make a step on right */
LE_LEFT    ,LEFT,              -1,/*  0402    */        /* make a step on left */
LE_DEOL    ,CTRL_D,            -1,/*  0403       */     /* delete end of line */
LE_DBGOL   ,CTRL_U,            -1,/*  0404       */     /* delete begining of line */
LE_BEG     ,PF2,LEFT,          -1,/*  0405      */      /* go to begining of line */
LE_BEG     ,Home,              -1,/*  0405      */      /* go to begining of line */
LE_END     ,PF2,RIGHT,         -1,/*  0406      */      /* goto end of line */
LE_END     ,End,               -1,/*  0406      */      /* goto end of line */
LE_IC      ,N9,                -1,/*  0407    */        /* insert space */
LE_IC      ,Insert,            -1,/*  0407    */        /* insert space */
LE_DC      ,N8,                -1,/*  0410      */      /* delete current character */
LE_DC      ,Delete,            -1,/*  0410      */      /* delete current character */
LE_BS      ,ZB,                -1,/*  0411       */     /* delete previous character */
LE_TAB     ,Tab,               -1,/*  0412       */     /* right tabulation */
LE_LTAB    ,LALT_Tab,          -1,/*  0413      */      /* left tabulation */
LE_NWORD   ,                   -1,/*  0414    */        /* go to next word */
LE_PWORD   ,                   -1,/*  0415    */        /* go to previous word */
LE_CCHAR   ,F1,                -1,/*  0416      */      /* save character */
LE_CDCHAR  ,SHIFT_F1,          -1,/*  0417      */      /* save character and delete it */
LE_PCHAR   ,F2,                -1,/*  0420     */       /* restore character(s) */
LE_CCCHR   ,SHIFT_F2,          -1,/*  0421     */       /* drop character(s) */
LE_CWORD   ,PF3,F1,            -1,/*  0422    */        /* save word */
LE_CWORD   ,LALT_Tilda,        -1,/*  0422    */        /* save word */
LE_CDWORD  ,PF3,SHIFT_F1,      -1,/*  0423      */      /* save word and delete it */
LE_DWORD   ,CTRL_A,            -1,/*  0424        */    /* delete current word with trailing spaces */
LE_DLWORD  ,PF3,CTRL_A,        -1,/*  0425     */       /* delete previous word */
LE_CENTR   ,CTRL_T,            -1,/*  0426    */        /* center current line */

LE_RINS    ,PF2,N9,            -1,/*  0427     */       /* set insert mode */
LE_RREP    ,PF2,N8,            -1,/*  0430     */       /* set replace mode */
LE_CCUP    ,Pause  ,UP,        -1,/*  0431    ->  */    /* current letter -> upper */
LE_CCDWN   ,Pause  ,DOWN,      -1,/*  0432    ->  */    /* current letter -> lower */
LE_CCRUS   ,Pause  ,RIGHT,     -1,/*  0433    ->  */    /* current letter -> russian */
LE_CCLAT   ,Pause  ,LEFT,      -1,/*  0434    ->  */    /* current letter -> english */
LE_CATTR   ,Pause  ,Enter,     -1,/*  0435    ->   */
LE_SPCHAR  ,CTRL_X,            -1,/*  0436     <> */
LE_ESCCHR  ,                   -1,/*  0437  ESC (033 ) */
LE_HFCHAR  ,                   -1,/*  0440           */
/* attributes */
LA_NONE    ,F10,' ',           -1,/*  0441    :=  */    /* input attributes = clear */
LA_LIGHT   ,F10,'S',           -1,/*  0442    ^=   */   /* input attributes |= light */
LA_LIGHT   ,F10,'s',           -1,/*  0442    ^=   */   /* input attributes |= light */
LA_ITALIC  ,F10,'M',           -1,/*  0443    ^=    */  /* input attributes |= italic */
LA_ITALIC  ,F10,'m',           -1,/*  0443    ^=    */  /* input attributes |= italic */
LA_UNDERL  ,F10,'P',           -1,/*  0444    ^=  */    /* input attributes |= underline */
LA_UNDERL  ,F10,'p',           -1,/*  0444    ^=  */    /* input attributes |= underline */
LA_BOLD    ,F10,'V',           -1,/*  0445    ^=   */   /* input attributes |= bold */
LA_BOLD    ,F10,'v',           -1,/*  0445    ^=   */   /* input attributes |= bold */
LA_SUPER   ,F10,'W',           -1,/*  0446    ^=   */   /* input attributes |= superscr */
LA_SUPER   ,F10,'w',           -1,/*  0446    ^=   */   /* input attributes |= superscr */
LA_SUBSCR  ,F10,'N',           -1,/*  0447    ^=   */   /* input attributes |= subscr */
LA_SUBSCR  ,F10,'n',           -1,/*  0447    ^=   */   /* input attributes |= subscr */
LA_INVERT  ,F10,'I',           -1,/*  0450    ^=  */    /* input attributes |= inverse */
LA_INVERT  ,F10,'i',         -1,/*  0450    ^=  */      /* input attributes |= inverse */

TE_UNDO    ,F8,                -1,/*  0451   */         /* undo */
TE_UNUNDO  ,SHIFT_F8,          -1,/*  0452    */        /* undo undo */
TE_SUNDO   ,                   -1,/*  0453   */
TE_SUNUNDO ,                   -1,/*  0454    */

TE_SRO     ,PF2, CTRL_T,       -1,/*  0455      */      /* set read-only mode */
TE_SRW     ,PF2, CTRL_E,       -1,/*  0456      */      /* cancel read-only mode */
TE_SWIGHT  ,PF2, CTRL_W,       -1,/*  0457     */

TE_SMARK   ,PF2,Return ,       -1,/*  0460    */        /* set marker */
TE_CMARK   ,PF2,Tab,           -1,/*  0461      */      /* swap marker and cursor */

TE_PTAB    ,PF3,Tab,           -1,/*  0462     */       /* edit tabulation pattern */
TE_PFORM   ,PF3,F9,            -1,/*  0463     */       /* edit formatting pattern */

TE_UP      ,UP,                -1,/*  0464    */        /* up */
TE_DOWN    ,DOWN,              -1,/*  0465    */        /* down */
TE_CUP     ,PF2, UP,           -1,/*  0466    */
TE_CDOWN   ,PF2, DOWN,         -1,/*  0467    */
TE_TBEG    ,PF1, UP,           -1,/*  0470     */       /* go to top of file */
TE_TBEG    ,PF1, PageUp,       -1,/*  0470     */       /* go to top of file */
TE_TEND    ,PF1, DOWN,         -1,/*  0471     */       /* go to end of file */
TE_TEND    ,PF1, PageDown,     -1,/*  0471     */       /* go to end of file */

TE_IL      ,N6,                -1,/*  0472    */        /* insert empty line */
TE_IL      ,N0,                -1,/*  0472    */        /* insert empty line */
TE_DL      ,N5,                -1,/*  0473    */        /* delete current line */
TE_DL      ,DOT,               -1,/*  0473    */        /* delete current line */
TE_CR      ,Return ,           -1,/*  0474      */      /* goto beginning of next line */
TE_RCR     ,PF3,Return,        -1,/*  0475      */      /* goto beginning of previous line */
TE_SENTR   ,PF2,N7,            -1,/*  0476      .  */   /* input search pattern */
TE_SENTR   ,PF2,  F7,          -1,/*  0476      .  */   /* input search pattern */
TE_SUP     ,PF3,N7,            -1,/*  0477    */        /* search up */
TE_SUP     ,SHIFT_F7,          -1,/*  0477    */        /* search up */
TE_SDOWN   ,N7,                -1,/*  0500    */        /* search down */
TE_SDOWN   ,F7,                -1,/*  0500    */        /* search down */
TE_CENTR   ,CTRL_E,            -1,/*  0501       */     /* move current line to middle of screen */
TE_CLIN    ,F3,                -1,/*  0502      */      /* save current line */
TE_CDLIN   ,SHIFT_F3,          -1,/*  0503      */      /* save current line and delete it */
TE_PLIN    ,F4,                -1,/*  0504     */       /* restore line(s) */
TE_CCLIN   ,SHIFT_F4,          -1,/*  0505     */       /* drop line(s) */
TE_BLIN    ,PF2,N6,            -1,/*  0506      */      /* cut line without horizontal displacement */
TE_SLIN    ,PF3,N6,            -1,/*  0507     */       /* paste line without horizontal displacement */
TE_NBLIN   ,PF2,N5,            -1,/*  0510    */        /* cut line with horizontal displacement */
TE_NSLIN   ,PF3,N5,            -1,/*  0511     */       /* paste line with horizontal displacement */
TE_IPAGE   ,                   -1,/*  0512     */
TE_NPAGE   ,PageDown,          -1,/*  0513     */       /* next screen */
TE_PPAGE   ,PageUp,            -1,/*  0514     */       /* previous screen */

TE_MCBLOCK ,F5,                -1,/*  0515    /  */     /* mark begin of block/save block */
TE_CDBLOCK ,PF2, F5,           -1,/*  0516      */      /* save and remove block */
TE_CCBLOCK ,PF3, F5,           -1,/*  0517      */      /* save and clear block */
TE_PBLOCK  ,F6,                -1,/*  0520    */        /* restore block over text */
TE_PIBLOCK ,PF2,F6,            -1,/*  0522    */        /* insert saved block */
TE_PRBLOCK ,PF3,F6,            -1,/*  0521       */     /* restore block over text from bottom right cornner */
TE_PIRBLOCK,PF2,PF3,F6,        -1,/*  0523       */     /* insert saved block from bottom right cornner */
TE_BLKEXEC ,PF1,Enter,         -1,/*  1021 */           /* executr OS command from block */
TE_BLKMOD  ,Pause, F5,         -1,/*  ???? */

TE_LPRINT  ,                   -1,/*  0524    */
TE_FPRINT  ,                   -1,/*  0525      */
TE_FORMAT  ,F9,                -1,/*  0526       */     /* reformat within intrinsic bounds */
TE_PFORMAT ,PF2,F9,            -1,/*  0527      */      /* reformat by pattern */
TE_LDUP    ,                   -1,/*  0530     */
TE_LDDOWN  ,                   -1,/*  0531     */
TE_LDLEFT  ,                   -1,/*  0532     */
TE_LDRIGHT ,                   -1,/*  0533     */
TE_RUP     ,PF3,PF1,N7,        -1,/*  0534      */
TE_RUP     ,PF3,PF1,F7,        -1,/*  0534      */
TE_RDOWN   ,PF1,N7,            -1,/*  0535      */
TE_RDOWN   ,PF1,F7,            -1,/*  0535      */

TE_PPRINT  ,                   -1,/*  0536  c   */
TE_IPRINT  ,                   -1,/*  0537  (  */
TE_RSPELL  ,                   -1,/*  0540  /  */
TE_LSPELL  ,                   -1,/*  0541 */
TE_SPADD   ,                   -1,/*  0542  $    */
TE_SPREM   ,                   -1,/*  0543     */

TE_HELP    ,PF3 ,'/',          -1,/*  0544  help */     /* help */
TE_HELP    ,PF3 ,'?',          -1,/*  0544  help */     /* help */
TE_HELP    ,                   -1,/*  0544  help */
TE_OHELP   ,PF3 ,'H',          -1,/*  0544  help */
TE_OHELP   ,PF3 ,'h',          -1,/*  0544  help */
TE_MSGS    ,PF3 ,'!',          -1,/* 01013     */
TE_FENTR   ,PF3 ,DOWN,         -1,/*  0545      */      /* enter file */
TE_FENTR   ,ESC ,DOWN,         -1,/*  0545      */      /* enter file */
TE_F2ENTR  ,PF3 ,PF3 ,         -1,/*  0546         */   /* obtain filename and enter file */
TE_FEXEC   ,PF3,Enter,         -1,/*  0547    */        /* execute command */
TE_EXEC    ,PF2,Enter,         -1,/*  0547    */        /* execute ??? */
TE_SHELL   ,Enter,             -1,/*  0550      cli */  /* OS shell */
TE_INFILE  ,PF3,RIGHT,         -1,/*  0551  enter filename */   /* enter filename in micros.dir */
TE_INFILE  ,ESC,RIGHT,         -1,/*  0551  enter filename */   /* enter filename in micros.dir */
TE_OUTFILE ,PF3,LEFT,          -1,/*  0552  leave filename */   /* leave filename in micros.dir */
TE_OUTFILE ,ESC,LEFT,          -1,/*  0552  leave filename */   /* leave filename in micros.dir */

TE_MODFILE ,PF3,CTRL_U,        -1,                      /* get file attribute of filename in cursor position */
TE_MODFILE ,ESC,CTRL_U,        -1,                      /* get file attribute of filename in cursor position */

TE_UPDATE  ,PF2,PF3,           -1,/*  0553      */      /* update file */
TE_EXIT    ,PF3,UP,            -1,/*  0554       */     /* exit file */
TE_EXIT    ,ESC,UP,            -1,/*  0554       */     /* exit file */
TE_EXOLD   ,PF3,CTRL_C,        -1,/*  0555       */     /* exit without save */
TE_EXOLD   ,ESC,CTRL_C,        -1,/*  0555       */     /* exit without save */
TE_EXBACK  ,PF3,ZB,            -1,/*  0556         */   /* exit file with bakup copy */
TE_EXBACK  ,ESC,ZB,            -1,/*  0556         */   /* exit file with bakup copy */
TE_QUIT    ,PF1,PF1,           -1,/*  0557      */      /* quit from editor */
TE_CRASH   ,CTRL_Y,            -1,/*  0557      */      /* quit from editor */
TE_NERROR  ,F10, DOWN,         -1,/*  0560      */      /* next error */
TE_PERROR  ,F10, UP,           -1,/*  0561     */       /* previous error */
TE_ERROFF  ,F10, F10,          -1,/* 01022     */       /* exit error mode */
TE_STATUS  ,CTRL_P,            -1,/* 01023     */       /* toggle status line */
TE_ZOOM    ,CTRL_R,            -1,/* 01024     */       /* zoom */
TE_UFORK   ,PF3,PF4,UP,        -1,/*  0562   fork    */ /* horizontal split of window */
TE_DFORK   ,PF3,PF4,DOWN,      -1,/* 01003   fork    */ /* horizontal split of window */
TE_RFORK   ,PF3,PF4,RIGHT,     -1,/* 01004     fork  */ /* vertical split of window */
TE_LFORK   ,PF3,PF4,LEFT,      -1,/*  0563     fork  */ /* vertical split of window */
TE_UWINDOW ,PF4,UP,            -1,/*  0564     */       /* switch to up window */
TE_DWINDOW ,PF4,DOWN,          -1,/*  0565      */      /* switch to down window */
TE_RWINDOW ,PF4,RIGHT,         -1,/*  0566      */      /* switch to right window */
TE_LWINDOW ,PF4,LEFT,          -1,/*  0567       */     /* switch to left window */

TK_BREAK   ,CTRL_C,            -1,/*  0570       */     /* interrupt a long command */
TK_CTRLO   ,CTRL_Z,            -1,/*  0571  /    */
TK_REPEAT  ,PF2 ,PF2 ,         -1,/*  0573  /   */      /* repeat previous command */
TK_REPEAT  ,Pause, Pause,        -1,/*  0573  /   */      /* repeat previous command */
TK_BRUS    ,                   -1,/*  0574  /   */
TK_SRUS    ,CTRL_O,            -1,/*  0575  a    */     /* switch to russian */
TK_BLAT    ,                   -1,/*  0576  /   */
TK_SLAT    ,CTRL_N,            -1,/*  0577  a    */     /* switch to latin */
TK_SM+0    ,PF2 ,N4,           -1,/*  0600     0 */     /* initiate capturing of macro */
TK_EM+0    ,N4,                -1,/*  0700   /  0 */    /* finish capturing of macro / excute macro */
TK_SM+1    ,PF2,PF1,'1',       -1,/*  0600     0 */     /* initiate capturing of macro 1 */
TK_EM+1    ,    PF1,'1',       -1,/*  0700   /  0 */    /* finish capturing of macro 1 / excute macro 1 */
TK_SM+2    ,PF2,PF1,'2',       -1,/*  0600     0 */     /* initiate capturing of macro 2 */
TK_EM+2    ,    PF1,'2',       -1,/*  0700   /  0 */    /* finish capturing of macro 2 / excute macro 2 */
TK_SM+3    ,PF2,PF1,'3',       -1,/*  0600     0 */     /* initiate capturing of macro 3 */
TK_EM+3    ,    PF1,'3',       -1,/*  0700   /  0 */    /* finish capturing of macro 3 / excute macro 3 */
TK_SM+4    ,PF2,PF1,'4',       -1,/*  0600     0 */     /* initiate capturing of macro 4 */
TK_EM+4    ,    PF1,'4',       -1,/*  0700   /  0 */    /* finish capturing of macro 4 / excute macro 4 */
TK_SM+5    ,PF2,PF1,'5',       -1,/*  0600     0 */     /* initiate capturing of macro 5 */
TK_EM+5    ,    PF1,'5',       -1,/*  0700   /  0 */    /* finish capturing of macro 5 / excute macro 5 */
TK_SM+6    ,PF2,PF1,'6',       -1,/*  0600     0 */     /* initiate capturing of macro 6 */
TK_EM+6    ,    PF1,'6',       -1,/*  0700   /  0 */    /* finish capturing of macro 6 / excute macro 6 */
TK_SM+7    ,PF2,PF1,'7',       -1,/*  0600     0 */     /* initiate capturing of macro 7 */
TK_EM+7    ,    PF1,'7',       -1,/*  0700   /  0 */    /* finish capturing of macro 7 / excute macro 7 */
TK_SM+8    ,PF2,PF1,'8',       -1,/*  0600     0 */     /* initiate capturing of macro 8 */
TK_EM+8    ,    PF1,'8',       -1,/*  0700   /  0 */    /* finish capturing of macro 8 / excute macro 8 */
TK_SM+9    ,PF2,PF1,'9',       -1,/*  0600     0 */     /* initiate capturing of macro 9 */
TK_EM+9    ,    PF1,'9',       -1,/*  0700   /  0 */    /* finish capturing of macro 9 / excute macro 9 */
TK_SM+10   ,PF2,PF1,'0',       -1,/*  0600     0 */     /* initiate capturing of macro 0 */
TK_EM+10   ,    PF1,'0',       -1,/*  0700   /  0 */    /* finish capturing of macro 0 / excute macro 0 */
TK_SM+11   ,                   -1,/*  0600     0 */     /* initiate capturing of macro 11 */
TK_EM+11   ,                   -1,/*  0700   /  0 */    /* finish capturing of macro 11 / excute macro 11 */

TD_SDEQ    ,PF1,'D',           -1,/*  01001* /  */
TD_SDEQ    ,PF1,'d',           -1,/*  01001* /  */
TD_STXT    ,                   -1,/*  01002* /  */
TK_CSET    ,PF1,N2,            -1,                      /* set counter value */
TK_CADD    ,PF1,N9,            -1,                      /* increment counter value */
TK_CSUB    ,PF1,N8,            -1,                      /* decrement counter value */
TK_CPUT    ,PF1,N3,            -1,                      /* put counter value */
TK_DATE    ,PF1,CTRL_D,        -1,/*  01005  "   */     /* put current date */
TK_TIME    ,PF1,CTRL_T,        -1,/*  01006  "   */     /* put current time */
MC_SAVE    ,PLUS ,PLUS ,N0,    -1,/*  01007  "    */
MC_LOAD    ,PLUS ,PLUS ,DOT,   -1,/*  01010  '    */
MC_2SAVE   ,PLUS ,N0,          -1,/*  01011  "    */
MC_2LOAD   ,PLUS ,DOT,         -1,/*  01012  '    */

EX_EXEC    ,N1,                -1,/*  02030  '    */
EX_CEXEC   ,PF2,N1,            -1,/*  02031  '    */
EX_FEXEC   ,PF3,N1,            -1,/*  02032  '    */

MC_INCL    ,PF3,'j',           -1,/*  02000  '    */
MC_INCL    ,PF3,'J',           -1,/*  02000  '    */
MC_PROC    ,PF3,'p',           -1,/*  02000  '    */
MC_PROC    ,PF3,'P',           -1,/*  02000  '    */
MC_SUBR    ,PF3,'s',           -1,/*  02000  '    */
MC_SUBR    ,PF3,'S',           -1,/*  02000  '    */
MC_FUNC    ,PF3,'f',           -1,/*  02000  '    */
MC_FUNC    ,PF3,'F',           -1,/*  02000  '    */
MC_WHIL    ,PF3,'w',           -1,/*  02000  '    */
MC_WHIL    ,PF3,'W',           -1,/*  02000  '    */
MC_DO      ,PF3,'d',           -1,/*  02000  '    */
MC_DO      ,PF3,'D',           -1,/*  02000  '    */
MC_IF      ,PF3,'i',           -1,/*  02000  '    */
MC_IF      ,PF3,'I',           -1,/*  02000  '    */

DB_DUMP    ,PF3,'V',           -1,/*  02000  '    */
DB_DUMP    ,PF3,'v',           -1,/*  02000  '    */
DB_TRACE   ,PF3,'T',           -1,/*  02000  '    */
DB_TRACE   ,PF3,'t',           -1,/*  02000  '    */

-1};

short special_table[] = {

TK_ESC     ,ESC,               /*  0572  Esc */
TK_ESC     ,PF3,               /*  0572  Esc */
-1};

#include "gentab.c"
