/*
 *     nanoMir
 *
 *     "cp.c"           strings & blocks copy
 */

#include "nm.h"
#include "tx.h"
#include "cm.h"
#include "te.h"
#include "le.h"
#include "kb.h"
#include "cp.h"
#include "fs.h"
#include "tm.h"
#include "wd.h"
#include <stdio.h>


txt *LCtxt, *Btxt;
static large tmarkx;
static small tmarky;
static bool lcempt;
static bool bempt;
bool BlockMark;
static small blkxmin, blkxsiz, blkydir;
small blkxdir;
static large blkysiz;

local void tecutlin (void), gcutblock (small), gpasteblock (small);
local void cpccase (char *pat), cpacase (unsigned bm);
local void temarkblock (void);

void
cpInit ()
{
  lcempt = FALSE;
  bempt = FALSE;
  BlockMark = FALSE;
  blkxdir = 0;
  LCtxt = Btxt = NIL;
}

#define otkaz( s ) { perror( s ); _exit( 1 ); }

static void
cptinit (bool qload)
{
  register txt *t;
  if (LCtxt == NIL)
    {
      char *SavFilNam;
      if (!(SavFilNam = (char *) getenv ("MIMLINESAVEFILE")))
	{
	  SavFilNam = DEFAULT_MIMLINESAVEFILE;
	  /* otkaz("MIMLINESAVEFILE undefined."); */
	}
      LCtxt = t = tmdesc (SavFilNam, FALSE);
      assert (t != NIL, "cptinit");
      if (qload)
	{
	  if (!tmload (t))
	    exc (E_CPFER);
	}
      else
	tmnew (t);		/* For optimizing purpose only simulate empty file */
      t->txstat |= (TS_PERM1 | TS_FILE);
    }
}

static void
cpbinit (bool qload)
{
  register txt *t;
  if (Btxt == NIL)
    {
      char *SavFilNam;
      if (!(SavFilNam = (char *) getenv ("MIMBLOCKSAVEFILE")))
	{
	  SavFilNam = DEFAULT_MIMBLOCKSAVEFILE;
	  /* otkaz("MIMBLOCKSAVEFILE undefined."); */
	}
      Btxt = t = tmdesc (SavFilNam, FALSE);
      assert (t != NIL, "cpbinit");
      if (qload)
	{
	  if (!tmload (t))
	    exc (E_CPFER);
	}
      else
	tmnew (t);		/* For optimizing purpose only simulate empty file */
      t->txstat |= (TS_PERM1 | TS_FILE);
    }
}

#undef otkaz

void
cpsave ()
{
  register txt *t;
  if ((t = LCtxt) != NIL && (t->txstat & TS_CHANGE))
    tmsave (t, FALSE);
  if ((t = Btxt) != NIL && (t->txstat & TS_CHANGE))
    tmsave (t, FALSE);
}

static void
tecutlin ()
{
  register txt *lct;
  cptinit (FALSE);
  lct = LCtxt;
  if (lcempt)
    {
      TxEmpt (lct);
      lcempt = FALSE;
    }
  TxSetY (lct, MAXLARGE);
  Lleng = TxFRead (Ttxt, Lebuf);
  TxTIL (lct, Lebuf, Lleng);
  lct->txstat |= TS_CHANGE;
}

static void
tecutplin (small x, small dx, bool q)
{
  register txt *lct;
  small i;
/*cptinit( FALSE );     */
  if (q)
    {
      lct = Btxt;
      if (bempt)
	{
	  TxEmpt (lct);
	  bempt = FALSE;
	}
      TxSetY (lct, MAXLARGE);
    }
  Lleng = TxFRead (Ttxt, Lebuf);
  if (q)
    {
      Lleng = min (dx, Lleng - x);
      TxTIL (lct, Lebuf + x, Lleng);
      lct->txstat |= TS_CHANGE;
    }
}

void
teclin ()
{
  tecutlin ();
  blkxdir = 0;
  Ty++;
}

void
tecdlin ()
{
  tecutlin ();
  blkxdir = 0;
  tedl ();
}

void
teplin ()
{
  register txt *lct;
  register small l;
  cptinit (TRUE);
  lcempt = TRUE;
  lct = LCtxt;
  TxSetY (lct, MAXLARGE);
  while (!qTxUp (lct))
    {
      TxUp (lct);
      l = TxTRead (lct, Lebuf);
      TxTIL (Ttxt, Lebuf, l);
      if (qKbBreak ())
	exc (E_KBREAK);
    }
}

void
tecclin ()
{
  cptinit (FALSE);
  TxEmpt (LCtxt);
  lcempt = FALSE;
}

#define C_NONE    0
#define C_DELETE  1
#define C_CLEAR   2

static void
temarkblock ()
{
  register txt *t;
  tmarkx = Tx;
  tmarky = Ty;
}

void
temcblock ()
{
  if (Tx == Ttxt->txrm)
    exc (E_MOVEND);
  if (BlockMark)
    gcutblock (C_NONE);
  else
    {
      temarkblock ();
      BlockMark = TRUE;
    }
}

void
tecdblock ()
{
  gcutblock (C_DELETE);
}
void
teccblock ()
{
  gcutblock (C_CLEAR);
}

void
teblkmod ()
{
  register small l;
  register large y, ymin, ymax;
  register txt *t;
  unsigned bm;
  if (!KbRadix || !KbCount)
    return;
  bm = KbCount;
  t = Ttxt;
  if (!BlockMark)
    exc (E_NOBLOCK);
  if (Tx == t->txrm)
    exc (E_MOVEND);
  if (t->txredit != TXED_YES)
    exc (E_CHANGE);
  if (Ty < tmarky)
    {
      ymin = Ty;
      ymax = tmarky;
    }
  else
    {
      ymax = Ty;
      ymin = tmarky;
    }
  if (Tx < tmarkx)
    {
      blkxmin = Tx;
      blkxsiz = tmarkx - Tx + 1;
    }
  else
    {
      blkxmin = tmarkx;
      blkxsiz = Tx - tmarkx + 1;
    }
  for (y = ymin; y <= ymax; y++)
    {
      TxSetY (t, y);
      tecutplin (blkxmin, blkxsiz, FALSE);
      switch (bm & 3)
	{
	case 0:
	  break;
	case 1:
	  cpccase ("az\340 \040");
	  break;
	case 2:
	  cpccase ("az\340 \040AZ\040\377\340");
	  break;
	case 3:
	  cpccase ("AZ\040\377\340");
	  break;
	}
      switch ((bm >> 2) & 3)
	{
	case 0:
	  break;
	case 1:
	  cpccase ("\240\140");
	  break;
	case 2:
	  cpccase ("az\140AZ\240\240\140");
	  break;
	case 3:
	  cpccase ("az\140AZ\240");
	  break;
	}
      cpacase ((bm >> 4) & 7);
      TxFRep (t, Lebuf);
    }
}

static void
cpacase (unsigned bm)
{
  register tchar tc, *tp;
  register char c, *p;
  register small l;
  for (tp = Lebuf + blkxmin, l = blkxsiz; l; l--)
    {
      tc = *tp;
      if (bm & 4)
	tc &= 0377;
      switch (bm & 3)
	{
	case 0:
	  break;
	case 1:
	  tc |= (Lattr);
	  break;
	case 2:
	  tc ^= (Lattr);
	  break;
	case 3:
	  tc &= ~(Lattr);
	  break;
	}
      *tp++ = tc;
    }
}

static void
cpccase (char *pat)
{
  register tchar tc, *tp;
  register char c, *p;
  register small l;
  for (tp = Lebuf + blkxmin, l = blkxsiz; l; l--)
    {
      c = (tc = *tp);
      for (p = pat; *p; p += 3)
	if (c >= p[0] && c <= p[1])
	  {
	    c += p[2];
	    break;
	  }
      *tp++ = (tc & ~0377) | (c & 0377);
    }
}

void
cpblockmark ()
{
  register wnd *w;
  box b;
  b.x0 = tmarkx;
  b.y0 = tmarky;
  b.x1 = Lx;
  b.y1 = Ty;
  for (w = (wnd *) Ttxt->txwndptr; w != NIL; w = w->wnext)
    wblock (w, b);
}

void
cpblockoff ()
{
  register wnd *w;
  BlockMark = FALSE;
#ifdef X11
  for (w = (wnd *) Ttxt->txwndptr; w != NIL; w = w->wnext)
    wblkoff (w);
#else
  wblkoff ();
#endif
}

static void
gcutblock (register small typ)
{
  register small l;
  register large y, ymin, ymax;
  register txt *t;
  t = Ttxt;
  if (!BlockMark)
    exc (E_NOBLOCK);
  if (Tx == t->txrm)
    exc (E_MOVEND);
  if (typ != C_NONE && t->txredit != TXED_YES)
    exc (E_CHANGE);
  BlockMark = FALSE;
  cpblockoff ();
  cpbinit (FALSE);
  if (Ty < tmarky)
    {
      ymin = Ty;
      ymax = tmarky;
      blkydir = (-1);
    }
  else
    {
      ymax = Ty;
      ymin = tmarky;
      blkydir = 1;
    }
  if (Tx < tmarkx)
    {
      blkxmin = Tx;
      blkxsiz = tmarkx - Tx + 1;
      blkxdir = (-1);
    }
  else
    {
      blkxmin = tmarkx;
      blkxsiz = Tx - tmarkx + 1;
      blkxdir = 1;
      if (typ == C_DELETE)
	Tx = tmarkx;
    }
  bempt = TRUE;
  blkysiz = ymax - ymin + 1;
  for (y = ymin; y <= ymax; y++)
    {
      TxSetY (t, y);
      tecutplin (blkxmin, blkxsiz, TRUE);
      if (typ != C_NONE)
	{
	  if (typ == C_CLEAR)
	    blktspac (Lebuf + blkxmin, blkxsiz);
	  if (typ == C_DELETE)
	    {
	      l = blkxmin + blkxsiz;
	      blkmov ((char *) (Lebuf + l), (char *) (Lebuf + blkxmin),
		      (MAXLPAC - l) * sizeof (tchar));
	      blktspac (Lebuf + (MAXLPAC - blkxsiz), blkxsiz);
	    }
	  TxFRep (t, Lebuf);
	}
    }
}

#define C_REV 1
#define C_INS 2

void
tepblock ()
{
  gpasteblock (0);
}
void
teprblock ()
{
  gpasteblock (C_REV);
}
void
tepiblock ()
{
  gpasteblock (C_INS);
}
void
tepirblock ()
{
  gpasteblock (C_INS | C_REV);
}

static void
gpasteblock (register small typ)
{
  register small xleft, xright, l;
  register large y, yup, ydown;
  small dx, dy;
  bool q;

  if (blkxdir == 0)
    exc (E_NOBLOCK);
  cpbinit (TRUE);
  dx = blkxdir;
  dy = blkydir;
  if (typ & C_REV)
    {
      dx = -dx;
      dy = -dy;
    }
  if (dy < 0)
    yup = (ydown = Ty + 1) - blkysiz;
  else
    ydown = (yup = Ty) + blkysiz;
  if (dx < 0)
    xleft = (xright = Tx + 1) - blkxsiz;
  else
    xright = (xleft = Tx) + blkxsiz;

  if (xleft < 0 || xright > MAXLPAC || (yup < 0 ||
					(q = !TxSetY (Ttxt, ydown)))
      && !(typ & C_INS))
  blockout:exc (E_BLOCKOUT);
  if (typ & C_INS)
    {
      for (y = ydown; y > yup;)
	{
	  TxSetY (Ttxt, --y);
	  if (blkxsiz + TxTRead (Ttxt, Lebuf) > MAXLPAC)
	    goto blockout;
	}
      if (yup < 0)
	{
	  TxSetY (Ttxt, 0);
	  for (y = yup; y < 0; y++)
	    teil ();
	  ydown -= yup;
	  Ty -= yup;
	  yup -= yup;
	}
      if (q)
	{
	  TxSetY (Ttxt, MAXLARGE);
	  for (y = Ttxt->txy; y < ydown; y++)
	    teil ();
	}
      TxSetY (Ttxt, ydown);
      if (dx > 0)
	Tx += blkxsiz;
    }
  for (TxSetY (Btxt, MAXLARGE), y = ydown; y > yup;)
    {
      TxSetY (Ttxt, --y);
      l = TxFRead (Ttxt, Lebuf);
      TxUp (Btxt);
      TxFRead (Btxt, lfbuf);
      if ((typ & C_INS) && l > xleft)
	blkmov ((char *) (Lebuf + xleft), (char *) (Lebuf + xright),
		(l - xleft) * sizeof (tchar));
      blkmov ((char *) (lfbuf), (char *) (Lebuf + xleft),
	      blkxsiz * sizeof (tchar));
      TxFRep (Ttxt, Lebuf);
    }
}

void
blkexec ()
{
  register small l, stat, maxstat;
  register char *p;

  cpbinit (FALSE);

#ifndef X11
  scErase ();
  scDetach ();
#endif
  kbDetach ();

  maxstat = 0;
  for (TxSetY (Btxt, 0); !qTxDown (Btxt); TxDown (Btxt))
    {
      l = TxRead (Btxt, txbuf);
      p = txbuf;
      p[l] = 0;
      if (*p)
	{
	  char fnbuf[512];

	  printf ("+ %s\n", txbuf + blkxmin);
#ifdef VMS
	  sprintf (fnbuf, "spa /out=tt/nolog /prompt=\"NM> \" %s",
		   txbuf + blkxmin);
#endif
#ifdef UNIX
	  sprintf (fnbuf, "%s", txbuf + blkxmin);
#endif
	  stat = system (fnbuf);
	  maxstat = max (maxstat, stat);
	}
    }
  printf ("NM: stat=%d - press Return\n", maxstat);
  getchar ();

  kbAttach ();
#ifndef X11
  scAttach ();
  scErase ();
#endif
}
