/*
 *     nanoMir
 *
 *     "dq.c"           Deqs
 */

/*#define DQDEBUG*/

#include "nm.h"
#include "fs.h"
#include "dq.h"
#include "me.h"
#include "cm.h"
#include "er.h"

static deq *freedeqs;
static deq deqs[MAXDEQS + 2];
deq *gapdeq;
deq *lockdeq;
extern void dqdump (void);
extern char *lblkmov (char *, char *, addr);
local void extgap (deq *, addr, bool);
local addr extragap (deq *);
small DqRLen;
addr DqReqMem;
bool DqEconomy;
void
dqInit ()
{
  register deq *p, *p1;
  register small i;
  p = &deqs[MAXDEQS + 1];
  p->dnext = NIL;
  for (i = MAXDEQS; i; i--)
    {
      p1 = p;
      (--p)->dnext = p1;
    }
  freedeqs = p;
  p--;
  p->dnext = p->dprev = p;
  p->dbeg = p->dend = NIL;
  p->dbext = p->deext = NIL;
  gapdeq = NIL;
  DqEconomy = FALSE;
}

deq *
DqNew (small typ, small bext, small eext)
{
  register deq *d, *dn;
  d = freedeqs;
  assert (d != NIL, "DqNew");
  freedeqs = d->dnext;
  dn = deqs[0].dprev;
  d->dprev = dn;
  dn->dnext = d;
  d->dnext = &deqs[0];
  deqs[0].dprev = d;
  d->dbeg = d->dend = NIL;
  d->dtyp = typ;
  d->dbext = bext;
  d->deext = eext;
  d->dbmem = NIL;
  d->demem = NIL;
/*
  extgap( d, (addr)(bext+eext) );
  d->dbeg += bext;
  d->dend += bext;
*/
#ifdef DQDEBUG
  printf ("dqNew: deq=%x \n", d);
#endif
  return (d);
}

bool
DqDel (register deq * d)
{
  register deq *p;
  if (d == lockdeq)
    return (FALSE);
  (d->dnext)->dprev = d->dprev;
  (d->dprev)->dnext = d->dnext;
  if (d == gapdeq)
    gapdeq = d->dprev;
  d->dnext = freedeqs;
  freedeqs = d;
  if (d->dbmem)
    meFree (d->dbmem);
  d->dbeg = d->dend = NIL;
  d->dbmem = NIL;
  d->demem = NIL;
#ifdef DQDEBUG
  printf ("dqDel: deq=%x \n", d);
#endif

  return (TRUE);
}

void
DqAddB (register deq * d, char *p, register small l)
{
  register char *pb;
  register small *pi;
  register small le;
  if (d->dtyp == DT_ASC)
    {
      extgap (d, (addr) (l + 1), TRUE);
      pb = blkmov (p, d->dbeg -= l + 1, l);
      *pb++ = LF;
    }
  else
    {
      le = (l + 1) & (~1);
      extgap (d, (addr) (le + 4), TRUE);
      pi = (small *) (d->dbeg -= le + 4);
      *pi++ = l;
      pi = (small *) blkmov (p, (char *) pi, le);
      *pi++ = l;
    }
}

void
DqAddE (register deq * d, char *p, register small l)
{
  register char *pb;
  register small *pi;
  register small le;
  if (d->dtyp == DT_ASC)
    {
      extgap (d, (addr) (l + 1), FALSE);
      pb = blkmov (p, d->dend, l);
      *pb++ = LF;
      d->dend = pb;
    }
  else
    {
      le = (l + 1) & (~1);
      extgap (d, (addr) (le + 4), FALSE);
      pi = (small *) d->dend;
      *pi++ = l;
      pi = (small *) blkmov (p, (char *) pi, le);
      *pi++ = l;
      d->dend = (char *) pi;
    }
}

void
DqExtend (register deq * d, large l)
{
#ifdef DQDEBUG
  printf ("dqExtend: deq=%x\n", d, l);
#endif
  extgap (d, l, FALSE);
  d->dend += l;
}

small
DqCopForw (register deq * d, register large pos, char *p)
{
  register long l;
  register char *pb;
  register small *pi;
  register char *pbeg;
  assert (pos >= 0 && pos < DqLen (d), "DqCopForw");
  if (d->dtyp == DT_ASC)
    {
      pbeg = pb = d->dbeg + pos;
      while (*pb++ != LF);
      l = (DqRLen = (small) (pb - pbeg)) - 1;
      if (l && *(pb - 2) == CR)
	l--;
    }
  else
    {
      pi = (small *) (d->dbeg + pos);
      l = *pi++;
      DqRLen = ((l + 1) & (~1)) + 4;
      pbeg = (char *) pi;
    }
  if (l > MAXLUP)
    l = MAXLUP;
  if (p != NIL)
    blkmov (pbeg, p, l);
  return (l);
}

small
DqCopBack (register deq * d, register large pos, char *p)
{
  register char *p1;
  register long l;
  register small le;
  register char tc, *tp;
  register char *pend;
  assert (pos > 0 && pos <= DqLen (d), "DqCopBack");
  if (d->dtyp == DT_ASC)
    {
      p1 = (pend = d->dbeg + pos) - 1;
      tc = *(tp = d->dbeg - 1);
      *tp = LF;
      while (*--p1 != LF);
      p1++;
      *tp = tc;
      l = (DqRLen = (small) (pend - p1)) - 1;
      if (l && *(pend - 2) == CR)
	l--;
    }
  else
    {
      p1 = d->dbeg + pos - 2;
      l = *((small *) p1);
      le = (l + 1) & (~1);
      DqRLen = le + 4;
      p1 -= le;
    }
  if (l > MAXLUP)
    l = MAXLUP;
  if (p != NIL)
    blkmov (p1, p, l);
  return (l);
}

small
DqGetB (register deq * d, char *p)
{
  register small l;
  l = DqCopB (d, p);
  d->dbeg += DqRLen;
  return (l);
}

small
DqGetE (register deq * d, char *p)
{
  register small l;
  l = DqCopE (d, p);
  d->dend -= DqRLen;
  return (l);
}

void
DqEmpt (register deq * d)
{
  deq *prev, *next, *p;
  large ln;
  char *l;

  if (d->dbext < d->deext)
    d->dend = d->dbeg;
  else
    d->dbeg = d->dend;
#ifdef DQDEBUG
  printf ("dqEmpt: deq=%x \n", d);
#endif
}

void
DqCutB (register deq * d, register large pos)
{
  assert (pos >= 0 && pos <= DqLen (d), "DqCutB");
#ifdef DQDEBUG
  printf ("dqCutB: deq=%x pos %d \n", d, pos);
#endif
  d->dbeg += pos;
}

void
DqCutE (register deq * d, register large pos)
{
  assert (pos >= 0 && pos <= DqLen (d), "DqCutE");
#ifdef DQDEBUG
  printf ("dqCutE: deq=%x pos %d \n", d, pos);
#endif
  d->dend = d->dbeg + pos;
}

bool
DqLoad (register deq * d, fid f, large size)
{
  register char *p;
  jmp_buf dqlenv;
  register jmp_buf *nextexc;
  bool q = FALSE;

  nextexc = excptr;
  excptr = (jmp_buf *) dqlenv;
  if (setjmp (dqlenv))
    goto exc;
  DqEmpt (d);
  extgap (d, (addr) (size + 2), FALSE);
  p = d->dbeg + FsRead (f, size, d->dbeg);
  if (d->dtyp == DT_ASC)
    {
      if (*--p == NM_EOF)
	p--;
      if (*p++ != LF)
	*p++ = LF;
    }
  d->dend = p;
  q = TRUE;
exc:
  excptr = nextexc;
  return (q);
}

bool
DqSave (register deq * d, fid f)
{
  register large l;
  l = DqLen (d);
  return (FsWrite (f, l, d->dbeg) == l);
}

local void
extgap (register deq * d, register addr l, bool qpr)
{
  char *newmem;
  addr quantum, offset, dl;

  lockdeq = d;
#ifdef DQDEBUG
  printf ("extgap: deq=%x \n", d);
#endif

  if (qpr)
    {
      if (cpdiff (d->dbeg, d->dbmem) >= l)
	goto ret;
      quantum = max (d->dbext, l);
    }
  else
    {
      if (cpdiff (d->demem, d->dend) >= l)
	goto ret;
      quantum = max (d->deext, l);
    }
#ifdef DQDEBUG
  printf ("        quantum=%d \n", quantum);
#endif
  dl = d->demem - d->dbmem + quantum;
  if ((newmem = meAlloc (d->dbmem, dl)) == NIL)
    {
      ErLert ("Not enough memory");
      exc (E_NOMEM);
    }

  offset = cpdiff (newmem, d->dbmem);
  d->dbmem = newmem;
  d->demem = newmem + dl;
  d->dbeg += offset;
  d->dend += offset;

  if (qpr)
    {
      addr size = cpdiff (d->dend, d->dbeg);
      lblkmov (d->dbeg, d->dbeg + quantum, size);
      d->dbeg += quantum;
      d->dend += quantum;
#ifdef DQDEBUG
      printf ("        copy %x +%d \n", d->dbeg, size);
#endif
    }
ret:
  lockdeq = NIL;
  gapdeq = NIL;
  DqEconomy = FALSE;
}

addr
DqFree ()
{
  register deq *d = deqs;
  register large f = 0;
  do
    f += extragap (d);
  while ((d = d->dnext) != deqs);
  return (f);
}
static addr
extragap (register deq * d)
{
  return 0;
}

#ifdef DEBUG
void
dqdump ()
{
  register deq *d = deqs;
  do
    {
      dbprintf ("DqDump: %lx(%d) - %lx(%d) : %ld\n",
		d->dbeg, d->dbext, d->dend, d->deext, DqLen (d));
    }
  while ((d = d->dnext) != &deqs[0]);
}
#endif

large
DqLines (deq * d)
{
  register char *p;
  large l = 0;
  if (d->dtyp != DT_ASC)
    return l;
  for (p = d->dbeg; cpdiff (d->dend, p) > 0; p++)
    if (*p == LF)
      l++;
  return l;
}
