
/*
 *     nanoMir
 *
 *     "ki.c"           Keyboard interface
 */

#include "nm.h"
#include "kb.h"
#include "cm.h"
#include "cmn.h"
#include "trm.h"
#include "er.h"
#ifndef X11
#include "sc.h"
#include "tv.h"
#else
#include "xi.h"
#include "xi_keys.h"
#endif

#ifdef VMS
#include <descrip.h>
#include <iodef.h>
#include <ssdef.h>
#include <ttdef.h>
#include <file.h>
#endif

#ifdef UNIX

#ifdef MIM_SVR4

#include <termios.h>
#include <sys/ioctl.h>
typedef struct termios ttystruct;
#ifndef VDSUSP
#define VDSUSP VSUSP
#endif /* In some Unix (eg Linux) VDSUSP is not defined */
#ifndef IUCLC
#define IUCLC 0
#endif /* In some Unixes (eg FreeBSD) IUCLC is not defined */
#ifndef ONOCR
#define ONOCR 0
#endif /* In some Unixes (eg FreeBSD) ONOCR is not defined */

#else /* not def MIM_SVR4 */

#include <termio.h>
typedef struct termio ttystruct;
#define KI_FLUSH TCFLSH
#define KI_GET   TCGETA
#define KI_SET   TCSETAW

#endif /* def MIM_SVR4 */

#include <fcntl.h>

#ifndef O_SYNC
#define O_SYNC 0
#endif /* In some Unixes (eg FreeBSD) O_SYNC is not defined */

#endif /* def UNIX */

#include <signal.h>
#include <stdio.h>
#include <errno.h>

#ifdef X11
unsigned short kb_buf;
#else
unsigned char kb_buf;
#endif
unsigned char kb_flags;

local token_tabs kb_tables;
local token_tabs cm_tables;
local unsigned char *prdcd_table;
local bool kbnofirst = FALSE;
local bool kblock = FALSE;

#ifndef X11
#ifdef UNIX
local ttystruct shell_termio;
local ttystruct nm_termio;
#endif /* def UNIX */
#endif /* not def X11 */
int term_file;
local int term_flags;
bool have_tty;

#if defined( UNIX ) || defined( X11 )
#define buf_no_empty()      (kb_flags &   KBBNE)
#define set_buf_no_empty()  (kb_flags |=  KBBNE)
#define buf_empty()       (!(kb_flags &   KBBNE))
#define set_buf_empty()     (kb_flags &= ~KBBNE)
#endif /* def UNIX or X11 */

#if defined( VMS ) && !defined( X11 )
#include "kbq.c"
#define buf_no_empty()      !qkbqempty()
#define set_buf_no_empty()
#define buf_empty()          qkbqempty()
#define set_buf_empty()
#endif /* def VMS without X11 */

int
set_big_letters ()
{
  return kb_flags |= KBSBL;
}
int
set_small_letters ()
{
  return kb_flags &= ~KBSBL;
}
#define big_letters()       (kb_flags &  KBSBL )

#ifndef X11
#ifdef VMS
void
clear_tty_in ()
{
  kbqinit ();
}
#define clear_tty_in()      kbqinit()
#endif /* def VMS */

#ifdef UNIX
#ifdef MIM_SVR4
void
clear_tty_in ()
{
  tcflush (term_file, TCOFLUSH);
}
#else /* not def MIM_SVR4 */
void
clear_tty_in ()
{
  ioctl (term_file, KI_FLUSH, 0);
}
#endif /* def MIM_SVR4 */
#define set_flag_no_delay() (term_flags |=  O_NDELAY)
#define set_flag_delay()    (term_flags &= ~O_NDELAY)
#define is_no_delay()       (term_flags &   O_NDELAY)
#define is_delay()          !(term_flags &   O_NDELAY)
#endif /* not def MIM_SVR4 */
#else /* not def UNIX */
void
clear_tty_in ()
{
}
#define set_flag_no_delay()
#define set_flag_delay()
#define is_no_delay()
#define is_delay()
#endif /* def UNIX */

bool
qkbhin ()
{
  register small i;
  if (buf_no_empty () || cm_tables.buf_count || kb_tables.buf_count)
    return TRUE;
#ifndef X11
#ifdef VMS
  return FALSE;
#endif /* def VMS */

#ifdef UNIX
  if (is_delay ())
    {
      set_flag_no_delay ();
      fcntl (term_file, F_SETFL, term_flags);
    }
/* In some Unix (eg Linux) non-blocking read returns -1  if input terminal */
/* buffer is empty                                                         */
  /* 
     if ( (i = read( term_file, &kb_buf, 1 )) == 0 || i == -1 ) return FALSE;
   */
  i = read (term_file, &kb_buf, 1);

  /*+++
     fprintf(stderr, "*** i=%d, key=%o ***\n", i, kb_buf);
     +++ */

  if (i == 0 || i == (-1))
    return FALSE;

  if (i != 1 && qkbhbrk ())
    kb_buf = 003;		/* CTRL_C */
  set_buf_no_empty ();
  return TRUE;
#endif /* def UNIX */
#else /* def X11  */
  if (!xiCheckEvent ())
    return FALSE;
  kb_buf = xiKey;
  set_buf_no_empty ();
  return TRUE;
#endif /* not def X11 */
}

static previous_was_esc = 0;

local short
get_tty ()
{
#ifndef X11
#ifdef UNIX
  if (buf_empty ())
    {
      if (is_no_delay ())
	{
	  set_flag_delay ();
	  fcntl (term_file, F_SETFL, term_flags);
	}
      while (read (term_file, &kb_buf, 1) != 1)
	{
	  if (qkbhbrk ())
	    {
	      kb_buf = 003;	/* CTRL_C */
	      break;
	    }
	}
      /*+++
         fprintf(stderr, "get_tty=%o ***\r\n", kb_buf);
         +++ */
    }
  set_buf_empty ();
  return ((short) kb_buf) & 0377;
#endif /* def UNIX */

#ifdef VMS
  while (buf_empty ())
    sys$waitfr (5);
  return ((short) kbqget ()) & 0377;
#endif /* def VMS */
#else /* def X11 */
  if (buf_empty ())
    {
      xiWaitEvent ();
      /*   Alt+ */
      if (xiKey & XI_ALT_MASK)
	{
	  xiKey &= ~XI_ALT_MASK;
	  previous_was_esc = 1;
	}
      kb_buf = xiKey;
    }
  set_buf_empty ();
  return kb_buf;
#endif /* not def X11 */
}

local short
get_key ()
{
  return get_token (&kb_tables, get_tty);
}
local short
get_cmd ()
{
  return get_token (&cm_tables, get_key);
}

#ifdef X11
void
quit_handler ()
{
  if (xiLocked)
    xiUnLock (kbnofirst);
  else
    {
      xiLock ();
      kbnofirst = !kblock;
    }
}
#else /* not def X11 */
#ifdef UNIX
local void
break_handler (int sig)
{
#ifdef LINUX
  /*  Ctrl+C  ,  ,       
     Ctrl+C.  ,      (, AIX),  
           ,    . */
  if (is_delay())
    {
      set_flag_no_delay ();
      fcntl (term_file, F_SETFL, term_flags);
    }
#endif /* LINUX */
  set_break_pressed ();
  signal (SIGINT, break_handler);
}

local void
quit_handler (int sig)
{
  signal (SIGQUIT, SIG_IGN);
  if (ScLocked)
    scunlock (kbnofirst);
  else
    {
      sclock ();
      kbnofirst = !kblock;
    }
  signal (SIGQUIT, quit_handler);
}

#ifdef MIM_SVR4
local void
winch_handler (int sig)
{
  struct winsize ws;
  ioctl (term_file, TIOCGWINSZ, &ws);
  if (ws.ws_col)
    columns = ws.ws_col;
  if (ws.ws_row)
    lines_ = ws.ws_row;
  scResize (ws.ws_col, ws.ws_row);
  signal (SIGWINCH, winch_handler);
}
#endif /* def MIM_SVR4 */
#endif /* def UNIX */

#ifdef VMS
local void
asthdl (int p)
{
  if (kb_buf == 003)
    {
      set_break_pressed ();
      kbBreak = TRUE;
    }
  if (kb_buf == 032)
    {
      if (ScLocked)
	scunlock (kbnofirst);
      else
	{
	  sclock ();
	  kbnofirst = !kblock;
	}
    }
  else
    kbqput (kb_buf);
  if (have_tty)
    sys$qio (5, term_file,
	     IO$_TTYREADALL | IO$M_NOECHO,
	     0, &asthdl, 0, &kb_buf, 1, 0, 0, 0, 0);
}

static struct
{
  unsigned s_class:8;
  unsigned s_type:8;
  unsigned s_width_page:16;
  union ttdef s_main_modes;
}
shell_termio, nm_termio;
#define LENMODBUF 8
static $DESCRIPTOR (dtty, "sys$input");
static char fd0[MAXPATH];
#endif /* def VMS */
#endif /* def X11 */

#ifdef SUPPORT_XTERM
int xterm_p = 0;
#endif /* SUPPORT_XTERM */

void
term_attach ()
{
  int i;
  char file_name[MAXPATH], *n, *t;
  char buf[256];

#define otkaz( s ) { perror( s ); _exit( 1 ); }

#ifndef X11
#ifdef UNIX
  ttystruct *st = &shell_termio;
  ttystruct *mt = &nm_termio;
  term_file = 0;
#endif /* def UNIX */

#ifdef VMS
  if ((i = sys$assign (&dtty, &term_file, 0, 0)) != SS$_NORMAL)
    {
      sprintf (buf, "Open_tty fail %s", strerror (EVMSERR, i));
      otkaz (buf);
    }
#endif /* def VMS */
#endif /* not def X11 */

#ifdef SUPPORT_XTERM
  if ((n = (char *) getenv ("TERM")) && !strcmp (n, "xterm"))
    {
      /*  - xterm. */
      xterm_p = 1;
    }
#endif /* SUPPORT_XTERM */

  if (!(n = (char *) getenv (NMTERM)))
    {
      n = DEFAULT_MIMTERM;
      /* otkaz("MIMTERM undefined."); */
    }
  if (!(t = (char *) getenv (NMTERMINFO)))
    {
      t = DEFAULT_MIMTERMINFO;
      /* otkaz("MIMTERMINFO undefined."); */
    }

#ifdef  VERSION_2
  {
    int l;
    strcpy (nmlocal, t);	/* Added by VVB */
    l = strlen (nmlocal);
    if (l > 0 && nmlocal[l - 1] != '/')
      {
	strcat (nmlocal, "/");
      }
  }
#endif /* VERSION_2 */

#ifdef UNIX
#ifdef SUPPORT_XTERM
#define PREFSTR (xterm_p ? "%s/xterm%s" : "%s/%s")
#else
#define PREFSTR "%s/%s"
#endif /* SUPPORT_XTERM */
#endif /* def UNIX */

#ifdef VMS
#define PREFSTR "%s%s"
#endif /* def VMS */

  sprintf (file_name, PREFSTR, t, n);
  strcat (file_name, ".kb");
  if (load_table (file_name, &kb_tables) != 0)
#ifdef X11
    otkaz ("XMIMTERM - invalid kb");
#else
    otkaz ("MIMTERM - invalid kb");
#endif /* def X11 */
  sprintf (file_name, PREFSTR, t, n);
  strcat (file_name, ".cm");
  if (load_table (file_name, &cm_tables) != 0)
#ifdef X11
    otkaz ("XMIMTERM - invalid cm");
#else
    otkaz ("MIMTERM - invalid cm");
#endif /* def X11 */
  sprintf (file_name, PREFSTR, t, n);
  strcat (file_name, ".cd");
  if (load_codes (file_name) != 0)
#ifdef X11
    otkaz ("XMIMTERM - invalid cd");
#else
    otkaz ("MIMTERM - invalid cd");
#endif /* def X11 */
  kb_flags = 0;

#ifdef X11
#ifdef  VERSION_2
  signal (SIGINT, SIG_DFL);
  signal (SIGQUIT, SIG_DFL);
#else
  signal (SIGQUIT, quit_handler);
#endif /* VERSION_2 */
#if 0
  signal (SIGSTOP, stop_handler);
  signal (SIGCONT, cont_handler);
#endif

#else /* not def X11 */

#ifdef UNIX

#ifdef MIM_SVR4
  if (tcgetattr (term_file, mt))
    otkaz ("tcgetattr");
#else /* not def MIM_SVR4 */
  if (ioctl (term_file, KI_GET, mt))
    otkaz ("ioctl:GET");
#endif /* def MIM_SVR4 */

  *st = *mt;
  mt->c_iflag &= ~(ISTRIP | IUCLC | ICRNL);
  mt->c_iflag |= (IXOFF | IGNBRK | IGNPAR);

#if 0
  if (xon_xoff)
    mt->c_iflag |= (IXON);
  else
    mt->c_iflag &= ~(IXON);
#else
#ifdef  USE_SYSTEM_KEYBORD_DRIVER
  mt->c_iflag |= (IXON);
#else
  mt->c_iflag &= ~(IXON);
#endif /* USE_SYSTEM_KEYBORD_DRIVER */
#endif /* if 0 */

  mt->c_oflag &= ~(OPOST);
  mt->c_lflag &= ~(ICANON | ECHO | ECHOE | ECHONL | ONOCR);
  mt->c_lflag |= (ISIG);
  mt->c_cc[VINTR] = '\003';
  mt->c_cc[VQUIT] = '\032';

#ifdef SVR3
  mt->c_cc[VSWTCH] = '\377';
#endif

#ifdef MIM_SVR4
  mt->c_cc[VSUSP] = '\000';
  mt->c_cc[VDSUSP] = '\000';
#endif /* def MIM_SVR4 */

  mt->c_cc[VMIN] = 1;
  mt->c_cc[VTIME] = 0;

#ifdef MIM_SVR4
  if (tcsetattr (term_file, TCSADRAIN, mt))
    otkaz ("tcsetattr");
#else /* not def MIM_SVR4 */
  if (ioctl (term_file, KI_SET, mt))
    otkaz ("ioctl:SET");
#endif /* def MIM_SVR4 */

#ifdef MIM_SVR4
#ifndef VERSION_2
  if (!alphanumeric_term)
#endif /* ! VERSION_2 */
    {
      struct winsize ws;
      ioctl (term_file, TIOCGWINSZ, &ws);
      if (ws.ws_col)
	columns = ws.ws_col;
      if (ws.ws_row)
	lines_ = ws.ws_row;
      signal (SIGWINCH, winch_handler);
    }
#endif /* def MIM_SVR4 */

  term_flags = fcntl (term_file, F_GETFL);
  term_flags |= O_SYNC;
  fcntl (term_file, F_SETFL, term_flags);
  signal (SIGINT, break_handler);
  signal (SIGQUIT, quit_handler);
#endif /* def UNIX */

#ifdef VMS
  if (sys$qiow
      (0, term_file, IO$_SENSEMODE, 0, 0, 0, &shell_termio, LENMODBUF, 0, 0,
       0, 0) != SS$_NORMAL)
    otkaz ("get terminal description fail");
  nm_termio = shell_termio;
  nm_termio.s_main_modes.tt$v_ttsync = 0;
  if (sys$qiow
      (0, term_file, IO$_SETMODE, 0, 0, 0, &nm_termio, LENMODBUF, 0, 0, 0,
       0) != SS$_NORMAL)
    otkaz ("set terminal description fail");
#endif /* def VMS  */
#endif /* def X11 */

  prdcd_table = kb_lat_ASCII;
  stCCaps (STC_SLAT);
#ifndef X11
  tputs (screen_attach, 1, tv_write_one);
#ifdef VMS

  kbqinit ();
  sys$qio (5, term_file,
	   IO$_TTYREADALL | IO$M_NOECHO | IO$M_PURGE,
	   0, &asthdl, 0, &kb_buf, 1, 0, 0, 0, 0);
#endif /* def VMS  */
#endif /* def X11 */
#undef otkaz
}

void
kbDetach ()
{
  int s;
  if (have_tty)
    {
#ifdef X11
      if (xiLocked)
	xiUnLock (FALSE);
      signal (SIGINT, SIG_IGN);
      signal (SIGQUIT, SIG_IGN);
#else
      if (ScLocked)
	scunlock (FALSE);
      tputs (exit_alt_charset_mode, 1, tv_write_one);
      tputs (exit_graph_mode, 1, tv_write_one);
      tputs (exit_attribute_mode, 1, tv_write_one);
      tputs (screen_detach, 1, tv_write_one);
      tv_flush ();

#ifdef UNIX

#ifdef MIM_SVR4
      tcsetattr (term_file, TCSADRAIN, &shell_termio);
#else /* not def MIM_SVR4 */
      ioctl (term_file, KI_SET, &shell_termio);
#endif /* def MIM_SVR4 */

      signal (SIGINT, SIG_IGN);
      signal (SIGQUIT, SIG_IGN);
#ifdef MIM_SVR4
      signal (SIGWINCH, SIG_DFL);
#endif /* def MIM_SVR4 */
#endif /* def UNIX */

#ifdef VMS
      s = sys$dassgn (term_file);
      if (sys$assign (&dtty, &term_file, 0, 0) != SS$_NORMAL)
	otkaz ("Open_tty fail");
      sys$qiow (5, term_file, IO$_SETMODE, 0, 0, 0, &shell_termio, LENMODBUF,
		0, 0, 0, 0);
#endif /* def VMS */
#endif /* def X11 */

      have_tty = FALSE;
    }
}

void
kbAttach ()
{
  if (!have_tty)
    {
#ifdef X11
      signal (SIGINT, SIG_DFL);
      signal (SIGQUIT, SIG_DFL);

#else

#ifdef UNIX
      signal (SIGINT, break_handler);
      signal (SIGQUIT, quit_handler);

#ifdef MIM_SVR4
#ifndef VERSION_2
      if (!alphanumeric_term)
#endif /* ! VERSION_2 */
	{
	  struct winsize ws;
	  ioctl (term_file, TIOCGWINSZ, &ws);
	  if (ws.ws_col)
	    columns = ws.ws_col;
	  if (ws.ws_row)
	    lines_ = ws.ws_row;
	  scResize ();
	  signal (SIGWINCH, winch_handler);
	}
#endif /* def MIM_SVR4 */

#ifdef MIM_SVR4
      tcsetattr (term_file, TCSADRAIN, &nm_termio);
#else /* not def MIM_SVR4 */
      ioctl (term_file, KI_SET, &nm_termio);
#endif /* def MIM_SVR4 */

      kb_flags = 0;
#endif /* def UNIX */

#ifdef VMS
      kbqinit ();
      sys$qio (5, term_file,
	       IO$_TTYREADALL | IO$M_NOECHO | IO$M_PURGE,
	       0, &asthdl, 0, &kb_buf, 1, 0, 0, 0, 0);
#endif /* def VMS */
      tputs (screen_attach, 1, tv_write_one);
#endif /* def X11 */
      have_tty = TRUE;
      kbBreak = FALSE;
    }
}

kchar
prdcd ()
{
  register kchar c, ci;
  kblock = TRUE;
  for (;;)
    {
      if (kbnofirst)
#ifdef X11
	if (xiLocked)
	  xiUnLock (FALSE);
#else /* not def X11 */
	if (ScLocked)
	  scunlock (FALSE);
#endif /* def X11 */
      ci = (kchar) get_cmd ();

      /*+++
         printf("get_cmd()=%o \r\n", ci);
         +++ */

      if (ci < CM_MIN)
	{
#ifdef SUPPORT_XTERM
	  if (xterm_p)
	    {
	      if (ci > 0200 && ci < 0400)
		{
		  /*  Alt  xterm */
		  ci -= 0200;
		  if (prdcd_table == kb_rus_ASCII)
		    ci = kb_lat_ASCII[ci];
		  else		/* if (prdcd_table == kb_lat_ASCII) */
		    ci = kb_rus_ASCII[ci];
		}
	      else
		ci = prdcd_table[ci];
	    }
	  else
#endif /* SUPPORT_XTERM */
	  if (previous_was_esc)
	    {
#ifndef X11
	      /*   */
	      if (ci == '`')
		ci = '[';
#endif /* X11 */
	      if (prdcd_table == kb_rus_ASCII)
		ci = kb_lat_ASCII[ci];
	      else		/* if (prdcd_table == kb_lat_ASCII) */
		ci = kb_rus_ASCII[ci];
	    }
	  else
	    ci = prdcd_table[ci];
	  if (big_letters () && (ci & 0100))
	    {
	      if (ci & 0200)
		ci ^= 040;
	      else
		{
		  c = (ci & 0337);
		  if (c >= 0101 && c <= 0132)
		    ci ^= 040;
		}
	    }
	  if (kbnofirst)
#ifdef X11
	    if (xiLocked)
	      xiUnLock (FALSE);
#else
	    if (ScLocked)
	      scunlock (FALSE);
#endif
	  kbnofirst = TRUE;
	  kblock = FALSE;
	  previous_was_esc = 0;
	  return ci;
	}
      switch (ci)
	{
	case TK_BRUS:
	  set_big_letters ();
	  prdcd_table = kb_rus_ASCII;
	  stCCaps (STC_BRUS);
	  break;
	case TK_SRUS:
	  set_small_letters ();
	  prdcd_table = kb_rus_ASCII;
	  stCCaps (STC_SRUS);
	  break;
	case TK_BLAT:
	  set_big_letters ();
	  prdcd_table = kb_lat_ASCII;
	  stCCaps (STC_BLAT);
	  break;
	case TK_SLAT:
	  set_small_letters ();
	  prdcd_table = kb_lat_ASCII;
	  stCCaps (STC_SLAT);
	  break;
	default:
	  if (kbnofirst)
#ifdef X11
	    if (xiLocked)
	      xiUnLock (FALSE);
#else
	    if (ScLocked)
	      scunlock (FALSE);
#endif
	  kbnofirst = TRUE;
	  kblock = FALSE;
	  previous_was_esc = (ci == TK_ESC);
	  return ci;
	}
    }
}
