/*
 *     nanoMir
 *
 *     "rt.c"           "Run time library"
 */

#include "nm.h"
#include <stdio.h>
#ifdef UNIX
#include <sys/types.h>
#endif
#ifdef VMS
#include <types.h>
#include <descrip.h>
#endif
#include <time.h>

local char *ltod (char *, large);

char *
sncpy (register char *from, register char *to, register small n)
{
  if (n)
    {
      do
	{
	  *to++ = *from++;
	}
      while (--n);
    }
  return (to);
}

char *
scpy (register char *from, register char *to)
{
  register char c;
  while (c = *from++)
    *to++ = c;
  return (to);
}

bool
scmp (register char *p1, register char *p2)
{
  register small c;
  while ((c = *p1++) == (*p2++))
    if (c == 0)
      return (TRUE);
  return (FALSE);
}

bool
sncmp (register char *p1, register char *p2, register small n)
{
  while (n--)
    if (*p1++ != *p2++)
      return FALSE;
  return TRUE;
}

small
slen (register char *p)
{
  register char *p1;
  for (p1 = p; *p1++;);
  return ((small) (p1 - p) - 1);
}

char
stoupper (register char c)
{
  if (c >= 'a' && c <= 'z')
    c -= 040;
  return (c);
}

char *
ltox (register char *p, register large l)
{
  register char *p1;
  register small i, j;
  char lbuf[9];
  p1 = lbuf + 9;
  *--p1 = 0;
  i = sizeof (large) * 2;
  do
    {
      if ((j = l & 017) <= 9)
	j += '0';
      else
	j += ('A' - 10);
      *--p1 = j;
      if ((l >>= 4) == 0)
	break;
    }
  while (--i);
  while (j = *p1++)
    *p++ = j;
  return (p);
}

char *
xtol (register char *p, large * pl)
{
  register large l;
  register small c;
  l = 0;
  while (1)
    {
      c = *p++;
      if ((c >= '0') && (c <= '9'))
	c = c - '0';
      else if ((c >= 'A') && (c <= 'F'))
	c = c - ('A' - 10);
      else
	break;
      l = (l << 4) + c;
    }
  *pl = l;
  return (p);
}

static char *
twodec (register char *p, register small s)
{
  *p++ = s / 10 + '0';
  *p++ = s % 10 + '0';
  return (p);
}

char *
timtoa (register char *p, register small t)
{
  p = twodec (p, (t >> 11) & 037);
  *p++ = ':';
  p = twodec (p, (t >> 5) & 077);
  *p++ = ':';
  p = twodec (p, (t << 1) & 077);
  return (p);
}

static char *mounths[13] = {
  " ", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

char *
dattoa (register char *p, register small d)
{
  p = twodec (p, d & 037);
  *p++ = '-';
  p = scpy (mounths[(d >> 5) & 017], p);
  *p++ = '-';
  p = stod (p, ((d >> 9) & 0177) + 1980);
  return (p);
}

char *
stod (char *p, small s)
{
  return (ltod (p, (large) s));
}

local char *
ltod (register char *p, register large l)
{
  register char *p1, c;
  register small i;
  char lbuf[12];
  if (l < 0)
    {
      l = -l;
      *p++ = '-';
    }
  p1 = lbuf + 12;
  *--p1 = 0;
  i = 11;
  do
    {
      *--p1 = l % 10 + '0';
      if ((l = l / 10) == 0)
	break;
    }
  while (--i);
  while (c = *p1++)
    *p++ = c;
  return (p);
}

char *
dtos (register char *p, small * s)
{
  large l;
  p = dtol (p, &l);
  *s = l;
  return (p);
}

char *
dtol (register char *p, large * lp)
{
  register large l;
  register small i, s;

  s = 0;
  if (*p == '-')
    {
      p++;
      s++;
    }
  l = 0;
  while ((i = *p++ - '0') >= 0 && i < 10)
    l = l * 10 + i;
  if (s)
    l = -l;
  *lp = l;
  return (--p);
}

void
br14 ()
{
}

long
atodate (register char *s)
{
  large y, m, d, h, mn;
  s = dtol (s, &d);
  s = dtol (++s, &m);
  s = dtol (++s, &y);
  s = dtol (++s, &h);
  s = dtol (++s, &mn);
  return (((y - 80) << 25) | (m << 21) | (d << 16) | (h << 11) | (mn << 5));
}

char *
sindex (register char *s, register char c)
{
  register char c1;
  while (c1 = *s++)
    if (c1 == c)
      return (--s);
  return (NIL);
}

char *
attrtoa (register char *p, register small attr)
{
  register small mask;
  register char *pm;

  switch (attr & 070000)
    {
    case 000000:
      *p++ = '-';
      break;
    case 010000:
      *p++ = 'p';
      break;
    case 020000:
      *p++ = 'c';
      break;
    case 040000:
      *p++ = 'd';
      break;
    case 060000:
      *p++ = 'b';
      break;
    default:
      *p++ = '?';
      break;
    }
  if (attr & 04000)
    *p++ = 's';
  else
    *p++ = '-';
  if (attr & 02000)
    if (attr & 010)
      *p++ = 's';
    else
      *p++ = 'l';
  else
    *p++ = '-';
  if (attr & 01000)
    *p++ = 't';
  else
    *p++ = '-';

  for (mask = 0400, pm = MASK; mask; mask >>= 1, p++, pm++)
    if (attr & mask)
      *p = *pm;
    else
      *p = '-';
  return (p);
}

char *
lblkmov (register char *from, register char *to, register addr len)
{
  register small m;
  register char *ret;
  if (len <= 0)
    return (to);
  ret = to + len;
#ifdef VMS
  memcpy (to, from, len);
#endif
#ifdef UNIX
#ifdef MIM_SVR4
  memmove (to, from, len);
#endif
#ifdef SVR3
  if (len > 0)
    {
      if (from < to)
	{
	  from += len;
	  to += len;
	  for (; len >= 4; len -= 4)
	    *--((addr *) to) = *--((addr *) from);
	  for (; len > 0; len--)
	    *--(to) = *--(from);
	}
      else
	{
	  for (; len >= 4; len -= 4)
	    *((addr *) to)++ = *((addr *) from)++;
	  for (; len > 0; len--)
	    *(to)++ = *(from)++;
	}
    }
#endif
#endif
  return (ret);
}

char *
blkmov (char *from, char *to, small len)
{
  return (lblkmov (from, to, (addr) len));
}

static char chx[80];
char *
dttm (time_t t)
{
  register char *p;
  char year[5];
  char date[3];
  char month[4];
  char time[9];
  p = ctime (&t) + 4;
  *sncpy (p, month, 3) = 0;
  *sncpy (p + 4, date, 2) = 0;
  *sncpy (p + 7, time, 8) = 0;
  *sncpy (p + 16, year, 4) = 0;
  sprintf (chx, "%s-%s-%s  %s", date, month, year, time);
  return (chx);
}

time_t
GetTime ()
{
  return time (0);
}
