/*
 *     nanoMir
 *
 *     "tm.c"           Text & Window manager
 */

/*
#define TMDEBUG
*/
#include "nm.h"
#include "fs.h"
#include "dq.h"
#include "tx.h"
#include "me.h"
#include "wd.h"
#include "kb.h"
#ifndef X11
#include "sc.h"
#else
#include "xi.h"
#endif
#include "le.h"
#include "te.h"
#include "ud.h"
#include "cm.h"
#include "cmn.h"
#include "tm.h"
#include "cp.h"
#include "er.h"
#include "ac.h"
#include <stdio.h>
#include <errno.h>
#include <string.h>

#define MINWWIN 4
#define MINHWIN 5

bool tmNoText = TRUE;

extern char *getenv (char *);
local small filrm;
local bool wmedin (wnd *, char *);
local void wattach (txt *, wnd *);
local void wdetach (txt *, wnd *);
local void wndcopy (wnd *, wnd *);
local void setwindow (register wnd * w);
local char *fenter (void), *f2enter (void);
local void sysenter ();
local void checkfiles (void);
local void tmdirlst (txt *);
local void infilnam (void);
local bool tm_modfile (void);
local bool tmChAttr (char *p, int attr, char *modfile_buf);
local char fnbuf[MAXLPAC];
#ifdef RUSSIAN
local char fprompt[] = ">";
#define lfprompt 5
#endif
#ifdef ENGLISH
local char fprompt[] = "File>";
#define lfprompt 5
#endif
local tchar f2buf[MAXLPAC];
local small f2len;

void
tmInit ()			/* TM.Initialize               */
{
  tchar *pt;
  char *p;
  pt = f2buf;
  for (p = fprompt; (*pt++) = ctotc (*p++););
  f2len = lfprompt;
  tmNoText = TRUE;
}

bool
wmInit (char *filnam)		/* WM.Initialize( <file> )     */
{
  register wnd *w;
#ifndef X11
  w = newWindow (0, 0, scx, scy, WND_NORMAL, 0);
#else
  w = newWindow (xiStdX (), xiStdY (),
		 xiStdWidth (), xiStdHeight (), WND_NORMAL, 0);
  w->wxstat &= ~WDX_REDRAW;
#endif
  if (w == NIL)
    otkaz ("No windows");
  Tx = Ty = 0;
  Twnd = w;
  filrm = 0;

  if (!wmedin (w, filnam))
    return FALSE;
  stCFile (Twnd->wtext->finame);
  stCLines (Twnd->wtext->flines);
  stModification (STM_NO);
  if (Lrins)
    stCMode (STM_INSERT);
  else
    stCMode (STM_REPLACE);
  stCAttr (Lattr);
  stCZoom (STZ_OFF);
  tmNoText = FALSE;
  return (TRUE);
}
/*
 * Create window ( <left up point>, <width>, <height>, <window type>, <window number> )
 */
wnd *
newWindow (int sx, int sy, int sw, int sh, small typ, small num)
{
  register wnd *w;

  WdShowboxes ();		/* .            */
  if ((w = WdNew (typ, num)) == NIL)
    {
      WdShowboxes ();
      return NIL;
    }
  w->wsx = sx;
  w->wsy = sy;
  w->wsh = sh;
  w->wsw = sw;
  w->winf = TxInfo;
  w->dirsp = w->stack;
  w->wty = w->wtx = 0;
  w->wtext = NIL;
#ifdef X11
  xiCreateWindow (w);
  if (w->wxmain == NIL)
    return NIL;
#endif
  return (w);
}

void
wattach (register txt * t, register wnd * w)
{
  if (BlockMark)
    cpblockoff ();
  w->wnext = (wnd *) t->txwndptr;
  w->wtext = t;
  w->wstat &= ~WS_NOTEXT;
  t->txwndptr = (void *) w;
  t->txstat |= TS_WND;
  w->wcx = t->cx;
  w->wcy = t->cy;
  w->wtx = t->tcx;
  w->wty = t->tcy;
}

local void
wdetach (register txt * t, register wnd * w)
{
  register wnd *w1;
#ifdef X11
  /*        */
  xiOnDetachWnd (w);
#endif /* X11 */
  if (w->wstat & WS_ERROR)
    return;
  t->txstat &= ~TS_WND;
  if (t->txwndptr == NIL)
    return;
  if (t->txwndptr == (void *) w)
    t->txwndptr = (void *) w->wnext;
  else
    {
      for (w1 = (wnd *) t->txwndptr; w1->wnext != w; w1 = w1->wnext)
	if (w1->wnext == NIL)
	  return;
      w1->wnext = w->wnext;
    }
  t->cx = Tx;
  t->cy = Ty;
  t->tcx = w->wtx;
  t->tcy = w->wty;
  w->wnext = NIL;
}

int tmemptyi = 0;
void
tmempty ()
{
  tmemptyi /= KbCount;
}

#ifndef X11
local bool tmZoomAlloc = FALSE;
bool
tmQZoom ()
{
  return tmZoomAlloc;
}

local small zwsx, zwsw;
small zwsy, zwsh;
local void
tmCreateZoom ()
{
  register wnd *w = Twnd;
  register wnd *w1;
  register small i = 0;

  if (Twnd->wstat & WS_ERROR)
    {
      scbell ();
      return;
    }
  if (tmQZoom ())
    return;
  if (erOpen)
    i++;
  if (stOpen)
    i += NSTLINE;
  if (erOpen || stOpen)
    i++;
  if (w->wsx == 0 && w->wsy == i && w->wsw == scx && w->wsh == scy - i)
    return;
  zwsx = w->wsx;
  w->wsx = 0;
  zwsy = w->wsy;
  w->wsy = i;
  zwsw = w->wsw;
  w->wsw = scx;
  zwsh = w->wsh;
  w->wsh = scy - i;
  for (i = MAXWNDWS - 2, w1 = windows; i; i--, w1++)
    if ((w1->wstat & WS_BUSY) && !(w1->wstat & WS_DEBUG) && w1 != w)
      w1->wstat |= WS_SHADOW;
  wredraw (w);
  stCZoom (STZ_ON);
  tmZoomAlloc = TRUE;
}

local void
tmDestroyZoom ()
{
  register wnd *w = Twnd;
  register wnd *w1;
  register small i;

  if (Twnd->wstat & WS_ERROR)
    {
      scbell ();
      return;
    }
  if (!tmQZoom ())
    return;
  w->wsx = zwsx;
  w->wsy = zwsy;
  w->wsw = zwsw;
  w->wsh = zwsh;
  for (i = MAXWNDWS - 2, w1 = windows; i; i--, w1++)
    if ((w1->wstat & WS_BUSY) && !(w1->wstat & WS_DEBUG) && w != w1)
      {
	w1->wstat &= ~WS_SHADOW;
	wredraw (w1);
      }
  if (w->wtext != NIL)
    wredraw (w);
  WdShowboxes ();
  stCZoom (STZ_OFF);
  tmZoomAlloc = FALSE;
}
#endif

void
wmEdit ()
{
  register txt *t;
  register small i, j;
  register char *p;
  register wnd *wind, *w;
  jmp_buf tmenv;
  bool stWasOpen;

  excptr = (jmp_buf *) tmenv;
  if (setjmp (tmenv))
    goto bell;
  for (;;)
    {
      t = Ttxt;
      if (TExec (RQ_WAIT) != E_NOCOM)
	goto bell;
      else
	{
	  /*+++
	     printf("wmEdit: kbCode = 0%o\n", KbCode);
	     +++ */

	  switch (KbCode)
	    {
	    case TE_UPDATE:
	      tmSaveAll ();
	      break;
	    case TE_EXIT:
	      if (t != LCtxt && t != Btxt)
		if (!tmsave (t, FALSE))
		  goto bell;
	      goto pop;
	    case TE_EXOLD:
	      if ((t->txstat & TS_CHANGE) && (t != LCtxt) && (t != Btxt))
		{
		  DqEmpt (t->txustk);
		  DqEmpt (t->txdstk);
		  t->txy = 0;
		  t->txudcptr = t->txudlptr = 0;
		  if (t->txudeq != NIL)
		    DqEmpt (t->txudeq);
		  t->txstat &= ~(TS_UNDO | TS_FILE | TS_CHANGE);
		}
	      goto pop;
	    case TE_EXBACK:
	      if (t != LCtxt && t != Btxt)
		if (!tmsave (t, TRUE))
		  goto bell;
	    pop:
#ifdef ERRVIEW
	      erHflag = FALSE;
#endif
	      dspop ();
	      stCFile (Twnd->wtext->finame);
	      stCLines (Twnd->wtext->flines);
	      break;
	    case TE_QUIT:
#ifndef X11
	      if (tmQZoom ())
		tmDestroyZoom ();
#endif
	      wdetach (t, Twnd);
	      t->txstat &= ~TS_PERM;
	      wmexit ();
	      stCFile (Twnd->wtext->finame);
	      stCLines (Twnd->wtext->flines);
	      break;
	    case TE_CRASH:
	      cpsave ();
#ifdef SPELL
	      erTerm ();
#endif
#ifdef X11
	      stufTerm ();
	      xiDetach ();
#else
	      scErase ();
	      scDetach ();
#endif
	      kbDetach ();
	      exit (0);
	    case TE_HELP:	/* Help, I need somebody... */
	      /* p=scpy( nmlocal, fnbuf ); */
	      if (!(p = (char *) getenv (NMTERMINFO)))
		p = DEFAULT_MIMTERMINFO;
	      p = scpy (p, fnbuf);

#ifdef UNIX
#ifdef SUPPORT_XTERM
	      {
		extern int xterm_p;
		p = scpy ((xterm_p ? "/xterm" : "/"), p);
	      }
#else
	      p = scpy ("/", p);
#endif /* SUPPORT_XTERM */
#endif
	      {
		char *tmp = getenv (NMTERM);
		if (!tmp)
		  tmp = DEFAULT_MIMTERM;
		p = scpy (tmp, p);
	      }
	      *(p = scpy (".hlp", p)) = 0;
	      p = fnbuf;
	      goto push;
	    case TE_MSGS:
	      p = scpy (nmlocal, fnbuf);
	      *(p = scpy (MSGFILE, p)) = 0;
	      p = fnbuf;
	      goto push;
	    case TE_FENTR:
	      p = fenter ();
	      goto push;
	    case TE_F2ENTR:
	      p = f2enter ();
	    push:if (Twnd->wstat & WS_ERROR || p == NIL)
		goto bell;
	      if (tmcheck (p) == NIL)
		{
		  ErFerr (p);
		  goto bell;
		}
#ifdef ERRVIEW
	      erHflag = FALSE;
#endif
	      dspush (p);
	      stCFile (Twnd->wtext->finame);
	      stCLines (Twnd->wtext->flines);
	      break;
	    case TE_EXEC:
#ifdef NM
	      if ((p = Ttxt->finame) == NIL)
		goto bell;
	      goto exec;
#endif
#ifdef C93
	      cdCompile ();
	      break;
#endif
	    case TE_FEXEC:
	      if ((p = fenter ()) == NIL)
		goto bell;
	      if ((p = tmcheck (p)) == NIL)
		goto bell;
	    exec:
#ifdef ERRVIEW
	      erLoaded = FALSE;
#endif
	      tmSaveAll ();
	      if (p == NIL)
		blkexec ();
	      else
		fileexec (p);
	      blkxdir = 0;
	      checkfiles ();
	      WdShowboxes ();
#ifndef X11
	      wind = Twnd;
	      for (i = MAXWNDWS, w = windows; --i >= 0; w++)
		if (((j = w->wstat) & WS_BUSY) && w != wind)
		  {
		    if (!(j & WS_DEBUG))
		      setwindow (w);
		    wredraw (w);
		  }
	      setwindow (wind);
	      wredraw (wind);
#endif
	      break;
	    case TE_SHELL:
	      p = "";
	      goto exec;
	    case TE_BLKEXEC:
	      if (Btxt == NIL)
		goto bell;
	      p = NIL;
	      goto exec;
#ifdef ERRVIEW
	    case TE_PERROR:
	      p = erPrev ();
	      goto showerr;
	    case TE_NERROR:
	      p = erNext ();
#endif
	    showerr:
	      ErShow (p);
	      break;
	    case TE_ERROFF:
	      ErOff ();
	      break;
	    case TE_STATUS:
	      if (stOpen)
		stOff ();
	      else
		stOn ();
	      break;
#ifndef X11
	    case TE_DFORK:
	      if (Twnd->wstat & WS_ERROR || tmQZoom ())
		goto bell;
	      if ((i = Twnd->wsh / 2) < MINHWIN)
		goto bell;
	      i++;
	      wind =
		newWindow (Twnd->wsx, Twnd->wsy + i, Twnd->wsw, Twnd->wsh - i,
			   WND_NORMAL, 0);
	      if (wind == NIL)
		goto bell;
	      w_clear (Twnd);
	      Twnd->wsh = i - 1;
	      goto fork;
	    case TE_UFORK:
	      if (Twnd->wstat & WS_ERROR || tmQZoom ())
		goto bell;
	      if ((i = Twnd->wsh / 2) < MINHWIN)
		goto bell;
	      wind =
		newWindow (Twnd->wsx, Twnd->wsy, Twnd->wsw, i, WND_NORMAL, 0);
	      if (wind == NIL)
		goto bell;
	      w_clear (Twnd);
	      i++;
	      Twnd->wsy += i;
	      Twnd->wsh -= i;
	      goto fork;
	    case TE_RFORK:
	      if (Twnd->wstat & WS_ERROR || tmQZoom ())
		goto bell;
	      if ((i = Twnd->wsw / 2) < MINWWIN)
		goto bell;
	      i++;
	      wind =
		newWindow (Twnd->wsx + i, Twnd->wsy, Twnd->wsw - i, Twnd->wsh,
			   WND_NORMAL, 0);
	      if (wind == NIL)
		goto bell;
	      w_clear (Twnd);
	      Twnd->wsw = i - 1;
	      goto fork;
	    case TE_LFORK:
	      if (Twnd->wstat & WS_ERROR || tmQZoom ())
		goto bell;
	      if ((i = Twnd->wsw / 2) < MINWWIN)
		goto bell;
	      wind =
		newWindow (Twnd->wsx, Twnd->wsy, i, Twnd->wsh, WND_NORMAL, 0);
	      if (wind == NIL)
		goto bell;
	      w_clear (Twnd);
	      i++;
	      Twnd->wsx += i;
	      Twnd->wsw -= i;
	    fork:
	      WdShowboxes ();
	      wndcopy (wind, Twnd);
	      wredraw (Twnd);
	      setwindow (wind);
	      wattach (Ttxt, Twnd);
	      wredraw (Twnd);
	      break;
#else
	    case TE_FORK:
	      {
		wnd *w;
		if (Twnd->wstat & WS_ERROR)
		  goto bell;
		wind =
		  newWindow (xiStdX (), xiStdY (), xiStdWidth (),
			     xiStdHeight (), WND_NORMAL, 0);
		if (wind == NIL)
		  goto bell;
		wndcopy (wind, Twnd);
		w = Twnd;
		setwindow (wind);
		wattach (Ttxt, Twnd);
		Twnd = w;
		wpos (wind, wind->wcx, wind->wcy);
		/*... break; ... */
		goto LSetFocus;	/* VVB */
	      }
#endif
#ifndef X11
	    case TE_ZOOM:
	      if (tmQZoom ())
		tmDestroyZoom ();
	      else
		tmCreateZoom ();
	      break;
	    case TE_UWINDOW:
	      i = 0;
	      j = (-1);
	      goto gwindow;
	    case TE_DWINDOW:
	      i = 0;
	      j = 1;
	      goto gwindow;
	    case TE_RWINDOW:
	      i = 1;
	      j = 0;
	      goto gwindow;
	    case TE_LWINDOW:
	      i = (-1);
	      j = 0;
	    gwindow:
	      wind =
		WdNear (Twnd, Tx - Twnd->wtx + Twnd->wsx,
			Ty - Twnd->wty + Twnd->wsy, i, j);
	      if (wind == NIL)
		goto bell;
#else
	    case TE_NWINDOW:
	      /*     */
	      xiNextWindow (Twnd);
	      wind = xiTWnd;
	      goto LSetFocus;

	    case TE_PWINDOW:
	      /*     */
	      xiPrevWindow (Twnd);
	      wind = xiTWnd;
	      goto LSetFocus;

	    case TE_FOCUS:

	      /*+++
	         printf("tm.c: SetFocus (key = TE_FOCUS)\n");
	         +++ */

	      wind = xiTWnd;
	      if (xiLocked)
		xiUnLock (TRUE);
#endif
	    LSetFocus:;

	      setwindow (wind);
	      wpos (Twnd, Tx, Ty);

#if defined (X11) && defined (STATUS_IN_MAIN_WINDOW)
	      linkStatusBarAndCurrentWnd ();
#endif /* X11 && STATUS_IN_MAIN_WINDOW */

	      stCFile (Twnd->wtext->finame);
	      stCLines (Twnd->wtext->flines);
	      break;
	    case TD_SDEQ:
	      tmempty ();
	      break;
#ifdef TMDEBUG
	    case TD_STXT:
	      for (i = MAXFILES, t = texts; i; t++, i--)
		if (t->txstat)
		  dbprintf ("%16s %o\n", t->finame, t->txstat);
	      break;
#endif
	    case MC_SAVE:
	      p = fenter ();
	      KbMSave (p);
	      break;
	    case MC_2SAVE:
	      p = f2enter ();
	      KbMSave (p);
	      break;
	    case MC_LOAD:
	      p = fenter ();
	      KbMLoad (p);
	      break;
	    case MC_2LOAD:
	      p = f2enter ();
	      KbMLoad (p);
	      break;
	    case TE_MODFILE:
	      if (!tm_modfile ())
		goto bell;
	      break;
	    case TE_INFILE:	/* enter filename */
	      infilnam ();
	      break;
	    case TK_BREAK:
	      if (BlockMark)
		cpblockoff ();
	    case TE_OUTFILE:	/* leave filename */
	    default:
	    bell:
	      scbell ();
	      break;
	    }
	}
    }
}
local void
wndcopy (register wnd * t, register wnd * f)
{
  register txt *tt;
  tt = Ttxt;
  t->wtext = f->wtext;
  t->dirsp = t->stack + (f->dirsp - f->stack);
  tt->tcy = t->wty = f->wty;
  tt->tcx = t->wtx = f->wtx;
  t->wcy = Ty;
  t->wcx = Tx;
  blkmov ((char *) (f->stack), (char *) (t->stack),
	  sizeof (dirstk) * MAXDIRSTK);
}

void
dspop ()
{
  register dirstk *dp;
  register wnd *w;
  w = Twnd;
  wdetach (Ttxt, w);
  Ttxt->txstat &= ~TS_PERM;
next:
  if ((dp = w->dirsp) > w->stack)
    {
      --dp;
      w->wtx = dp->dswx;
      w->wty = dp->dswy;
      Tx = dp->dsx;
      Ty = dp->dsy;
      w->dirsp = dp;
      if (!wmedin (w, dp->dsname))
	goto next;
      stModification ((Twnd->wtext->txstat & TS_CHANGE) ? STM_YES : STM_NO);
    }
  else
    wmexit ();
}

void
dspush (char *filnam)
{
  register dirstk *dp;
  register wnd *w;
  w = Twnd;
  wdetach (Ttxt, w);
  Ttxt->txstat &= ~TS_PERM;
  dp = w->dirsp;
  if (dp < w->stack + MAXDIRSTK)
    {
      dp->dswx = w->wtx;
      dp->dswy = w->wty;
      dp->dsx = Tx;
      dp->dsy = Ty;
      blkmov (Ttxt->finame, dp->dsname, MAXPATH);
      w->dirsp++;
      w->wtx = 0;
      w->wty = 0;
      if (!wmedin (Twnd, filnam))
	dspop ();
      stModification ((Twnd->wtext->txstat & TS_CHANGE) ? STM_YES : STM_NO);
      if (filrm)
	{
	  Ttxt->txrm = filrm;
	  filrm = 0;
	}
    }
  else
    scbell ();
}

local bool
wmedin (register wnd * w, char *filnam)
{
  register txt *t;
  register bool q;
  if ((t = tmdesc (filnam, TRUE)) == NIL)
    {
      ErFerr (filnam);
      return (FALSE);
    }
/*ScJustMin( TRUE );    */
  q = tmload (t);
/*ScJustMin( FALSE );   */
  if (!q)
    {
      ErFerr (t->finame);
      return (FALSE);
    }
  Ttxt = t;
  Tx = t->cx;
  Ty = t->cy;
  wattach (t, w);
  wredraw (w);
  return (TRUE);
}

void
tmSaveAll ()
{
  register small i;
  register txt *t;
/*ScJustMin( TRUE );    */
  for (i = MAXFILES, t = texts; --i >= 0; t++)
    if (t->txstat & TS_CHANGE)
      tmsave (t, FALSE);
/*ScJustMin( FALSE );   */
}

/*      */
void
tmSaveBackAll ()
{
  register small i;
  register txt *t;
  for (i = MAXFILES, t = texts; --i >= 0; t++)
    if (t->txstat & TS_CHANGE)
      tmsave (t, TRUE);
}

/*         */
void
tmSaveBackExit ()
{
  register dirstk *dp;
  register wnd *w;

  if (Ttxt->txstat & TS_CHANGE)
    tmsave (Ttxt, TRUE);

  w = Twnd;
  wdetach (Ttxt, w);
  Ttxt->txstat &= ~TS_PERM;

  /*     */
  while (w->dirsp > w->stack)
    {
      w->dirsp --;
      w->wtx = w->dirsp->dswx;
      w->wty = w->dirsp->dswy;
      Tx = w->dirsp->dsx;
      Ty = w->dirsp->dsy;
      if ((Ttxt = tmdesc (w->dirsp->dsname, TRUE)) == NIL
	  || !tmload (Ttxt))
	continue;
      if (Ttxt->txstat & TS_CHANGE)
	tmsave (Ttxt, TRUE);
      Ttxt->txstat &= ~TS_PERM;
      Tx = Ttxt->cx;
      Ty = Ttxt->cy;
    }
  wmexit ();
}

local void
checkfiles ()
{
  register small i, s;
  register txt *t;
  for (i = MAXFILES, t = texts; i--; t++)
    if (t != Atxt)
      {
	s = t->txstat;
	if ((s & TS_BUSY) && !FsCheck (t))
	  {
	    if (s & TS_FILE)
	      {
		DqEmpt (t->txustk);
		DqEmpt (t->txdstk);
		t->txy = 0;
		t->txstat &= ~TS_FILE;
	      }
	    if (s & TS_UNDO)
	      {
		t->txudcptr = t->txudlptr = 0;
		DqEmpt (t->txudeq);
		t->txstat &= ~TS_UNDO;
	      }
	  }
      }
}

local void
setwindow (register wnd * w)
{
  register txt *t;
  t = w->wtext;
  Ttxt->txstat &= ~TS_PERM;
  if (!tmload (t))
    {
      ErFerr (t->finame);
      Ttxt->txstat |= TS_PERM;
      return;
    }
  Ttxt = t;
  Twnd->wcx = Tx;
  Twnd->wcy = Ty;
  Tx = w->wcx;
  Ty = w->wcy;
  Ty = w->wcy = min (Ty, TxLines (Ttxt));
  Twnd = w;
}

void
wmexit ()
{
  register wnd *w;

#ifndef X11
  if (tmQZoom ())
    tmDestroyZoom ();
#endif
  if ((w = WdFree (Twnd)) != NIL)
    {
      WdShowboxes ();
      setwindow (w);
#if defined (X11) && defined (STATUS_IN_MAIN_WINDOW)
      linkStatusBarAndCurrentWnd ();
#endif /* X11 && STATUS_IN_MAIN_WINDOW */
#ifndef X11
      wredraw (w);
#endif
      return;
    }
  cpsave ();
#ifdef SPELL
  erTerm ();
#endif
#ifdef X11
  stufTerm ();
  xiDetach ();
#else
  scErase ();
#ifdef COLOR
  SetDefaultTextColor ();
#endif
  scDetach ();
#endif
  kbDetach ();
#ifdef COLOR
  SetDefaultTextColor ();
#endif
  exit (0);
}

local unsigned char fnterms[] =
  { MICROS_DIR_FIELDS_SEPARATOR, ' ', '\"', '\'', '>', '<', 0 };
local bool
qfnchar (small i)
{
  register unsigned char *p, c, c1;
  c = ((unsigned char) Lebuf[i]);
  for (p = fnterms; c1 = *p++;)
    if (c1 == c)
      return (FALSE);
  return (TRUE);
}

local char *
fenter ()
{
  register char *p;
  register tchar *tp;
  register small l, lm, rm, cl;
  char tmp[4];

  cl = TRUE;
  TxSetY (Ttxt, Ty);
  if (qTxDown (Ttxt))
    return NIL;
  l = TxTRead (Ttxt, Lebuf);
#ifdef LONG_FILE_NAME
  if (Ttxt->txstat & TS_DIRLST)
    rm = FULL_FILE_NAME_POSITION;
  else
#endif
    rm = (qmcd (Ttxt) ? MCD_RIGHT + 1 : Tx);
  if ((lm = rm) >= l)
    return NIL;
  while (rm < l && qfnchar (rm))
    rm++;
  if (rm >= l || (char) Lebuf[rm] != '>')
    cl = FALSE;
  while (lm > 0)
    if (!qfnchar (--lm))
      {
	lm++;
	break;
      }
  if (lm == 0 || (char) Lebuf[lm - 1] != '<')
    cl = FALSE;
  if ((l = rm - lm) == 0 || l >= MAXPATH)
    return NIL;
  p = fnbuf;
  if (cl)
    p = scpy (clibpath, p);
  while (--l >= 0)
    *p++ = (char) Lebuf[lm++];
  *p = 0;
  if (qmcd (Ttxt))
    {
      for (p = tmp, tp = Lebuf + MCD_LEFT + 1, l = 3; --l >= 0;)
	if ((cl = *tp++) != ' ')
	  *p++ = cl;
      *p++ = 0;
      dtos (tmp, &filrm);
      if (filrm < 0 || filrm > MAXLPAC)
	filrm = 0;
    }
  return fnbuf;
}

local small fterms[] = { TK_BREAK, TE_CR, 0 };
local char *
f2enter ()
{
  register char *p;
  register tchar *pt;
  register small l;

  for (p = fprompt, pt = f2buf; *p; *pt++ = ctotc (*p++));
  Lenter (f2buf, &f2len, lfprompt, fterms, TRUE);
  if (KbCode == TK_BREAK)
    return NIL;
  if ((l = f2len - lfprompt) <= 0)
    return NIL;
  for (pt = f2buf + lfprompt, p = fnbuf; l; l--)
    *p++ = (char) (*pt++);
  *p = 0;
  return fnbuf;
}

local void
infilnam ()
{
  register txt *t;
  register small c;
  t = Ttxt;
  if (!qmcd (t))
    {
      scbell ();
      return;
    }
  TxSetY (t, Ty);
  tleload ();
/* Lxlm = Lxle = MCD_LEFT;*/
  Lxlm = Lxle = MCD_RIGHT + 1;
  Lxrm = Lxre = MAXLPAC;
  Lx = MCD_RIGHT + 1;
  for (;;)
    {
      if (LExec (RQ_WAIT) != E_NOCOM)
	{
	  scbell ();
	  continue;
	}
      if ((c = KbCode) == TK_BREAK || c == TE_CR || c == TE_OUTFILE)
	break;
    }
  Lx = 0;
  tleunload ();
}

bool
qmcd (register txt * t)
{
  return t->txstat & TS_DIRFIL;
}

/*
 * Text manager
 */

local char tmFileName[MAXPATH];
char *
tmcheck (char *filnam)
{
  if (FsParse (filnam, tmFileName))
    return (tmFileName);
  return (NIL);
}

txt *
tmdesc (char *filnam, bool qundo)
{
  register txt *t;
  register small i;
  register char *p, *p1;
  register small nmlen, ddlen, extlen;
  register bool dirlist;

  if ((p = tmcheck (filnam)) == NIL)
    return (NIL);
  ddlen = Fsddlen;
  nmlen = Fsnmlen;
  extlen = Fsext;
  dirlist = Fsiswild;

  for (i = MAXFILES, t = texts; i--; t++)
    if ((t->txstat & TS_BUSY) && scmp (t->finame, p))
      if (t->txstat & TS_NEW)
	return (t);
      else if (FsCheck (t))
	return (t);
  while ((t = TxNew (qundo)) == NIL)
    tmswap (TRUE);
  blkmov (p, t->finame, MAXPATH);

  t->fiddl = ddlen;
  t->finml = nmlen;
  t->fiext = extlen;
  if (dirlist)
    {
      t->txstat |= TS_DIRLST;
      t->txredit = TXED_NONO;
    }
  else
    t->txredit = TXED_YES;
#ifdef UNIX
  if (t->finml >= 4 && scmp (DIREXT, t->finame + t->fiddl + t->finml - 4))
    t->txstat |= TS_DIRFIL;
#endif
#ifdef VMS
  {
    char *p;
    for (p = t->finame + t->fiddl; *p != '.'; p++);
    if (sncmp (p, DIREXT, sizeof (DIREXT) - 1))
      t->txstat |= TS_DIRFIL;
  }
#endif
  return (t);
}

bool
tmload (register txt * t)
{
  register fid f;
  register small i;
  register char *p;
  t->txstat |= TS_PERM;
  if (!(t->txstat & TS_FILE))
    {
      t->txy = 0;
      if (FsInfo (t))
	{
	  if (t->txstat & TS_DIRLST)
	    {
	      tmdirlst (t);
	      /*     */
	      t->flines = TxLines (t);
	    }
	  else
	    {
	      if (!(f = FsOpen (t->finame, FO_READ)))
		goto err;
	      if (!DqLoad (t->txdstk, f, t->fisize))
		{
		  FsClose (f);
		  goto err;
		}
	      if (!FsClose (f))
		goto err;

	      /*     */
	      t->flines = TxLines (t);
	    }
	}
      else if (errno == FE_NSF)
	{
	  /*     */
	  t->flines = 0;
	  tmnew (t);
	}
      else
	goto err;
    ok:
      t->txstat &= ~TS_CHANGE;
      t->txstat |= TS_FILE;
      if (t->txudeq != NIL)
	t->txstat |= TS_UNDO;
      if (qmcd (t))
	t->txrm = MCD_LEFT;
    }
  t->txlructr = 0;
  FsSetDef (t);
  for (i = MAXFILES, t = texts; i; t++, i--)
    t->txlructr++;

  return (TRUE);
err:
  t->txstat &= ~TS_PERM;
  return (FALSE);
}

void
tmnew (register txt * t)
{
  t->fiattr = 0;
  t->fitime = FsTime ();
  t->fisize = 0;
  t->txstat |= TS_NEW;
}

local deq *dd;
local void
dput (char *s)
{
  DqAddE (dd, s, slen (s));
}

local char dir[MAXPATH];
local void
tmdirlst (register txt * t)
{
  register bool q;
  register char *p0, *p1, *p;
  txt tt = *t;
  dd = t->txdstk;
  p0 = afbuf;
  *sncpy (t->finame, dir, t->fiddl) = 0;
#ifdef RUSSIAN
  p = scpy ("惑  ", p0);
  *scpy (t->finame, p) = 0;
  dput (p0);
#endif
#ifdef ENGLISH
  p = scpy ("Directory   ", p0);
  *scpy (t->finame, p) = 0;
  dput (p0);
#endif
#ifdef VMS
#ifdef RUSSIAN
  p = scpy ("˂Ճ    [-]*.*;*", p0);
#endif
#ifdef ENGLISH
  p = scpy ("Previous directory      [-]*.*;*", p0);
#endif
  while (p < p0 + FULL_FILE_NAME_POSITION)
    *p++ = ' ';
  *scpy ("[-]*.*;*", p) = 0;
  dput (p0);
#endif
  for (q = TRUE; FsSearch (&tt, q); q = FALSE)
    {
      if (!sncmp (fullname, dir, FsName - fullname))
	{
	  *p0 = 0;
	  dput (p0);
	  *sncpy (fullname, dir, FsName - fullname) = 0;
	  /*p = scpy( "  ", p0 );
	     p=scpy( dir, p  );
	     while(p<p0+FULL_FILE_NAME_POSITION )*p++=' ';
	     *scpy( dir, p ) = 0;
	     dput( p0 );      */
	}
      p = p0;
      p1 = FsName;
      while (p < p0 + MAXFILELN1 && *p1)
	*p++ = *p1++;
      while (p < p0 + MAXFILELN1)
	*p++ = ' ';
#ifdef VMS
      if (tt.fisize || tt.fitime || tt.fiattr)
	{
#endif
	  sprintf (p, "%8ld  ", tt.fisize);
	  p = p0 + MAXFILELN1 + 8 + 2;
	  p = sncpy (dttm (tt.fitime), p, 21);
	  *p++ = ' ';
	  *p++ = ' ';
	  p = attrtoa (p, tt.fiattr);
	  *p++ = ' ';
	  *p++ = ' ';
	  p = sncpy (FsUname, p, MAXUSRLN);
	  *p++ = ' ';
	  *p++ = ' ';
#ifdef VMS
	}
      else
	p = scpy (" no privilege for attempted operation", p);
#endif
#ifdef UNIX
      p = sncpy (FsGname, p, MAXGRPLN);
#endif
#ifdef LONG_FILE_NAME
      while (p < p0 + FULL_FILE_NAME_POSITION)
	*p++ = ' ';
      p1 = fullname;
      while (p < p0 + MAXLPAC - 1 && *p1)
	*p++ = *p1++;
#endif
      *p = '\0';
      dput (p0);
    }
}

bool
tmsave (register txt * t, bool qback)
{
  register char *p, *p1;
  register fid f;
  register small mode;
  register large filsize;
  if (t->txstat & TS_CHANGE)
    {
      if (!(t->txstat & TS_DIRLST))
	{
#ifdef TMDEBUG
	  dbprintf ("Save %-24s ", t->finame);
#endif
	  if (qback)
	    {
	      p1 = p = scpy (t->finame, tmFileName);
#ifdef UNIX
	      if (t->finml > (MAXFILELN - 3))
		p = tmFileName + t->fiddl + (MAXFILELN - 2);
	      *scpy (".bak", p) = 0;
#endif
#ifdef VMS
	      while (*--p != '.')
		if (*p == ']')
		  {
		    p = p1;
		    break;
		  }
	      *scpy (".BAK", p) = 0;
#endif
#ifdef TMDEBUG
	      dbprintf ("Delete .bak");
#endif
	      FsDelete (tmFileName);
#ifdef TMDEBUG
	      dbprintf ("Rename ");
#endif
	      if (!(t->txstat & TS_NEW))
		{
		  if (!FsRename (t->finame, tmFileName))
		    goto err;
		  t->txstat |= TS_NEW;
		}
	    }
	  TxSetY (t, 0L);
	  filsize = DqLen (t->txdstk);
	  if (filsize == 0L)
	    {
#ifdef TMDEBUG
	      dbprintf ("Delete ");
#endif
	      if (!(t->txstat & TS_NEW))
		{
		  if (!FsDelete (t->finame))
		    goto err;
		}
	    }
	  else
	    {
#ifdef TMDEBUG
	      dbprintf ("Open ");
#endif
#ifdef VMS
	      mode = FO_NEW;
#else
	      mode = (t->fisize > filsize || (t->txstat & TS_NEW))
		? FO_NEW : FO_WRITE;
#endif
	      if (!(f = FsOpen (t->finame, mode)))
		goto err;
#ifdef TMDEBUG
	      dbprintf ("Write ");
#endif
	      if (!DqSave (t->txdstk, f))
		{
		  FsClose (f);
		  goto err;
		}
	      t->fitime = FsTime ();
	      t->fisize = filsize;
#ifdef TMDEBUG
	      dbprintf ("Close ");
#endif
	      if (!FsClose (f))
		{
		  goto err;
		}
	      t->txstat &= ~TS_NEW;
#ifdef TMDEBUG
	      dbprintf ("Ok\n");
#endif
	    }
	}
      t->txstat &= ~(TS_CHANGE);
      stModification (STM_NO);
    }
  else
    t->txstat &= ~TS_NEW;

  return (TRUE);
err:
  ErFerr (t->finame);
#ifdef TMDEBUG
  dbprintf ("Error %d txstat %o \n", errno, t->txstat);
#endif
  return (FALSE);
}
bool
qFilSwap ()
{
  register small i;
  register txt *t;
  for (i = MAXFILES, t = texts; --i >= 0; t++)
    if (t->txstat && t->txudeq != NIL)
      udcut (t);
  if (DqFree () >= DqReqMem)
    return (TRUE);
/*if( fntdeq && DqDel( fntdeq ) )fntdeq=NIL;*/
#ifdef SPELL
  if (spdeq)
    erTerm ();
#endif
  return (tmswap (FALSE));
}
#define S_SFILE  001
#define S_EFILE  002
#define S_EUNDO  004
#define S_EDESC  010
local small memswr[] = {
  TS_BUSY | TS_FILE | TS_UNDO | TS_DIRLST, S_EFILE | S_EUNDO | S_EDESC,
  TS_BUSY | TS_FILE | TS_UNDO, S_EFILE,
  TS_BUSY | TS_UNDO, S_EUNDO | S_EDESC,
  TS_BUSY | TS_FILE | TS_UNDO | TS_WND | TS_DIRLST, S_EFILE | S_EUNDO,
  TS_BUSY | TS_FILE | TS_UNDO | TS_WND, S_EFILE,
  TS_BUSY | TS_UNDO | TS_WND, S_EUNDO,
  TS_BUSY | TS_FILE | TS_UNDO | TS_WND | TS_CHANGE | TS_DIRLST,
    S_SFILE | S_EFILE | S_EUNDO,
  TS_BUSY | TS_FILE | TS_UNDO | TS_WND | TS_CHANGE, S_SFILE | S_EFILE,
  TS_BUSY | TS_FILE | TS_UNDO | TS_WND | TS_DIRLST, S_EFILE | S_EUNDO,
  TS_BUSY | TS_FILE | TS_UNDO | TS_WND, S_EFILE,
  TS_BUSY | TS_UNDO | TS_WND, S_EUNDO,
  0
};
local small descswr[] = {
  TS_BUSY | TS_FILE | TS_UNDO | TS_DIRLST, S_EFILE | S_EUNDO | S_EDESC,
  TS_BUSY | TS_UNDO, S_EUNDO | S_EDESC,
  TS_BUSY | TS_FILE | TS_UNDO, S_EFILE | S_EUNDO | S_EDESC,
  0
};
bool
tmswap (bool freedesc)
{
  register small *pcode;
  register txt *t, *tm;
  register small i, lrum, mask;
  pcode = freedesc ? descswr : memswr;
  for (; mask = *pcode++; pcode++)
    {
      for (;;)
	{
	  for (i = MAXFILES, t = texts, tm = NIL, lrum = 0; i; t++, i--)
	    if ((t->txstat & ~TS_NEW) == mask && t->txlructr > lrum)
	      {
		lrum = t->txlructr;
		tm = t;
	      }
	  if ((t = tm) == NIL)
	    break;
	  lrum = *pcode;
#ifdef TMDEBUG
	  dbprintf ("Swapped %16s programm %o ", t->finame, lrum);
#endif
	  if (lrum & S_SFILE)
	    if (!tmsave (t, FALSE))
	      {
		t->txstat |= TS_ERR;
		continue;
	      }
	  if (lrum & S_EFILE)
	    {
	      DqEmpt (t->txustk);
	      DqEmpt (t->txdstk);
	      t->txy = 0;
	      t->txstat &= ~TS_FILE;
	    }
	  if (lrum & S_EUNDO)
	    {
	      t->txudcptr = t->txudlptr = 0;
	      DqEmpt (t->txudeq);
	      t->txstat &= ~TS_UNDO;
	    }
	  if (lrum & S_EDESC)
	    {
	      TxDel (t);
	      if (freedesc)
		return (TRUE);
	    }
#ifdef TMDEBUG
	  dbprintf ("given:%ld required:%ld\n", DqFree (), DqReqMem);
#endif
	  if (DqFree () >= DqReqMem)
	    return (TRUE);
	}
    }
  return (FALSE);
}
local char modfile_buf[MAXPATH + 15];
local char modfile_rep[MAXPATH + 15];
local small modfile_exits[] = { TK_BREAK, TE_CR, 0 };

#ifdef VMS
#include <types.h>
#include <stat.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif

local bool
tm_modfile ()
{
  register bool repeat;
  register char *pf, *p, *p1, *p2;
  register tchar *pt, *pt1;
  struct stat statbuf;
  register small attr, len, i, nmlen, fileln;
  small llen;
  tchar buf[MAXLPAC];

  if ((pf = fenter ()) == NIL)
    return FALSE;

  repeat = TRUE;
  while (repeat)
    {
      repeat = FALSE;
      if (!FsParse (pf, modfile_buf) || Fsiswild)
	return FALSE;
#ifdef UNIX
      if (Fsddlen == 1)
	if (access ("/", 02))
	  return FALSE;
	else
	  {
	    p = modfile_buf + Fsddlen - 1;
	    *p = 0;
	    if (access (modfile_buf, 02))
	      return FALSE;
	    *p = '/';
	  }
#endif
      if (stat (modfile_buf, &statbuf))
	return FALSE;
      attr = statbuf.st_mode;
      nmlen = Fsnmlen;
#ifdef LONG_FILE_NAME
      fileln = (nmlen <= MAXFILELN) ? nmlen : MAXFILELN;
#else
      fileln = MAXFILELN;
      p = modfile_buf + Fsddlen + nmlen;
      for (i = nmlen; i <= fileln; i++)
	*p++ = ' ';
#endif
      p = modfile_buf + Fsddlen + fileln;
      *p++ = ' ';
      *attrtoa (p, attr) = 0;
      for (p = modfile_buf, pt = Lebuf; *p;)
	*pt++ = ctotc (*p++);
      len = llen = p - modfile_buf;
      Lenter (NIL, &llen, Fsddlen, modfile_exits, FALSE);
      if (KbCode == TK_BREAK)
	return FALSE;

      p = modfile_rep;
      pt = Lebuf;
      for (i = llen; i; i--)
	*p++ = (*pt++) & 0377;
      *p = 0;
      p = modfile_buf + Fsddlen;
      p1 = modfile_rep + Fsddlen;

      *(modfile_buf + Fsddlen + nmlen) = 0;

      if (*p1 == ' ')
	{			/*   */
#ifdef UNIX
	  struct stat st;
	  if (stat (".", &st))
	    return FALSE;
#endif
	  if (FsDelete (modfile_buf))
	    {
	      if (Ttxt->txstat & TS_DIRLST)
		tedl ();
#ifdef UNIX
	      if (st.st_mtime <= Ttxt->fitime)
		Ttxt->fitime = FsTime ();
#endif
	      return TRUE;
	    }
	  else
	    return FALSE;
	}
      for (i = 0; i < MAXFILELN && p1[i] != ' '; i++);
      p1[i] = 0;
      /*       */
      p2 = p1 + i + 1;
      while (*p2 == ctotc (' '))
	p2++;
      switch (tmChAttr (p2, attr, modfile_buf))
	{
	case TRUE:
	  repeat = TRUE;	/*     */
	case FALSE:
	  break;
	case 2:
	  return FALSE;
	}
      /*   */
      for (i = 0; p1[i]; i++)
	if (p1[i] == '/')
	  {
	    repeat = TRUE;
	    continue;
	  }
      if (!scmp (p, p1))
	{
	  if ((pf = FsRename (modfile_buf, modfile_rep)) == NIL)
	    return FALSE;
	  if (Ttxt->txstat & TS_DIRLST)
	    {
	      TxFRead (Ttxt, buf);
	      for (p1 = pf + Fsddlen, pt = buf, i = 0;
		   *p1 && i < MAXFILELN1; *pt++ = ctotc (*p1++), i++);
	      for (; i < MAXFILELN1; *pt++ = ctotc (' '), i++);
#ifdef LONG_FILE_NAME
	      for (p1 = pf, pt = buf + FULL_FILE_NAME_POSITION, i =
		   FULL_FILE_NAME_POSITION; *p1; *pt++ = *p1++, i++);
	      for (; i < MAXLPAC; *pt++ = ctotc (' '), i++);
#endif
	      TxFRep (Ttxt, buf);

	    }
	}
      if (repeat)
	scbell ();
    }
  return TRUE;
}

local bool
tmChAttr (char *p, int attr, char *modfile_buf)
{
  register small new_attr = 0, mask, i;
  register char *p1, *ps;
  tchar buf[MAXLPAC], *pt;
  bool repeat = FALSE;
  struct stat statbuf;

  ps = p;
  p++;
  if (*p == 's')
    new_attr |= 04000;
  else if (*p != '-')
    {
      repeat = TRUE;
      new_attr |= attr & 04000;
    }
  p++;
#ifdef UNIX
  if (*p == 's' && *(p + 7) == 'x')
    new_attr |= 02010;
#endif
#ifdef VMS
  if (*p == 's' && *(p + 7) == 'e')
    new_attr |= 02010;
#endif
  else if (*p == 'l' && *(p + 7) == '-')
    new_attr |= 02000;
  else if (*p != '-')
    {
      repeat = TRUE;
      new_attr |= attr & 02000;
      *(p + 7) = '?';
    }
  p++;
  if (*p == 't')
    new_attr |= 01000;
  else if (*p != '-')
    {
      repeat = TRUE;
      new_attr |= attr & 01000;
    }
  p++;
  for (mask = 0400, p1 = MASK; mask; mask >>= 1, p++, p1++)
    if (*p == *p1)
      new_attr |= mask;
    else if (*p != '-')
      {
	repeat = TRUE;
	new_attr |= attr & mask;
      }
  if (!repeat && (new_attr != (attr & 07777)))
    if (chmod (modfile_buf, (int) new_attr) == 0)
      {
	if (Ttxt->txstat & TS_DIRLST)
	  {
	    TxFRead (Ttxt, buf);
	    if (stat (modfile_buf, &statbuf) == 0)
	      new_attr = statbuf.st_mode;
	    *attrtoa (ps, new_attr) = 0;
	    for (pt = buf + MAXFILELN1 + 8 + 2 + 21 + 2;
		 *ps; *pt++ = ctotc (*ps++), i++);
	    TxFRep (Ttxt, buf);
	  }
	return FALSE;
      }
    else
      return 2;
  return repeat;
}
