/*
 *     nanoMir
 *
 *     "ud.c"           Undo
 */

#include "nm.h"
#include "fs.h"
#include "dq.h"
#include "tx.h"
#include "le.h"
#include "te.h"
#include "cm.h"
#include "ud.h"
#include "kb.h"
#ifndef X11
#include "sc.h"
#else
#include "xi.h"
#endif

/*#define UDEBUG*/

typedef struct
{
  small utyp;
  large uytext;
}
textlundo;

typedef struct
{
  small utyp;
  large uytext;
  small uslen;
/* old/new string */
}
text1undo;

typedef struct
{
  small utyp;
  large uytext;
  small udindex;
  small unslen;
  small uoslen;
/* new string */
/* old string */
}
text2undo;

typedef struct
{
  small utyp;
  small ulx;
  small ulxl, ulxr, uldx;
  small ulnsl, ulosl;
}
lineundo;

bool UdMark;
static char *ubuf, *eubuf;
small undodir;

local void tsundo (bool, bool), tsunundo (bool, bool), genundo (void),
repsubstr (void), undogadd (txt *, small);
local small midind (char *, small *, char *, small *);

void
udInit ()
{
  ubuf = GetMain (UBUFSIZ);
  undodir = 0;
  UdMark = FALSE;
}

void
tundoload (register txt * t)
{
  register char *p;

  if ((undodir == 0) && (t->txudeq))
    {
      p = ubuf;
      ((textlundo *) p)->uytext = t->txy;
      eubuf = p + sizeof (textlundo);
      undogadd (t, UT_LOAD);
    }
}

void
tundounload (register txt * t)
{
  if ((undodir == 0) && (t->txudeq))
    {
      DqCutE (t->txudeq, t->txudlptr);
      t->txudcptr = t->txudlptr;
    }
  UdMark = TRUE;
#ifdef UDEBUG
  dbprintf ("UndoUnload %ld %ld %ld\n", t->txudlptr, t->txudcptr,
	    DqLen (t->txudeq));
#endif
}

void
tundo1add (register txt * t, small typ, char *a, small l)
{
  register char *p;

  if ((undodir == 0) && (t->txudeq))
    {
      p = ubuf;
      ((text1undo *) p)->uytext = t->txy;
      ((text1undo *) p)->uslen = l;
      p += sizeof (text1undo);
      eubuf = blkmov (a, p, l);
      undogadd (t, typ);
    }
}

void
tundo2add (register txt * t, char *ao, small lo, char *an, small ln)
/*  char *ao; small lo;
 *  char *an; small ln;
 */
{
  register char *p;
  register small k;

  if ((undodir == 0) && (t->txudeq))
    {
      k = midind (ao, &lo, an, &ln);
      p = ubuf;
      ((text2undo *) p)->uytext = t->txy;
      ((text2undo *) p)->udindex = k;
      ((text2undo *) p)->unslen = ln;
      ((text2undo *) p)->uoslen = lo;
      p += sizeof (text2undo);
      p = blkmov (an + k, p, ln);
      eubuf = blkmov (ao + k, p, lo);
      undogadd (t, UT_REP);
    }
}

void
lundoadd (register txt * t, register small xl, register small xr, small dx,
	  char *os, char *ns)
/* register small xl, xr;
 * small dx;
 * tchar *os, *ns;
 */
{
  register small slen;
  register char *p;
#define pl  ((lineundo *)p)

  if ((undodir == 0) && (t) && (t->txudeq))
    {
      slen = (dx == REPLACE ? xr - xl : ABS (dx));
      p = ubuf;
      pl->ulx = Lx;
      pl->ulxl = xl;
      pl->ulxr = xr;
      pl->uldx = dx;
      pl->ulnsl = xl = lstrlen (slen, (tchar *) ns);
      pl->ulosl = xr = lstrlen (slen, (tchar *) os);
      p += sizeof (lineundo);
      p = blkmov (ns, p, xl * sizeof (tchar));
      eubuf = blkmov (os, p, xr * sizeof (tchar));
      undogadd (t, UL_GEN);
    }
#undef pl
}

static void
undogadd (register txt * t, register small typ)
{
  register deq *d;
  register large l;

  ((text1undo *) ubuf)->utyp = (UdMark ? typ | U_MARK : typ);
  UdMark = FALSE;
  l = t->txudcptr;
  if (typ <= UT_LOAD)
    {
      if (t->txudlptr < l)
	l = t->txudlptr;
      t->txudlptr = l;
    }
  d = t->txudeq;
  DqCutE (d, l);
  DqAddE (d, ubuf, (small) (eubuf - ubuf));
  t->txudcptr = l = DqLen (d);
  if (typ < UT_LOAD)
    t->txudlptr = l;
#ifdef UDEBUG
  dbprintf ("UndoAdd %d  %ld %ld %ld\n", typ, t->txudlptr, t->txudcptr,
	    DqLen (d));
#endif
}

static small
midind (char *ao, register small * plo, char *an, register small * pln)
{
  register char *p1, *p2;
  register small i, j;

  for (p1 = ao, p2 = an, j = min (*plo, *pln); j; j--)
    if (*p1++ != *p2++)
      {
	p1--;
	break;
      }
  if ((j == 0) && (*plo == *pln))
    {
      *plo = 0;
      *pln = 0;
      return (0);
    }
  i = p1 - ao;
  for (p1 = ao + *plo, p2 = an + *pln; j; j--)
    if (*--p1 != *--p2)
      {
	p1++;
	p2++;
	break;
      }
  *plo = (p1 - ao) - i;
  *pln = (p2 - an) - i;
  return (i);
}

static small usl, uosl, unsl;
static char *usptr, *unsptr;

static void
repsubstr ()
{
  if (uosl != unsl)
    {
      usl -= uosl;
      blkmov (usptr + uosl, usptr + unsl, usl);
      usl += unsl;
    }
  blkmov (unsptr, usptr, unsl);
}

void
teundo ()
{
  tsundo (FALSE, FALSE);
}
void
tesundo ()
{
  tsundo (TRUE, FALSE);
}
void
teunundo ()
{
  tsunundo (FALSE, FALSE);
}
void
tesunundo ()
{
  tsunundo (TRUE, FALSE);
}
void
leundo ()
{
  tsundo (FALSE, TRUE);
}
void
lesundo ()
{
  tsundo (TRUE, TRUE);
}
void
leunundo ()
{
  tsunundo (FALSE, TRUE);
}
void
lesunundo ()
{
  tsunundo (TRUE, TRUE);
}


static void
tsundo (bool slow, bool lud)
{
  register deq *d;
  register txt *t;
  register small typ;
  t = lud ? Ltxt : Ttxt;
  if (t && (d = t->txudeq) && t->txudcptr)
    {
      while (t->txudcptr != 0)
	{
	  qKbReq ();
	  DqCopBack (d, t->txudcptr, ubuf);
	  typ = ((text1undo *) ubuf)->utyp;
	  if (lud && (typ & (~U_MARK)) != UL_GEN)
	    {
	      Lchange = FALSE;
	      exc (E_UP);
	    }
	  t->txudcptr -= DqRLen;
	  undodir = (-1);
	  genundo ();
	  if (!((typ & (~U_MARK)) == UT_LOAD) && (slow || (typ & U_MARK)))
	    break;
	}
#ifdef UDEBUG
      dbprintf ("Undo     %ld %ld %ld\n", t->txudlptr, t->txudcptr,
		DqLen (t->txudeq));
#endif
    }
  else
    exc (E_NOUNDO);
}

static void
tsunundo (bool slow, bool lud)
{
  register txt *t;
  register deq *d;
  register small typ;
  register small ctr;
  t = lud ? Ltxt : Ttxt;
  if (t && (d = t->txudeq) && (t->txudcptr != DqLen (d)))
    {
      ctr = 0;
      while (t->txudcptr != DqLen (d))
	{
	  DqCopForw (d, t->txudcptr, ubuf);
	  typ = ((text1undo *) ubuf)->utyp;
	  if (ctr && (typ & (~U_MARK)) != UT_LOAD && (slow || (typ & U_MARK)))
	    break;
	  if ((!lud) && (typ & (~U_MARK)) == UL_GEN)
	    exc (E_DOWN);
	  t->txudcptr += DqRLen;
	  undodir = 1;
	  genundo ();
	  ctr++;
	}
#ifdef UDEBUG
      dbprintf ("UnUndo     %ld %ld %ld\n", t->txudlptr, t->txudcptr,
		DqLen (t->txudeq));
#endif
    }
  else
    exc (E_NOUNDO);
}

static void
genundo ()
{
  register char *p;
#define pt1 ((text1undo *)p)
#define pt2 ((text2undo *)p)
#define pl  ((lineundo *)p)
  register tchar *ptc;
  register small i, dx, len;
#ifdef UDEBUG
  register small k;
#endif
  p = ubuf;
  i = pt1->utyp & (~U_MARK);
  if (i != UL_GEN)
    {
      qsety (pt1->uytext);
      switch (i)
	{
	case UT_IL:
	  if (undodir < 0)
	    goto dl;
	il:TxIL (Ttxt, ubuf + sizeof (text1undo), pt1->uslen);
	  break;
	case UT_DL:
	  if (undodir < 0)
	    goto il;
	dl:TxDL (Ttxt);
	  break;
	case UT_REP:
	  i = pt2->udindex;
	  usl = TxRead (Ttxt, afbuf) - i;
	  usptr = afbuf + i;
	  unsptr = p + sizeof (text2undo);
	  if (undodir > 0)
	    {
	      uosl = pt2->uoslen;
	      unsl = pt2->unslen;
	    }
	  else
	    {
	      uosl = pt2->unslen;
	      unsl = pt2->uoslen;
	      unsptr += uosl;
	    }
#ifdef UDEBUG
	  dbprintf ("New string#");
	  for (k = 0; k < unsl; k++)
	    dbprintf ("%c", unsptr[k]);
	  dbprintf ("#\n");

	  dbprintf ("before #");
	  for (k = 0; k < usl + i; k++)
	    dbprintf ("%c", afbuf[k]);
	  dbprintf ("#\n");
#endif
	  repsubstr ();
#ifdef UDEBUG
	  dbprintf ("after  #");
	  for (k = 0; k < usl + i; k++)
	    dbprintf ("%c", afbuf[k]);
	  dbprintf ("#\n");
#endif
	  TxRep (Ttxt, afbuf, usl + i);
	  break;
	case UT_LOAD:

	  break;
	default:
	  otkaz ("Undo");
	}
    }
  else
    {
      Lx = pl->ulx;
      dx = pl->uldx;
      len = pl->ulnsl;
      unsptr = p + sizeof (lineundo);
      i = (dx == REPLACE) ? (pl->ulxr - pl->ulxl) : ABS (dx);
      len = pl->ulnsl;
      unsptr = p + sizeof (lineundo);
      if (undodir < 0)
	{
	  if (dx != REPLACE)
	    dx = -dx;
	  unsptr += len * sizeof (tchar);
	  len = pl->ulosl;
	}
      for (ptc = ((tchar *) unsptr) + len; len < i; len++)
	*ptc++ = (tchar) ' ';
      llmove (pl->ulxl, pl->ulxr, dx, (tchar *) unsptr);
      Lchange = TRUE;
    }
  undodir = 0;
}

bool
udcut (register txt * t)
{
  register deq *d;
  register large l, m;
  if ((d = t->txudeq) != NIL && (l = DqLen (d)) > UDHQUOTA)
    {
      if ((m = min (t->txudcptr, t->txudlptr)) > 0)
	{
	  if ((l -= UDLQUOTA) < m)
	    m = l;
	  for (l = 0; l < m;)
	    {
	      DqCopForw (d, l, NIL);
	      l += DqRLen;
	    }
	  t->txudcptr -= l;
	  t->txudlptr -= l;
	  DqCutB (d, l);
	  return (TRUE);
	}
    }
  return (FALSE);
}
