/*
 *     nanoMir
 *
 *     "wd.c"           Windows
 */

#include "nm.h"
#include "tx.h"
#include "wd.h"
#ifndef X11
#include "sc.h"
#else
#include "xi.h"
#endif

small wxmin, wxmax, wdx, wymin, wymax, wdy;
wnd windows[MAXWNDWS];
static small x0, x1, y0, y1;

local void wrdrw01 (wnd *), wtxblt01 (wnd *), mergeemp (wnd *);
local bool wdmerge (wnd *, wnd *);

wnd *
wdWindow (large wd)
{
  return windows + wd;
}
large
wdDesc (wnd * w)
{
  return w - windows;
}

void
wdInit ()
{
  register wnd *w;
  register small i;
/* windows=GetMain(MAXWNDWS*sizeof(wnd));       */
  for (i = MAXWNDWS, w = windows; i; i--, w++)
    w->wstat = 0;
}

wnd *
WdNew (small typ, small num)
{
  register wnd *w;
  register small i;
  switch (typ)
    {
    case WND_NORMAL:
      for (i = MAXWNDWS - 2, w = windows; i; i--, w++)
	if (w->wstat == 0)
	  {
	    w->wstat |= WS_BUSY | WS_NOTEXT;
	    w->wtyp = typ;
	    return (w);
	  }
      return (NIL);
    case WND_SPECIAL:
    case WND_STATUS:
    case WND_ERROR:
      w = windows + MAXWNDWS - num;
      if (w->wstat & WS_BUSY)
	return NIL;
      w->wstat |= WS_BUSY | WS_NOTEXT;
      w->wtyp = typ;
      return w;
    default:
      return NIL;
    }
}

wnd *
WdFree (register wnd * w)
{
  register wnd *w1;
  register small i, j;

  w->wstat &= ~WS_BUSY;
  if (w->wtyp == WND_NORMAL)
    {
#ifndef X11
      for (i = MAXWNDWS, w1 = windows; i; i--, w1++)
	if ((w1->wstat & WS_BUSY) && !(w1->wstat & WS_DEBUG) &&
	    wdmerge (w, w1))
	  {
	    mergeemp (w1);
	    return w1;
	  }
      mergeemp (w);
      for (i = MAXWNDWS, w1 = windows; i; i--, w1++)
	if ((w1->wstat & WS_BUSY) && !(w1->wstat & WS_DEBUG) &&
	    wdmerge (w, w1))
	  {
	    mergeemp (w1);
	    return w1;
	  }
      if (!(w->wstat & WS_ERROR))
	{
	  w->wstat |= (WS_EMPTY | WS_BUSY);
	  w->winf = NoInfo;
	  w_clear (w);
	}
      for (i = MAXWNDWS, w1 = windows; i; i--, w1++)
	if (((j = w1->wstat) & WS_BUSY) && !(j & WS_DEBUG) && !(j & WS_EMPTY))
	  return w1;
#else
      xiDestroyWindow (w);
      for (i = MAXWNDWS, w1 = windows; i; i--, w1++)
	if ((w1->wstat & WS_BUSY) && !(w1->wstat & WS_DEBUG))
	  return w1;
#endif
      return NIL;
    }
  else
    {
#ifdef X11
      xiDestroyWindow (w);
#endif
      return NIL;
    }
}

#ifndef X11
static bool
wdmerge (register wnd * w, register wnd * w1)
{
  if (w1->wsh == w->wsh && w1->wsy == w->wsy)
    {
      if (w1->wsx + w1->wsw + 1 == w->wsx)
	{
	  w1->wsw += w->wsw + 1;
	  return TRUE;
	}
      if (w->wsx + w->wsw + 1 == w1->wsx)
	{
	  w1->wsw += w->wsw + 1;
	  w1->wsx = w->wsx;
	  return TRUE;
	}
    }
  if (w1->wsw == w->wsw && w1->wsx == w->wsx)
    {
      if (w1->wsy + w1->wsh + 1 == w->wsy)
	{
	  w1->wsh += w->wsh + 1;
	  return TRUE;
	}
      if (w->wsy + w->wsh + 1 == w1->wsy)
	{
	  w1->wsh += w->wsh + 1;
	  w1->wsy = w->wsy;
	  return TRUE;
	}
    }
  return FALSE;
}

static void
mergeemp (register wnd * w)
{
  register wnd *w1;
  register int i, j;
restart:
  for (i = MAXWNDWS, w1 = windows; i; i--, w1++)
    if (((j = w1->wstat) & WS_BUSY) && !(j & WS_DEBUG) &&
	(j & WS_EMPTY) && wdmerge (w1, w))
      {
	w1->wstat &= ~WS_BUSY;
	goto restart;
      }
}
#endif

void
WdShowboxes ()
{
#ifndef X11
  register wnd *w;
  register small i, j;
  for (i = MAXWNDWS, w = windows; i; i--, w++)
    if (((j = w->wstat) & WS_BUSY) && !(j & WS_DEBUG) && !(j & WS_SHADOW))
      scdrwnd (w->wsx, w->wsy, w->wsw, w->wsh);
#endif
}
#if 0
wnd *
WdNext (register small x, register small y)
{
  register wnd *w;
  register small i, j;
  for (i = MAXWNDWS, w = windows; i; i--, w++)
    {
      if (((j = w->wstat) & WS_BUSY) && !(j & WS_DEBUG) && !(j & WS_EMPTY))
	{
	  if (w->wsx <= x && w->wsx + w->wsw > x &&
	      w->wsy <= y && w->wsy + w->wsh > y)
	    return (w);
	}
    }
  return (NIL);
}
#endif

/*      */
wnd *
WdGetNext (wnd * w_curr)
{
  /*    */
  int id = wdDesc (w_curr);
  wnd *w;
  int i, j, wstat;

  if (id < 0 || id >= MAXWNDWS)
    return 0;	/*   -    */

  /*    ,    */
  for (j = 1; j < MAXWNDWS; j++)
    {
      i = (id + j) % MAXWNDWS;	/*    */
      w = windows + i;		/*    */
      wstat = w->wstat; 	/*   */

      if ((wstat & WS_BUSY) && !(wstat & WS_DEBUG)
	   && !(wstat & WS_EMPTY) && !(wstat & WS_ERROR))
	{
	  /*   */
#ifdef X11
	  if (w->wnd_shown)	/*     */
#endif
	  return w;
	}
    }
  return 0;	/*   */
}

/*      */
wnd *
WdGetPrev (wnd * w_curr)
{
  /*    */
  int id = wdDesc (w_curr);
  wnd *w;
  int i, j, wstat;

  if (id < 0 || id >= MAXWNDWS)
    return 0;	/*   -    */

  /*    ,    */
  for (j = MAXWNDWS - 1; j > 0; j--)
    {
      i = (id + j) % MAXWNDWS;	/*    */
      w = windows + i;		/*    */
      wstat = w->wstat; 	/*   */

      if ((wstat & WS_BUSY) && !(wstat & WS_DEBUG)
	   && !(wstat & WS_EMPTY) && !(wstat & WS_ERROR))
	{
	  /*   */
#ifdef X11
	  if (w->wnd_shown)	/*     */
#endif
	  return w;
	}
    }
  return 0;	/*   */
}

wnd *
WdNear (wnd * ww, small x, small y, small dx, small dy)
{
  register wnd *w;
  register small i, j;
  small ddx, ddy, tx, ty;
  wnd *wn;

  ddx = ddy = 1000;
  wn = NIL;
  for (i = MAXWNDWS, w = windows; i; i--, w++)
    {
      if ((w != ww) && ((j = w->wstat) & WS_BUSY) && !(j & WS_DEBUG) &&
	  !(j & WS_EMPTY) && !(j & WS_ERROR))
	{
	  if ((dx == 0) && ((w->wsx > x) || (w->wsx + w->wsw <= x)))
	    continue;
	  if ((dy == 0) && ((w->wsy > y) || (w->wsy + w->wsh <= y)))
	    continue;
	  tx = dx * (w->wsx - x);
	  ty = dy * (w->wsy - y);
	  if (tx < 0 || ty < 0 || tx > ddx || ty > ddy)
	    continue;
	  ddx = tx;
	  ddy = ty;
	  wn = w;
	}
    }
  return (wn);
}

void
wndop (small op, register txt * t)
{
  register large y;
  register wnd *w;
  register small i;
  y = t->txy;
  for (w = (wnd *) t->txwndptr; w != NIL; w = w->wnext)
    if (!(w->wstat & WS_SHADOW))
      {
	wxmin = 0;
	wxmax = w->wsw;
	wdx = 0;
	wymin = 0;
	wymax = w->wsh;
	wdy = 0;
	switch (op)
	  {
	  case TW_EM:
	    w->wty = 0;
	    w->wtx = 0;
	    wrdrw (w);
	    break;
	  case TW_IL:
	    if (y < w->wty)
	      w->wty++;
	    else if (y >= w->wty + w->wsh);	/* noop */
	    else
	      {
		wymin = y - w->wty;
		wdy++;
		wroll (w);
	      }			/* Roldown */
	    break;
	  case TW_DL:
	    if (y < w->wty)
	      w->wty--;
	    else if (y >= w->wty + w->wsh);	/* noop */
	    else
	      {
		wymin = y - w->wty;
		wdy--;
		wroll (w);
	      }			/* Rolup */
	    break;
	  case TW_RP:
	    if ((y >= w->wty) && (y < w->wty + w->wsh))
	      {
		wymin = y - w->wty;
		wymax = wymin + 1;
		wrdrw (w);	/* Current line */
	      }
	    break;
	  default:
	    otkaz ("Wop");
	  }
#ifdef X11
	wpos (w, w->wxcx + w->wtx, w->wxcy + w->wty);
#endif
      }
}

/*
 * Redraw window ( w )
 */
void
wredraw (register wnd * w)
{
  if (w->wstat & WS_SHADOW)
    return;
  x0 = 0;
  x1 = w->wsw;
  y0 = 0;
  y1 = w->wsh;
  wrdrw01 (w);
}

void
wadjust (register wnd * w, small x, large y)
{
  register small ddx;
  register large ddy;
  register small d;

  if (((ddx = x - w->wtx) >= 0) && ((ddx = ddx - w->wsw + 1) <= 0))
    ddx = 0;

#if (ROLL_LINE > 1)
  ddy = y - w->wty;

  if (ddy < ROLL_LINE - 1)
    {
      /* Looks like we are going to scroll up */
      ddy -= min (ROLL_LINE - 1, y);
    }
  else
    {
      /* Looks like we are going to scroll down */
      large text_length = 0;

      ddy = ddy - w->wsh;
      if (w->wtext)
	text_length = TxLines (w->wtext);

      if (text_length)
	ddy += min (ROLL_LINE, text_length - y + 1);
      else
	ddy++;

      if (ddy < 0)
	ddy = 0;
    }
#else
  if (((ddy = y - w->wty) >= 0) && ((ddy = ddy - w->wsh + 1) <= 0))
    ddy = 0;
#endif

  d = w->wsw >> 2;
  if (ddx > 0)
    {
      if (ddx < d)
	ddx = d;
    }
  d = -d;
  if (ddx < 0)
    {
      if (ddx > d)
	ddx = d;
      if (w->wtx + ddx < 0)
	ddx = -w->wtx;
    }
  wxmin = 0;
  wxmax = w->wsw;
  wdx = -ddx;
  wymin = 0;
  wymax = w->wsh;
  wdy = -ddy;
  w->wtx += ddx;
  w->wty += ddy;
  if (ddx < 0)
    ddx = -ddx;
  if (ddy < 0)
    ddy = -ddy;
  if ((ddx == 0) && (ddy == 0));
  else if ((ddx > w->wsw - 1) || (ddy > w->wsh - 1))
    wrdrw (w);
  else
    wroll (w);
#ifdef X11
  if (!xiLocked)
    xiFlush ();
#else
  if (!ScLocked)
    tv_flush ();
#endif
}

bool
qwclip (register wnd * w)
{
  if (wxmin < 0)
    wxmin = 0;
  if (wxmax > w->wsw)
    wxmax = w->wsw;
  if (wymin < 0)
    wymin = 0;
  if (wymax > w->wsh)
    wymax = w->wsh;
  return (wxmin < wxmax && wymin < wymax);
}

void
wrdrw (wnd * w)
{
  if (qwclip (w))
    {
      x0 = wxmin;
      x1 = wxmax;
      y0 = wymin;
      y1 = wymax;
      wrdrw01 (w);
    }
}

void
wroll (register wnd * w)
{
  if (w->wstat & WS_SHADOW)
    return;
  if (qwclip (w) && ((wdx != 0) || (wdy != 0)))
    {
/*
 * Box0 := Box n ( Box - vector )
 */
      if (wdx >= 0)
	{
	  x0 = wxmin;
	  x1 = wxmax - wdx;
	}
      else
	{
	  x0 = wxmin - wdx;
	  x1 = wxmax;
	}
      if (wdy >= 0)
	{
	  y0 = wymin;
	  y1 = wymax - wdy;
	}
      else
	{
	  y0 = wymin - wdy;
	  y1 = wymax;
	}
      if (x0 < x1 && y0 < y1)
	{
	  wtxblt01 (w);
	  if (wdy)
	    {
	      if (wdy > 0)
		{
		  y0 = wymin;
		  y1 = wymin + wdy;
		}
	      else
		{
		  y0 = wymax + wdy;
		  y1 = wymax;
		}
	      x0 = wxmin;
	      x1 = wxmax;
	      wrdrw01 (w);
	    }
	  if (wdx)
	    {
	      if (wdx > 0)
		{
		  x0 = wxmin;
		  x1 = wxmin + wdx;
		}
	      else
		{
		  x0 = wxmax + wdx;
		  x1 = wxmax;
		}
	      if (wdy > 0)
		{
		  y0 = wymin + wdy;
		  y1 = wymax;
		}
	      else
		{
		  y0 = wymin;
		  y1 = wymax + wdy;
		}
	      wrdrw01 (w);
	    }
	}
      else
	wrdrw (w);
    }
}

/*
 *   Simple operations in window
 */

void
wpos (register wnd * w, small x, small y)
{
  x -= w->wtx;
  y -= w->wty;
#ifndef X11
  scpos (w->wsx + x, w->wsy + y);
#else
  xiPos (w, x, y);
#endif
}

void
w_clear (register wnd * w)
{
#ifndef X11
  scclear (w->wsx, w->wsx + w->wsw, w->wsy, w->wsy + w->wsh);
#else
  xiClear (w);
#endif
}

void
wblock (register wnd * w, box b)
{
  large m = 0;
  box win;
  if (b.x0 > b.x1)
    {
      m = b.x0;
      b.x0 = b.x1;
      b.x1 = m;
    }
  if (b.y0 > b.y1)
    {
      m = b.y0;
      b.y0 = b.y1;
      b.y1 = m;
    }
  b.x1++;
  b.y1++;
  win.x0 = w->wsx;
  win.y0 = w->wsy;
  win.x1 = w->wsx + w->wsw;
  win.y1 = w->wsy + w->wsh;
  b.x0 -= w->wtx;
  b.x1 -= w->wtx;
  b.y0 -= w->wty;
  b.y1 -= w->wty;
  if (b.x0 < 0)
    b.x0 = 0;
  if (b.x0 > w->wsw)
    b.x0 = w->wsw;
  if (b.x1 < 0)
    b.x1 = 0;
  if (b.x1 > w->wsw)
    b.x1 = w->wsw;
  if (b.y0 < 0)
    b.y0 = 0;
  if (b.y0 > w->wsh)
    b.y0 = w->wsh;
  if (b.y1 < 0)
    b.y1 = 0;
  if (b.y1 > w->wsh)
    b.y1 = w->wsh;
#ifndef X11
  b.x0 += w->wsx;
  b.x1 += w->wsx;
  b.y0 += w->wsy;
  b.y1 += w->wsy;
  scblkon (b, win);
#else
  xiBlkOn (w, b);
#endif
}

void
wblkoff (register wnd * w)
{
#ifndef X11
  scblkoff ();
#else
  xiBlkOff (w);
#endif
}

#ifndef X11
static void
wrdrw01 (register wnd * w)
{
  register small y, m;
  register tchar *p;
  small l;
  for (y = y0; y < y1; y++)
    {
      p = (*(w->winf)) (w, y, &l) + x0;
      if ((m = l - x0) < 0)
	m = 0;
      if (m > x1 - x0)
	m = x1 - x0;
      if (m)
	sctext (p, m, w->wsx + x0, w->wsy + y);
      scspaces (w->wsx + x0 + m, w->wsy + y, x1 - x0 - m);
    }
}
#else /* def X11 */
static void
wrdrw01 (register wnd * w)
{
  if (!(w->wxstat & WDX_REDRAW))
    w->wxstat |= WDX_REDRAW;
  else
    wrdrwcallback (w, x0, x1, y0, y1);
/* else xiRedraw( w, x0, x1, y0, y1 ); */
}

void
wrdrwcallback (register wnd * w, small x0, small x1, small y0, small y1)
{
  register small y, m;
  register tchar *p;
  small l;
  for (y = y0; y < y1; y++)
    {
      p = (*(w->winf)) (w, y, &l) + x0;
      if ((m = l - x0) < 0)
	m = 0;
      if (m > x1 - x0)
	m = x1 - x0;
      if (m)
	xiText (w, x0, y, p, m);
      xiSpaces (w, x0 + m, y, x1 - x0 - m);
    }
}
#endif /* def X11 */

static void
wtxblt01 (register wnd * w)
{
#ifndef X11
  sctxblt (w->wsx + x0, w->wsx + x1, w->wsy + y0, w->wsy + y1, wdx, wdy);
#else
  xiTxblt (w, x0, y0, x1, y1, wdx, wdy);
#endif
}
